"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NotAFunction = exports.UnreachableCode = exports.InvalidCall = exports.InvalidOverload = exports.UnexpectedType = exports.NotAValidType = exports.CannotInferReturnType = exports.TypeMismatch = exports.LysCompilerError = exports.LysSemanticError = exports.LysTypeError = exports.LysScopeError = exports.AstNodeError = exports.PositionCapableError = exports.printDebugType = exports.isSamePositionOrInside = exports.isSamePosition = exports.DEBUG_TYPES = void 0;
const types_1 = require("./types");
exports.DEBUG_TYPES = process.env.DEBUG_TYPES === '1' || process.env.DEBUG_TYPES === 'true';
function isSamePosition(lhs, rhs) {
    return lhs === rhs || (lhs.moduleName === rhs.moduleName && lhs.end === rhs.end && lhs.start === rhs.start);
}
exports.isSamePosition = isSamePosition;
function isSamePositionOrInside(parent, child) {
    return (isSamePosition(parent, child) ||
        (parent.moduleName === child.moduleName && parent.end >= child.end && parent.start <= child.start));
}
exports.isSamePositionOrInside = isSamePositionOrInside;
function printDebugType(type) {
    if (type) {
        return `"${type}"${exports.DEBUG_TYPES ? ' ' + type.inspect(1) : ''}`;
    }
    return `<Type ${type}>`;
}
exports.printDebugType = printDebugType;
class PositionCapableError extends Error {
    constructor(message, position, warning = false) {
        super(message);
        this.message = message;
        this.position = position;
        this.warning = warning;
        if (!position) {
            console.trace();
            throw new Error('position is required');
        }
    }
}
exports.PositionCapableError = PositionCapableError;
class AstNodeError extends PositionCapableError {
    constructor(message, node, warning = false) {
        super(message, AstNodeError.ensureNodePosition(node), warning);
        this.message = message;
        this.node = node;
        this.warning = warning;
    }
    static ensureNodePosition(node) {
        if (!node) {
            throw new Error('node is required');
        }
        if (!node.astNode) {
            console.dir(node);
            throw new Error('node.astNode is required');
        }
        return node.astNode;
    }
}
exports.AstNodeError = AstNodeError;
class LysScopeError extends AstNodeError {
}
exports.LysScopeError = LysScopeError;
class LysTypeError extends AstNodeError {
}
exports.LysTypeError = LysTypeError;
class LysSemanticError extends AstNodeError {
}
exports.LysSemanticError = LysSemanticError;
class LysCompilerError extends AstNodeError {
}
exports.LysCompilerError = LysCompilerError;
class TypeMismatch extends LysTypeError {
    constructor(givenType, expectedType, node) {
        super(expectedType instanceof types_1.TraitType
            ? `Type mismatch: Type ${printDebugType(givenType)} does not implement ${printDebugType(expectedType)}`
            : `Type mismatch: Type ${printDebugType(givenType)} is not assignable to ${printDebugType(expectedType)}`, node);
        this.givenType = givenType;
        this.expectedType = expectedType;
    }
}
exports.TypeMismatch = TypeMismatch;
class CannotInferReturnType extends LysTypeError {
    constructor(node) {
        super(`Cannot infer return type`, node);
    }
}
exports.CannotInferReturnType = CannotInferReturnType;
class NotAValidType extends LysTypeError {
    constructor(node, type) {
        super(type ? `${printDebugType(type)} is not a type` : `This is not a type`, node);
    }
}
exports.NotAValidType = NotAValidType;
class UnexpectedType extends LysTypeError {
    constructor(type, node) {
        super(`${printDebugType(type)} is not a value, constructor or function.`, node);
        this.type = type;
    }
}
exports.UnexpectedType = UnexpectedType;
class InvalidOverload extends LysTypeError {
    constructor(functionType, givenTypes, node) {
        super(`Could not find a valid overload for function of type ${printDebugType(functionType)} with the arguments of type (${givenTypes.join(', ')})`, node);
        this.functionType = functionType;
        this.givenTypes = givenTypes;
    }
}
exports.InvalidOverload = InvalidOverload;
class InvalidCall extends LysTypeError {
    constructor(expectedTypes, givenTypes, node) {
        super(`Invalid signature. Expecting arguments type:\n  (${expectedTypes.join(', ')})\nbut got:\n  (${givenTypes.join(', ')})`, node);
        this.expectedTypes = expectedTypes;
        this.givenTypes = givenTypes;
    }
}
exports.InvalidCall = InvalidCall;
class UnreachableCode extends LysSemanticError {
    constructor(node) {
        super(`Unreachable code`, node);
    }
}
exports.UnreachableCode = UnreachableCode;
class NotAFunction extends LysTypeError {
    constructor(type, node) {
        super(`Type mismatch: Type ${printDebugType(type)} is not a function`, node);
        this.type = type;
    }
}
exports.NotAFunction = NotAFunction;
//# sourceMappingURL=NodeError.js.map