"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Scope = void 0;
const nodes_1 = require("./nodes");
const Reference_1 = require("./Reference");
const NodeError_1 = require("./NodeError");
const astPrinter_1 = require("../utils/astPrinter");
const types_1 = require("./types");
class Scope {
    constructor(parsingContext, moduleName, parent = null, nameHint = '') {
        this.parsingContext = parsingContext;
        this.moduleName = moduleName;
        this.parent = parent;
        this.nameHint = nameHint;
        this.localScopeDeclares = new Set();
        this.exportedNames = new Set();
        this.nameMappings = {};
        this.localUsages = {};
        this.importedModules = new Map();
        this.childrenScopes = [];
        this.name = this.parsingContext.getInternalName(moduleName, nameHint + '[child_scope]');
    }
    isDescendantOf(parentScope) {
        if (this.parent) {
            if (parentScope === this.parent) {
                return true;
            }
            return this.parent.isDescendantOf(parentScope);
        }
        return false;
    }
    registerForeginModule(moduleName) {
        if (moduleName && !this.importedModules.has(moduleName)) {
            this.importedModules.set(moduleName, new Set());
        }
        return this.importedModules.get(moduleName);
    }
    registerImport(moduleName, names) {
        const mod = this.registerForeginModule(moduleName);
        if (!mod)
            throw new Error('ModuleNotFound:' + moduleName);
        names.forEach($ => mod.add($));
    }
    getInternalIdentifier(node) {
        let prefix = node.name;
        if (this.nameHint && (this.nameHint.endsWith('#') || this.nameHint.endsWith('.'))) {
            prefix = this.nameHint + prefix;
        }
        prefix = prefix.replace(/([^A-Za-z:0-9_\.<>!+*^&%$@~/=|-])/g, '_');
        return this.parsingContext.getInternalName(this.moduleName, prefix);
    }
    incrementUsage(name) {
        const reference = this.get(name);
        this.localUsages[name] = (this.localUsages[name] || 0) + 1;
        reference.usages++;
    }
    incrementUsageQName(name) {
        this.incrementUsage(name.text);
    }
    set(nameNode, type, exported) {
        const localName = nameNode.name;
        if (localName === '_')
            return null;
        if (this.localScopeDeclares.has(localName)) {
            throw new NodeError_1.LysScopeError(`"${localName}" is already declared`, nameNode);
        }
        this.nameMappings[localName] = new Reference_1.Reference(nameNode, this, type, null);
        if (exported) {
            this.exportedNames.add(localName);
        }
        this.localScopeDeclares.add(localName);
        if (!nameNode.internalIdentifier) {
            nameNode.internalIdentifier = this.getInternalIdentifier(nameNode);
        }
        return this.nameMappings[localName];
    }
    canResolveName(localName) {
        try {
            return !!this.get(localName);
        }
        catch (e) {
            return false;
        }
    }
    canResolveQName(qname) {
        return this.canResolveName(qname.text);
    }
    getQName(qname) {
        return this.get(qname.text);
    }
    canGetFromOutside(localName) {
        return this.exportedNames.has(localName);
    }
    getFromOutside(localName) {
        if (localName in this.nameMappings) {
            if (!this.canGetFromOutside(localName)) {
                throw new Error(`Name "${localName}" is private in module "${this.moduleName}"`);
            }
            return this.nameMappings[localName];
        }
        return null;
    }
    get(localName) {
        if (localName in this.nameMappings) {
            return this.nameMappings[localName];
        }
        if (localName.includes('::')) {
            const parts = localName.split('::');
            const moduleName = parts.slice(0, -1).join('::');
            const name = parts[parts.length - 1];
            const moduleDocument = this.parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.NameInitialization);
            const ref = moduleDocument.scope.getFromOutside(name);
            if (ref) {
                this.registerForeginModule(moduleName);
                return ref.withModule(moduleName);
            }
        }
        if (this.parent && this.parent.canResolveName(localName)) {
            return this.parent.get(localName);
        }
        for (let [moduleName, importsSet] of this.importedModules) {
            const moduleDocument = this.parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.NameInitialization);
            if (importsSet.has(localName)) {
                const ref = moduleDocument.scope.getFromOutside(localName);
                if (ref) {
                    return ref.withModule(moduleName);
                }
            }
            else if (importsSet.has('*') && moduleDocument.scope.canGetFromOutside(localName)) {
                const ref = moduleDocument.scope.getFromOutside(localName);
                if (ref) {
                    return ref.withModule(moduleName);
                }
            }
        }
        throw new Error("Cannot find name '" + localName + "'");
    }
    inspect(voidIfEmpty = false, includePrivate = false) {
        let ret = [];
        for (let name in this.nameMappings) {
            const ref = this.nameMappings[name];
            const type = types_1.TypeHelpers.getNodeType(ref.referencedNode);
            if (includePrivate || this.exportedNames.has(name)) {
                const priv = this.exportedNames.has(name) ? '' : 'private ';
                const typeStr = type ? ': ' + type.inspect(10) : '';
                ret.push(priv + 'let ' + name + typeStr);
            }
        }
        this.importedModules.forEach((set, $) => {
            let list = [];
            set.forEach($ => list.push($));
            ret.push('import ' + $ + '::{' + list.join(', ') + '}');
        });
        this.childrenScopes.forEach($ => {
            const childScope = $.inspect(true, includePrivate);
            if (childScope.trim().length) {
                ret.push(childScope);
            }
        });
        if (!ret.length && voidIfEmpty)
            return '';
        return `${this.name}: {\n${(0, astPrinter_1.indent)(ret.join('\n'))}\n}`;
    }
    newChildScope(nameHint) {
        const newScope = new Scope(this.parsingContext, this.moduleName, this, nameHint);
        this.childrenScopes.push(newScope);
        return newScope;
    }
}
exports.Scope = Scope;
//# sourceMappingURL=Scope.js.map