import { TokenError } from 'ebnf';
import { Scope } from './Scope';
import { Type, FunctionType } from './types';
import { Annotation, IAnnotationConstructor } from './annotations';
import { Reference } from './Reference';
export interface LocalGlobalHeapReference {
    type?: Type;
    name: string;
    declarationNode?: Nodes.Node;
}
export declare class Global implements LocalGlobalHeapReference {
    declarationNode: Nodes.NameIdentifierNode;
    type?: Type;
    name: string;
    constructor(declarationNode: Nodes.NameIdentifierNode);
}
export declare class Local implements LocalGlobalHeapReference {
    index: number;
    name: string;
    declarationNode?: Nodes.Node | undefined;
    type?: Type;
    /** index in the function */
    constructor(index: number, name: string, declarationNode?: Nodes.Node | undefined);
}
export declare enum PhaseFlags {
    Semantic = 0,
    NameInitialization = 1,
    Scope = 2,
    TypeInitialization = 3,
    TypeCheck = 4,
    PreCompilation = 5,
    Compilation = 6,
    CodeGeneration = 7
}
export declare namespace Nodes {
    interface ASTNode {
        readonly type: string;
        readonly text: string;
        readonly start: number;
        readonly end: number;
        readonly errors: TokenError[];
        readonly moduleName: string;
        readonly children: ReadonlyArray<ASTNode>;
    }
    abstract class Node {
        astNode: ASTNode;
        hasParentheses: boolean;
        isTypeResolved: boolean;
        parent?: Node;
        private ownScope;
        private annotations?;
        constructor(astNode: ASTNode);
        get scope(): Scope | undefined;
        set scope(scope: Scope | undefined);
        /** Name of the node constructor */
        get nodeName(): string;
        get children(): Node[];
        abstract get childrenOrEmpty(): (Node | void)[];
        hasAnnotation<T extends Annotation = Annotation>(name: Annotation | IAnnotationConstructor<T>): boolean;
        getAnnotationsByClass<T extends Annotation>(klass: IAnnotationConstructor<T>): T[];
        getAnnotation<T extends Annotation>(klass: IAnnotationConstructor<T>): T | null;
        removeAnnotation<T extends Annotation = Annotation>(name: Annotation | IAnnotationConstructor<T>): void;
        annotate(annotation: Annotation): void;
        getAnnotations(): Annotation[];
        getAnnotations<T extends Annotation>(klass: IAnnotationConstructor<T>): T[];
        toString(): string;
    }
    abstract class ExpressionNode extends Node {
    }
    class NameIdentifierNode extends Node {
        name: string;
        impls: Set<ImplDirective>;
        internalIdentifier?: string;
        get childrenOrEmpty(): Node[];
        constructor(astNode: ASTNode, name: string);
        static fromString(name: string, astNode: ASTNode): NameIdentifierNode;
    }
    class QNameNode extends Node {
        readonly names: NameIdentifierNode[];
        constructor(astNode: ASTNode, names: NameIdentifierNode[]);
        static fromString(name: string, astNode: ASTNode): QNameNode;
        get childrenOrEmpty(): NameIdentifierNode[];
        deconstruct(): {
            moduleName: string;
            variable: string;
        };
        get text(): string;
    }
    abstract class TypeNode extends Node {
    }
    class TypeReducerNode extends Node {
        get childrenOrEmpty(): Node[];
    }
    class SignatureParameterNode extends TypeNode {
        readonly parameterType: TypeNode;
        parameterName?: NameIdentifierNode;
        constructor(astNode: ASTNode, parameterType: TypeNode);
        get childrenOrEmpty(): (NameIdentifierNode | TypeNode | undefined)[];
    }
    class FunctionTypeNode extends TypeNode {
        readonly parameters: SignatureParameterNode[];
        typeParameters?: string[];
        effect?: TypeNode;
        returnType?: TypeNode;
        constructor(astNode: ASTNode, parameters: SignatureParameterNode[]);
        get childrenOrEmpty(): (TypeNode | undefined)[];
    }
    class EffectMemberDeclarationNode extends TypeNode {
        name?: NameIdentifierNode;
        typeParameters?: string[];
        parameters?: SignatureParameterNode[];
        returnType?: TypeNode;
        get childrenOrEmpty(): (NameIdentifierNode | undefined)[];
    }
    class ReferenceNode extends ExpressionNode {
        readonly variable: QNameNode;
        isLocal: boolean;
        resolvedReference?: Reference;
        constructor(astNode: ASTNode, variable: QNameNode);
        get childrenOrEmpty(): QNameNode[];
    }
    class BlockNode extends ExpressionNode {
        readonly statements: Node[];
        label?: string;
        constructor(astNode: ASTNode, statements: Node[]);
        get childrenOrEmpty(): Node[];
    }
    class MemberNode extends ExpressionNode {
        readonly lhs: ExpressionNode;
        readonly operator: string;
        readonly memberName: NameIdentifierNode;
        resolvedReference?: Reference;
        constructor(astNode: ASTNode, lhs: ExpressionNode, operator: string, memberName: NameIdentifierNode);
        get childrenOrEmpty(): (NameIdentifierNode | ExpressionNode)[];
    }
    class DecoratorNode extends Node {
        readonly decoratorName: NameIdentifierNode;
        readonly args: LiteralNode<any>[];
        constructor(astNode: ASTNode, decoratorName: NameIdentifierNode, args: LiteralNode<any>[]);
        get childrenOrEmpty(): (NameIdentifierNode | LiteralNode<any>)[];
    }
    abstract class DirectiveNode extends Node {
        decorators?: DecoratorNode[];
        isPublic: boolean;
    }
    class DocumentNode extends Node {
        readonly directives: DirectiveNode[];
        moduleName: string;
        fileName: string;
        content: string;
        analysis: {
            version: number;
            nextPhase: PhaseFlags;
            isTypeAnalysisPassNeeded: boolean;
            areTypesResolved: boolean;
        };
        typeNumbers: Map<string, number>;
        nameIdentifiers: Set<string>;
        importedModules: Set<string>;
        importedBy: Set<string>;
        get childrenOrEmpty(): DirectiveNode[];
    }
    class ParameterNode extends Node {
        readonly parameterName: NameIdentifierNode;
        parameterType?: TypeNode;
        defaultValue?: ExpressionNode;
        constructor(astNode: ASTNode, parameterName: NameIdentifierNode);
        get childrenOrEmpty(): (NameIdentifierNode | TypeNode | ExpressionNode | undefined)[];
    }
    class FunctionNode extends Node {
        readonly functionName: NameIdentifierNode;
        functionReturnType?: TypeNode;
        parameters: ParameterNode[];
        body?: ExpressionNode;
        get childrenOrEmpty(): (NameIdentifierNode | ParameterNode | TypeNode | ExpressionNode | undefined)[];
        /** Array of locals by index. */
        localsByIndex: Local[];
        /** List of additional non-parameter locals. */
        additionalLocals: Local[];
        private tempI32s;
        private tempI64s;
        private tempF32s;
        private tempF64s;
        constructor(astNode: ASTNode, functionName: NameIdentifierNode);
        processParameters(): void;
        /** Adds a local of the specified type. */
        addLocal(type: Type, declaration?: NameIdentifierNode): Local;
        /** Gets a free temporary local of the specified type. */
        getTempLocal(type: Type): Local;
        /** Frees the temporary local for reuse. */
        freeTempLocal(local: Local): void;
        /** Gets and immediately frees a temporary local of the specified type. */
        getAndFreeTempLocal(type: Type): Local;
    }
    class ImplDirective extends DirectiveNode {
        readonly targetImpl: ReferenceNode;
        readonly directives: DirectiveNode[];
        baseImpl?: ReferenceNode;
        selfTypeName?: NameIdentifierNode;
        readonly namespaceNames: Map<string, NameIdentifierNode>;
        constructor(astNode: ASTNode, targetImpl: ReferenceNode, directives: DirectiveNode[]);
        get childrenOrEmpty(): (DirectiveNode | ReferenceNode | DecoratorNode)[];
    }
    class ImportDirectiveNode extends DirectiveNode {
        readonly module: QNameNode;
        allItems: boolean;
        alias: NameIdentifierNode | null;
        constructor(astNode: ASTNode, module: QNameNode);
        get childrenOrEmpty(): (NameIdentifierNode | QNameNode | DecoratorNode | undefined)[];
    }
    class FunDirectiveNode extends DirectiveNode {
        readonly functionNode: FunctionNode;
        readonly decorators: DecoratorNode[];
        constructor(astNode: ASTNode, functionNode: FunctionNode, decorators: DecoratorNode[]);
        get childrenOrEmpty(): (DecoratorNode | FunctionNode)[];
    }
    class EffectDirectiveNode extends DirectiveNode {
        effect?: EffectDeclarationNode;
        get childrenOrEmpty(): (DecoratorNode | EffectDeclarationNode | undefined)[];
    }
    class OverloadedFunctionNode extends DirectiveNode {
        readonly functionName: NameIdentifierNode;
        functions: FunDirectiveNode[];
        constructor(astNode: ASTNode, functionName: NameIdentifierNode);
        get childrenOrEmpty(): (NameIdentifierNode | DecoratorNode | FunDirectiveNode)[];
    }
    class VarDeclarationNode extends Node {
        readonly variableName: NameIdentifierNode;
        value: ExpressionNode;
        variableType?: TypeNode;
        constructor(astNode: ASTNode, variableName: NameIdentifierNode, value: ExpressionNode);
        get childrenOrEmpty(): (NameIdentifierNode | TypeNode | ExpressionNode | undefined)[];
    }
    class VarDirectiveNode extends DirectiveNode {
        readonly decl: VarDeclarationNode;
        constructor(astNode: ASTNode, decl: VarDeclarationNode);
        get childrenOrEmpty(): (DecoratorNode | VarDeclarationNode)[];
    }
    class AssignmentNode extends Node {
        readonly lhs: ExpressionNode;
        readonly rhs: ExpressionNode;
        constructor(astNode: ASTNode, lhs: ExpressionNode, rhs: ExpressionNode);
        get childrenOrEmpty(): ExpressionNode[];
    }
    class TypeDirectiveNode extends DirectiveNode {
        readonly variableName: NameIdentifierNode;
        valueType?: TypeNode;
        constructor(astNode: ASTNode, variableName: NameIdentifierNode);
        get childrenOrEmpty(): (NameIdentifierNode | TypeNode | DecoratorNode | undefined)[];
    }
    class TraitDirectiveNode extends DirectiveNode {
        readonly traitName: NameIdentifierNode;
        directives: DirectiveNode[];
        readonly namespaceNames: Map<string, NameIdentifierNode>;
        selfTypeName?: NameIdentifierNode;
        constructor(astNode: ASTNode, traitName: NameIdentifierNode, directives: DirectiveNode[]);
        get childrenOrEmpty(): (NameIdentifierNode | DirectiveNode | DecoratorNode)[];
    }
    class EnumDirectiveNode extends DirectiveNode {
        readonly variableName: NameIdentifierNode;
        readonly declarations: StructDeclarationNode[];
        constructor(astNode: ASTNode, variableName: NameIdentifierNode, declarations: StructDeclarationNode[]);
        get childrenOrEmpty(): (NameIdentifierNode | DecoratorNode | StructDeclarationNode)[];
    }
    abstract class LiteralNode<T> extends ExpressionNode {
        typeName: string;
        get value(): T;
        set value(_: T);
        rawValue: string;
        resolvedReference?: Reference;
        constructor(astNode: ASTNode, typeName: string);
    }
    class FloatLiteral extends LiteralNode<number> {
        suffixReference?: ReferenceNode;
        get value(): number;
        set value(value: number);
        get childrenOrEmpty(): (ReferenceNode | undefined)[];
    }
    class IntegerLiteral extends LiteralNode<number> {
        suffixReference?: ReferenceNode;
        get value(): number;
        set value(value: number);
        get childrenOrEmpty(): (ReferenceNode | undefined)[];
    }
    class UnknownExpressionNode extends ExpressionNode {
        get childrenOrEmpty(): never[];
    }
    class StructTypeNode extends TypeNode {
        readonly parameters: ParameterNode[];
        constructor(astNode: ASTNode, parameters: ParameterNode[]);
        get childrenOrEmpty(): ParameterNode[];
    }
    class StackTypeNode extends TypeNode {
        metadata: Record<string, LiteralNode<any>>;
        get childrenOrEmpty(): never[];
    }
    class InjectedTypeNode extends TypeNode {
        get childrenOrEmpty(): never[];
    }
    class HexLiteral extends IntegerLiteral {
        suffixReference?: ReferenceNode;
        get value(): number;
        set value(value: number);
        get childrenOrEmpty(): (ReferenceNode | undefined)[];
    }
    class BooleanLiteral extends LiteralNode<boolean> {
        get value(): boolean;
        set value(value: boolean);
        get childrenOrEmpty(): never[];
    }
    class StringLiteral extends LiteralNode<string> {
        get value(): string;
        set value(value: string);
        offset?: number;
        length?: number;
        get childrenOrEmpty(): never[];
    }
    abstract class AbstractFunctionCallNode extends ExpressionNode {
        abstract argumentsNode: ExpressionNode[];
        resolvedFunctionType?: FunctionType;
    }
    class InjectedFunctionCallNode extends AbstractFunctionCallNode {
        argumentsNode: ExpressionNode[];
        get childrenOrEmpty(): ExpressionNode[];
    }
    class FunctionCallNode extends AbstractFunctionCallNode {
        functionNode: ExpressionNode;
        readonly argumentsNode: ExpressionNode[];
        isInfix: boolean;
        constructor(astNode: ASTNode, functionNode: ExpressionNode, argumentsNode: ExpressionNode[]);
        get childrenOrEmpty(): ExpressionNode[];
    }
    class BinaryExpressionNode extends ExpressionNode {
        readonly operator: NameIdentifierNode;
        lhs: ExpressionNode;
        rhs: ExpressionNode;
        constructor(astNode: ASTNode, operator: NameIdentifierNode, lhs: ExpressionNode, rhs: ExpressionNode);
        get childrenOrEmpty(): (NameIdentifierNode | ExpressionNode)[];
    }
    class AsExpressionNode extends AbstractFunctionCallNode {
        get lhs(): ExpressionNode;
        set lhs(value: ExpressionNode);
        argumentsNode: ExpressionNode[];
        rhs: TypeNode;
        constructor(astNode: ASTNode, lhs: ExpressionNode, rhs: TypeNode);
        get childrenOrEmpty(): (TypeNode | ExpressionNode)[];
    }
    class IsExpressionNode extends AbstractFunctionCallNode {
        get lhs(): ExpressionNode;
        set lhs(value: ExpressionNode);
        booleanReference?: Reference;
        argumentsNode: ExpressionNode[];
        rhs: TypeNode;
        constructor(astNode: ASTNode, lhs: ExpressionNode, rhs: TypeNode);
        get childrenOrEmpty(): (TypeNode | ExpressionNode)[];
    }
    class UnaryExpressionNode extends AbstractFunctionCallNode {
        readonly operator: NameIdentifierNode;
        get rhs(): TypeNode;
        set rhs(value: TypeNode);
        argumentsNode: Node[];
        constructor(astNode: ASTNode, operator: NameIdentifierNode, rhs: TypeNode);
        get childrenOrEmpty(): TypeNode[];
    }
    class WasmAtomNode extends Node {
        readonly symbol: string;
        readonly args: ExpressionNode[];
        constructor(astNode: ASTNode, symbol: string, args: ExpressionNode[]);
        get childrenOrEmpty(): ExpressionNode[];
    }
    class WasmExpressionNode extends ExpressionNode {
        readonly atoms: WasmAtomNode[];
        constructor(astNode: ASTNode, atoms: WasmAtomNode[]);
        get childrenOrEmpty(): WasmAtomNode[];
    }
    abstract class MatcherNode extends ExpressionNode {
        readonly rhs: ExpressionNode;
        declaredName?: NameIdentifierNode;
        parent?: PatternMatcherNode;
        booleanReference?: Reference;
        constructor(astNode: ASTNode, rhs: ExpressionNode);
        get childrenOrEmpty(): (NameIdentifierNode | ExpressionNode | undefined)[];
    }
    class IfNode extends ExpressionNode {
        readonly condition: ExpressionNode;
        readonly truePart: ExpressionNode;
        readonly falsePart?: ExpressionNode | undefined;
        booleanReference?: Reference;
        constructor(astNode: ASTNode, condition: ExpressionNode, truePart: ExpressionNode, falsePart?: ExpressionNode | undefined);
        get childrenOrEmpty(): (ExpressionNode | undefined)[];
    }
    class MatchConditionNode extends MatcherNode {
        readonly condition: ExpressionNode;
        readonly rhs: ExpressionNode;
        constructor(astNode: ASTNode, condition: ExpressionNode, rhs: ExpressionNode);
        get childrenOrEmpty(): (NameIdentifierNode | ExpressionNode | undefined)[];
    }
    class MatchCaseIsNode extends MatcherNode {
        readonly typeReference: ReferenceNode;
        rhs: ExpressionNode;
        deconstructorNames?: NameIdentifierNode[];
        resolvedFunctionType?: FunctionType;
        constructor(astNode: ASTNode, typeReference: ReferenceNode, rhs: ExpressionNode);
        get childrenOrEmpty(): (NameIdentifierNode | ExpressionNode | undefined)[];
    }
    class MatchLiteralNode extends MatcherNode {
        readonly literal: LiteralNode<any>;
        readonly rhs: ExpressionNode;
        resolvedFunctionType?: FunctionType;
        constructor(astNode: ASTNode, literal: LiteralNode<any>, rhs: ExpressionNode);
        get childrenOrEmpty(): (NameIdentifierNode | ExpressionNode | undefined)[];
    }
    class UnionTypeNode extends TypeNode {
        of: TypeNode[];
        get childrenOrEmpty(): TypeNode[];
    }
    class IntersectionTypeNode extends TypeNode {
        of: TypeNode[];
        get childrenOrEmpty(): TypeNode[];
    }
    class StructDeclarationNode extends TypeNode {
        readonly declaredName: NameIdentifierNode;
        readonly parameters: ParameterNode[];
        isPublic: boolean;
        constructor(astNode: ASTNode, declaredName: NameIdentifierNode, parameters: ParameterNode[]);
        get childrenOrEmpty(): (NameIdentifierNode | ParameterNode)[];
    }
    class EffectDeclarationNode extends Node {
        readonly name: NameIdentifierNode;
        elements?: FunctionTypeNode[];
        constructor(astNode: ASTNode, name: NameIdentifierNode);
        get childrenOrEmpty(): (NameIdentifierNode | FunctionTypeNode)[];
    }
    class MatchDefaultNode extends MatcherNode {
    }
    class PatternMatcherNode extends ExpressionNode {
        readonly lhs: ExpressionNode;
        readonly matchingSet: MatcherNode[];
        carryType?: Type;
        constructor(astNode: ASTNode, lhs: ExpressionNode, matchingSet: MatcherNode[]);
        get childrenOrEmpty(): Node[];
    }
    class LoopNode extends ExpressionNode {
        readonly body: ExpressionNode;
        constructor(astNode: ASTNode, body: ExpressionNode);
        get childrenOrEmpty(): ExpressionNode[];
    }
    class ContinueNode extends Node {
        get childrenOrEmpty(): never[];
    }
    class BreakNode extends Node {
        get childrenOrEmpty(): never[];
    }
}
export declare function findNodesByType<T>(astRoot: {
    children: any[];
}, type: {
    new (...args: any[]): T;
}, list?: T[]): T[];
export declare function findNodesByTypeInChildren<T>(astRoot: {
    children: any[];
}, type: {
    new (...args: any[]): T;
}, list?: T[]): T[];
