"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.findNodesByTypeInChildren = exports.findNodesByType = exports.Nodes = exports.PhaseFlags = exports.Local = exports.Global = void 0;
const types_1 = require("./types");
const annotations_1 = require("./annotations");
class Global {
    constructor(declarationNode) {
        this.declarationNode = declarationNode;
        this.name = declarationNode.internalIdentifier;
    }
}
exports.Global = Global;
class Local {
    /** index in the function */
    constructor(index, name, declarationNode) {
        this.index = index;
        this.name = name;
        this.declarationNode = declarationNode;
    }
}
exports.Local = Local;
var PhaseFlags;
(function (PhaseFlags) {
    PhaseFlags[PhaseFlags["Semantic"] = 0] = "Semantic";
    PhaseFlags[PhaseFlags["NameInitialization"] = 1] = "NameInitialization";
    PhaseFlags[PhaseFlags["Scope"] = 2] = "Scope";
    PhaseFlags[PhaseFlags["TypeInitialization"] = 3] = "TypeInitialization";
    PhaseFlags[PhaseFlags["TypeCheck"] = 4] = "TypeCheck";
    PhaseFlags[PhaseFlags["PreCompilation"] = 5] = "PreCompilation";
    PhaseFlags[PhaseFlags["Compilation"] = 6] = "Compilation";
    PhaseFlags[PhaseFlags["CodeGeneration"] = 7] = "CodeGeneration";
})(PhaseFlags = exports.PhaseFlags || (exports.PhaseFlags = {}));
var Nodes;
(function (Nodes) {
    class Node {
        constructor(astNode) {
            this.astNode = astNode;
            // TODO: remove this VVVVVVVVVV
            this.hasParentheses = false;
            this.isTypeResolved = false;
        }
        get scope() {
            if (this.ownScope)
                return this.ownScope;
            if (this.parent)
                return this.parent.scope;
            return undefined;
        }
        set scope(scope) {
            this.ownScope = scope;
        }
        /** Name of the node constructor */
        get nodeName() {
            return this.constructor.name;
        }
        get children() {
            return this.childrenOrEmpty.filter($ => !!$);
        }
        hasAnnotation(name) {
            if (!this.annotations)
                return false;
            if (typeof name === 'object') {
                return this.annotations.has(name);
            }
            else {
                return !!this.getAnnotation(name);
            }
        }
        getAnnotationsByClass(klass) {
            const ret = [];
            if (this.annotations) {
                this.annotations.forEach($ => {
                    if ($ instanceof klass)
                        ret.push($);
                });
            }
            return ret;
        }
        getAnnotation(klass) {
            const ret = [];
            if (this.annotations) {
                this.annotations.forEach($ => {
                    if ($ instanceof klass)
                        ret.push($);
                });
            }
            return ret[ret.length - 1] || null;
        }
        removeAnnotation(name) {
            if (typeof name === 'object') {
                this.annotations && this.annotations.delete(name);
            }
            else {
                this.annotations &&
                    this.getAnnotationsByClass(name).forEach($ => this.annotations && this.annotations.delete($));
            }
        }
        annotate(annotation) {
            if (!this.annotations)
                this.annotations = new Set();
            this.annotations.add(annotation);
        }
        getAnnotations(klass) {
            const ret = [];
            if (this.annotations) {
                this.annotations.forEach($ => {
                    if (!klass || $ instanceof klass)
                        ret.push($);
                });
            }
            return ret;
        }
        toString() {
            console.trace();
            throw new Error('Cannot print nodes with toString');
        }
    }
    Nodes.Node = Node;
    class ExpressionNode extends Node {
    }
    Nodes.ExpressionNode = ExpressionNode;
    class NameIdentifierNode extends Node {
        constructor(astNode, name) {
            super(astNode);
            this.name = name;
            this.impls = new Set();
        }
        get childrenOrEmpty() {
            return [];
        }
        static fromString(name, astNode) {
            const r = new NameIdentifierNode(astNode, name);
            return r;
        }
    }
    Nodes.NameIdentifierNode = NameIdentifierNode;
    class QNameNode extends Node {
        constructor(astNode, names) {
            super(astNode);
            this.names = names;
        }
        static fromString(name, astNode) {
            const names = name.split('::').map($ => NameIdentifierNode.fromString($, astNode));
            const r = new QNameNode(astNode, names);
            return r;
        }
        get childrenOrEmpty() {
            return this.names || [];
        }
        deconstruct() {
            const moduleName = this.names
                .slice(0, -1)
                .map($ => $.name)
                .join('::');
            const variable = this.names[this.names.length - 1].name;
            return { moduleName, variable };
        }
        get text() {
            return this.names.map($ => $.name).join('::');
        }
    }
    Nodes.QNameNode = QNameNode;
    class TypeNode extends Node {
    }
    Nodes.TypeNode = TypeNode;
    class TypeReducerNode extends Node {
        get childrenOrEmpty() {
            return [];
        }
    }
    Nodes.TypeReducerNode = TypeReducerNode;
    class SignatureParameterNode extends TypeNode {
        constructor(astNode, parameterType) {
            super(astNode);
            this.parameterType = parameterType;
        }
        get childrenOrEmpty() {
            return [this.parameterName, this.parameterType];
        }
    }
    Nodes.SignatureParameterNode = SignatureParameterNode;
    class FunctionTypeNode extends TypeNode {
        constructor(astNode, parameters) {
            super(astNode);
            this.parameters = parameters;
        }
        get childrenOrEmpty() {
            return [...(this.parameters || []), this.returnType, this.effect];
        }
    }
    Nodes.FunctionTypeNode = FunctionTypeNode;
    class EffectMemberDeclarationNode extends TypeNode {
        get childrenOrEmpty() {
            return [this.name];
        }
    }
    Nodes.EffectMemberDeclarationNode = EffectMemberDeclarationNode;
    class ReferenceNode extends ExpressionNode {
        constructor(astNode, variable) {
            super(astNode);
            this.variable = variable;
            this.isLocal = false;
        }
        get childrenOrEmpty() {
            return [this.variable];
        }
    }
    Nodes.ReferenceNode = ReferenceNode;
    class BlockNode extends ExpressionNode {
        constructor(astNode, statements) {
            super(astNode);
            this.statements = statements;
        }
        get childrenOrEmpty() {
            return this.statements || [];
        }
    }
    Nodes.BlockNode = BlockNode;
    class MemberNode extends ExpressionNode {
        constructor(astNode, lhs, operator, memberName) {
            super(astNode);
            this.lhs = lhs;
            this.operator = operator;
            this.memberName = memberName;
        }
        get childrenOrEmpty() {
            return [this.lhs, this.memberName];
        }
    }
    Nodes.MemberNode = MemberNode;
    class DecoratorNode extends Node {
        constructor(astNode, decoratorName, args) {
            super(astNode);
            this.decoratorName = decoratorName;
            this.args = args;
        }
        get childrenOrEmpty() {
            return [this.decoratorName, ...(this.args || [])];
        }
    }
    Nodes.DecoratorNode = DecoratorNode;
    class DirectiveNode extends Node {
        constructor() {
            super(...arguments);
            this.isPublic = false;
        }
    }
    Nodes.DirectiveNode = DirectiveNode;
    class DocumentNode extends Node {
        constructor() {
            super(...arguments);
            this.directives = [];
            this.analysis = {
                version: 0,
                nextPhase: PhaseFlags.Semantic,
                isTypeAnalysisPassNeeded: true,
                areTypesResolved: false
            };
            this.typeNumbers = new Map();
            this.nameIdentifiers = new Set();
            this.importedModules = new Set();
            this.importedBy = new Set();
        }
        get childrenOrEmpty() {
            return this.directives;
        }
    }
    Nodes.DocumentNode = DocumentNode;
    class ParameterNode extends Node {
        constructor(astNode, parameterName) {
            super(astNode);
            this.parameterName = parameterName;
        }
        get childrenOrEmpty() {
            return [this.parameterName, this.parameterType, this.defaultValue];
        }
    }
    Nodes.ParameterNode = ParameterNode;
    class FunctionNode extends Node {
        constructor(astNode, functionName) {
            super(astNode);
            this.functionName = functionName;
            this.parameters = [];
            /** Array of locals by index. */
            this.localsByIndex = [];
            /** List of additional non-parameter locals. */
            this.additionalLocals = [];
            this.tempI32s = null;
            this.tempI64s = null;
            this.tempF32s = null;
            this.tempF64s = null;
        }
        get childrenOrEmpty() {
            return [this.functionName, ...this.parameters, this.functionReturnType, this.body];
        }
        processParameters() {
            let localIndex = 0;
            this.parameters.forEach(parameter => {
                let local = new Local(localIndex++, parameter.parameterName.name, parameter.parameterName);
                this.localsByIndex[local.index] = local;
                parameter.annotate(new annotations_1.annotations.LocalIdentifier(local));
            });
        }
        /** Adds a local of the specified type. */
        addLocal(type, declaration) {
            // if it has a name, check previously as this method will throw otherwise
            let localIndex = this.parameters.length + this.additionalLocals.length;
            let local = new Local(localIndex, declaration ? declaration.internalIdentifier : 'var$' + localIndex.toString(10), declaration);
            local.type = type;
            this.localsByIndex[local.index] = local;
            this.additionalLocals.push(local);
            return local;
        }
        /** Gets a free temporary local of the specified type. */
        getTempLocal(type) {
            let temps;
            switch (type.binaryenType) {
                case types_1.NativeTypes.i32: {
                    temps = this.tempI32s;
                    break;
                }
                case types_1.NativeTypes.i64: {
                    temps = this.tempI64s;
                    break;
                }
                case types_1.NativeTypes.f32: {
                    temps = this.tempF32s;
                    break;
                }
                case types_1.NativeTypes.f64: {
                    temps = this.tempF64s;
                    break;
                }
                default:
                    console.trace('concrete type expected ' + type.toString());
                    throw new Error('concrete type expected');
            }
            let local;
            if (temps && temps.length) {
                local = temps.pop();
                local.type = type;
            }
            else {
                local = this.addLocal(type);
            }
            return local;
        }
        /** Frees the temporary local for reuse. */
        freeTempLocal(local) {
            let temps;
            if (!local.type)
                throw new Error('type is null'); // internal error
            switch (local.type.binaryenType) {
                case types_1.NativeTypes.i32: {
                    temps = this.tempI32s || (this.tempI32s = []);
                    break;
                }
                case types_1.NativeTypes.i64: {
                    temps = this.tempI64s || (this.tempI64s = []);
                    break;
                }
                case types_1.NativeTypes.f32: {
                    temps = this.tempF32s || (this.tempF32s = []);
                    break;
                }
                case types_1.NativeTypes.f64: {
                    temps = this.tempF64s || (this.tempF64s = []);
                    break;
                }
                default:
                    console.trace('concrete type expected');
                    throw new Error('concrete type expected');
            }
            if (local.index < 0)
                throw new Error('invalid local index');
            temps.push(local);
        }
        /** Gets and immediately frees a temporary local of the specified type. */
        getAndFreeTempLocal(type) {
            let temps;
            switch (type.binaryenType) {
                case types_1.NativeTypes.i32: {
                    temps = this.tempI32s || (this.tempI32s = []);
                    break;
                }
                case types_1.NativeTypes.i64: {
                    temps = this.tempI64s || (this.tempI64s = []);
                    break;
                }
                case types_1.NativeTypes.f32: {
                    temps = this.tempF32s || (this.tempF32s = []);
                    break;
                }
                case types_1.NativeTypes.f64: {
                    temps = this.tempF64s || (this.tempF64s = []);
                    break;
                }
                default:
                    console.trace('concrete type expected');
                    throw new Error('concrete type expected');
            }
            let local;
            if (temps.length) {
                local = temps[temps.length - 1];
                local.type = type;
            }
            else {
                local = this.addLocal(type);
                temps.push(local);
            }
            return local;
        }
    }
    Nodes.FunctionNode = FunctionNode;
    class ImplDirective extends DirectiveNode {
        constructor(astNode, targetImpl, directives) {
            super(astNode);
            this.targetImpl = targetImpl;
            this.directives = directives;
            this.namespaceNames = new Map();
        }
        get childrenOrEmpty() {
            return [
                ...(this.decorators || []),
                this.targetImpl,
                ...(this.baseImpl ? [this.baseImpl] : []),
                ...(this.directives || [])
            ];
        }
    }
    Nodes.ImplDirective = ImplDirective;
    class ImportDirectiveNode extends DirectiveNode {
        constructor(astNode, module) {
            super(astNode);
            this.module = module;
            this.allItems = true;
            this.alias = null;
        }
        get childrenOrEmpty() {
            return [...(this.decorators || []), this.module, this.alias || void 0];
        }
    }
    Nodes.ImportDirectiveNode = ImportDirectiveNode;
    class FunDirectiveNode extends DirectiveNode {
        constructor(astNode, functionNode, decorators) {
            super(astNode);
            this.functionNode = functionNode;
            this.decorators = decorators;
        }
        get childrenOrEmpty() {
            return [...(this.decorators || []), this.functionNode];
        }
    }
    Nodes.FunDirectiveNode = FunDirectiveNode;
    class EffectDirectiveNode extends DirectiveNode {
        get childrenOrEmpty() {
            return [...(this.decorators || []), this.effect];
        }
    }
    Nodes.EffectDirectiveNode = EffectDirectiveNode;
    class OverloadedFunctionNode extends DirectiveNode {
        constructor(astNode, functionName) {
            super(astNode);
            this.functionName = functionName;
            this.functions = [];
        }
        get childrenOrEmpty() {
            return [...(this.decorators || []), this.functionName, ...this.functions];
        }
    }
    Nodes.OverloadedFunctionNode = OverloadedFunctionNode;
    class VarDeclarationNode extends Node {
        constructor(astNode, variableName, value) {
            super(astNode);
            this.variableName = variableName;
            this.value = value;
        }
        get childrenOrEmpty() {
            return [this.variableName, this.variableType, this.value];
        }
    }
    Nodes.VarDeclarationNode = VarDeclarationNode;
    class VarDirectiveNode extends DirectiveNode {
        constructor(astNode, decl) {
            super(astNode);
            this.decl = decl;
        }
        get childrenOrEmpty() {
            return [...(this.decorators || []), this.decl];
        }
    }
    Nodes.VarDirectiveNode = VarDirectiveNode;
    class AssignmentNode extends Node {
        constructor(astNode, lhs, rhs) {
            super(astNode);
            this.lhs = lhs;
            this.rhs = rhs;
        }
        get childrenOrEmpty() {
            return [this.lhs, this.rhs];
        }
    }
    Nodes.AssignmentNode = AssignmentNode;
    class TypeDirectiveNode extends DirectiveNode {
        constructor(astNode, variableName) {
            super(astNode);
            this.variableName = variableName;
        }
        get childrenOrEmpty() {
            return [...(this.decorators || []), this.variableName, this.valueType];
        }
    }
    Nodes.TypeDirectiveNode = TypeDirectiveNode;
    class TraitDirectiveNode extends DirectiveNode {
        constructor(astNode, traitName, directives) {
            super(astNode);
            this.traitName = traitName;
            this.directives = directives;
            this.namespaceNames = new Map();
        }
        get childrenOrEmpty() {
            return [...(this.decorators || []), this.traitName, ...this.directives];
        }
    }
    Nodes.TraitDirectiveNode = TraitDirectiveNode;
    class EnumDirectiveNode extends DirectiveNode {
        constructor(astNode, variableName, declarations) {
            super(astNode);
            this.variableName = variableName;
            this.declarations = declarations;
        }
        get childrenOrEmpty() {
            return [...(this.decorators || []), this.variableName, ...(this.declarations || [])];
        }
    }
    Nodes.EnumDirectiveNode = EnumDirectiveNode;
    class LiteralNode extends ExpressionNode {
        constructor(astNode, typeName) {
            super(astNode);
            this.typeName = typeName;
            this.rawValue = '';
            if (astNode) {
                this.rawValue = astNode.text;
            }
        }
        get value() { throw new Error('not implemented'); }
        set value(_) { throw new Error('not implemented'); }
    }
    Nodes.LiteralNode = LiteralNode;
    class FloatLiteral extends LiteralNode {
        get value() {
            return parseFloat(this.rawValue);
        }
        set value(value) {
            this.rawValue = value.toString();
        }
        get childrenOrEmpty() {
            return [this.suffixReference];
        }
    }
    Nodes.FloatLiteral = FloatLiteral;
    class IntegerLiteral extends LiteralNode {
        get value() {
            return parseInt(this.rawValue, 10);
        }
        set value(value) {
            this.rawValue = value.toString();
        }
        get childrenOrEmpty() {
            return [this.suffixReference];
        }
    }
    Nodes.IntegerLiteral = IntegerLiteral;
    class UnknownExpressionNode extends ExpressionNode {
        get childrenOrEmpty() {
            return [];
        }
    }
    Nodes.UnknownExpressionNode = UnknownExpressionNode;
    class StructTypeNode extends TypeNode {
        constructor(astNode, parameters) {
            super(astNode);
            this.parameters = parameters;
        }
        get childrenOrEmpty() {
            return this.parameters;
        }
    }
    Nodes.StructTypeNode = StructTypeNode;
    class StackTypeNode extends TypeNode {
        constructor() {
            super(...arguments);
            this.metadata = {};
        }
        get childrenOrEmpty() {
            return [];
        }
    }
    Nodes.StackTypeNode = StackTypeNode;
    class InjectedTypeNode extends TypeNode {
        get childrenOrEmpty() {
            return [];
        }
    }
    Nodes.InjectedTypeNode = InjectedTypeNode;
    class HexLiteral extends IntegerLiteral {
        get value() {
            return parseInt(this.rawValue, 16);
        }
        set value(value) {
            this.rawValue = value.toString(16);
        }
        get childrenOrEmpty() {
            return [this.suffixReference];
        }
    }
    Nodes.HexLiteral = HexLiteral;
    class BooleanLiteral extends LiteralNode {
        get value() {
            return this.rawValue.trim() === 'true';
        }
        set value(value) {
            this.rawValue = value.toString();
        }
        get childrenOrEmpty() {
            return [];
        }
    }
    Nodes.BooleanLiteral = BooleanLiteral;
    class StringLiteral extends LiteralNode {
        get value() {
            return this.rawValue;
        }
        set value(value) {
            this.rawValue = value;
        }
        get childrenOrEmpty() {
            return [];
        }
    }
    Nodes.StringLiteral = StringLiteral;
    class AbstractFunctionCallNode extends ExpressionNode {
    }
    Nodes.AbstractFunctionCallNode = AbstractFunctionCallNode;
    class InjectedFunctionCallNode extends AbstractFunctionCallNode {
        constructor() {
            super(...arguments);
            this.argumentsNode = [];
        }
        get childrenOrEmpty() {
            return this.argumentsNode;
        }
    }
    Nodes.InjectedFunctionCallNode = InjectedFunctionCallNode;
    class FunctionCallNode extends AbstractFunctionCallNode {
        constructor(astNode, functionNode, argumentsNode) {
            super(astNode);
            this.functionNode = functionNode;
            this.argumentsNode = argumentsNode;
            this.isInfix = false;
        }
        get childrenOrEmpty() {
            return [this.functionNode, ...this.argumentsNode];
        }
    }
    Nodes.FunctionCallNode = FunctionCallNode;
    class BinaryExpressionNode extends ExpressionNode {
        constructor(astNode, operator, lhs, rhs) {
            super(astNode);
            this.operator = operator;
            this.lhs = lhs;
            this.rhs = rhs;
        }
        get childrenOrEmpty() {
            return [this.lhs, this.operator, this.rhs];
        }
    }
    Nodes.BinaryExpressionNode = BinaryExpressionNode;
    class AsExpressionNode extends AbstractFunctionCallNode {
        constructor(astNode, lhs, rhs) {
            super(astNode);
            this.argumentsNode = [];
            this.lhs = lhs;
            this.rhs = rhs;
        }
        get lhs() {
            return this.argumentsNode[0];
        }
        set lhs(value) {
            this.argumentsNode[0] = value;
        }
        get childrenOrEmpty() {
            return [this.lhs, this.rhs];
        }
    }
    Nodes.AsExpressionNode = AsExpressionNode;
    class IsExpressionNode extends AbstractFunctionCallNode {
        constructor(astNode, lhs, rhs) {
            super(astNode);
            this.argumentsNode = [];
            this.lhs = lhs;
            this.rhs = rhs;
        }
        get lhs() {
            return this.argumentsNode[0];
        }
        set lhs(value) {
            this.argumentsNode[0] = value;
        }
        get childrenOrEmpty() {
            return [this.lhs, this.rhs];
        }
    }
    Nodes.IsExpressionNode = IsExpressionNode;
    class UnaryExpressionNode extends AbstractFunctionCallNode {
        constructor(astNode, operator, rhs) {
            super(astNode);
            this.operator = operator;
            this.argumentsNode = [];
            this.rhs = rhs;
        }
        get rhs() {
            return this.argumentsNode[0];
        }
        set rhs(value) {
            this.argumentsNode[0] = value;
        }
        get childrenOrEmpty() {
            return [this.rhs];
        }
    }
    Nodes.UnaryExpressionNode = UnaryExpressionNode;
    class WasmAtomNode extends Node {
        constructor(astNode, symbol, args) {
            super(astNode);
            this.symbol = symbol;
            this.args = args;
        }
        get childrenOrEmpty() {
            return [...this.args];
        }
    }
    Nodes.WasmAtomNode = WasmAtomNode;
    class WasmExpressionNode extends ExpressionNode {
        constructor(astNode, atoms) {
            super(astNode);
            this.atoms = atoms;
        }
        get childrenOrEmpty() {
            return this.atoms || [];
        }
    }
    Nodes.WasmExpressionNode = WasmExpressionNode;
    class MatcherNode extends ExpressionNode {
        constructor(astNode, rhs) {
            super(astNode);
            this.rhs = rhs;
        }
        get childrenOrEmpty() {
            return [this.declaredName, this.rhs];
        }
    }
    Nodes.MatcherNode = MatcherNode;
    class IfNode extends ExpressionNode {
        constructor(astNode, condition, truePart, falsePart) {
            super(astNode);
            this.condition = condition;
            this.truePart = truePart;
            this.falsePart = falsePart;
        }
        get childrenOrEmpty() {
            return [this.condition, this.truePart, this.falsePart];
        }
    }
    Nodes.IfNode = IfNode;
    class MatchConditionNode extends MatcherNode {
        constructor(astNode, condition, rhs) {
            super(astNode, rhs);
            this.condition = condition;
            this.rhs = rhs;
        }
        get childrenOrEmpty() {
            return [...super.childrenOrEmpty, this.condition];
        }
    }
    Nodes.MatchConditionNode = MatchConditionNode;
    class MatchCaseIsNode extends MatcherNode {
        constructor(astNode, typeReference, rhs) {
            super(astNode, rhs);
            this.typeReference = typeReference;
            this.rhs = rhs;
        }
        get childrenOrEmpty() {
            return [...super.childrenOrEmpty, this.typeReference, ...(this.deconstructorNames || [])];
        }
    }
    Nodes.MatchCaseIsNode = MatchCaseIsNode;
    class MatchLiteralNode extends MatcherNode {
        constructor(astNode, literal, rhs) {
            super(astNode, rhs);
            this.literal = literal;
            this.rhs = rhs;
        }
        get childrenOrEmpty() {
            return [...super.childrenOrEmpty, this.literal];
        }
    }
    Nodes.MatchLiteralNode = MatchLiteralNode;
    class UnionTypeNode extends TypeNode {
        constructor() {
            super(...arguments);
            this.of = [];
        }
        get childrenOrEmpty() {
            return this.of;
        }
    }
    Nodes.UnionTypeNode = UnionTypeNode;
    class IntersectionTypeNode extends TypeNode {
        constructor() {
            super(...arguments);
            this.of = [];
        }
        get childrenOrEmpty() {
            return this.of || [];
        }
    }
    Nodes.IntersectionTypeNode = IntersectionTypeNode;
    class StructDeclarationNode extends TypeNode {
        constructor(astNode, declaredName, parameters) {
            super(astNode);
            this.declaredName = declaredName;
            this.parameters = parameters;
            this.isPublic = true;
        }
        get childrenOrEmpty() {
            return [this.declaredName, ...(this.parameters || [])];
        }
    }
    Nodes.StructDeclarationNode = StructDeclarationNode;
    class EffectDeclarationNode extends Node {
        constructor(astNode, name) {
            super(astNode);
            this.name = name;
        }
        get childrenOrEmpty() {
            return [this.name, ...(this.elements || [])];
        }
    }
    Nodes.EffectDeclarationNode = EffectDeclarationNode;
    class MatchDefaultNode extends MatcherNode {
    }
    Nodes.MatchDefaultNode = MatchDefaultNode;
    class PatternMatcherNode extends ExpressionNode {
        constructor(astNode, lhs, matchingSet) {
            super(astNode);
            this.lhs = lhs;
            this.matchingSet = matchingSet;
        }
        get childrenOrEmpty() {
            return [this.lhs, ...(this.matchingSet || [])];
        }
    }
    Nodes.PatternMatcherNode = PatternMatcherNode;
    class LoopNode extends ExpressionNode {
        constructor(astNode, body) {
            super(astNode);
            this.body = body;
        }
        get childrenOrEmpty() {
            return [this.body];
        }
    }
    Nodes.LoopNode = LoopNode;
    class ContinueNode extends Node {
        get childrenOrEmpty() {
            return [];
        }
    }
    Nodes.ContinueNode = ContinueNode;
    class BreakNode extends Node {
        get childrenOrEmpty() {
            return [];
        }
    }
    Nodes.BreakNode = BreakNode;
})(Nodes = exports.Nodes || (exports.Nodes = {}));
function findNodesByType(astRoot, type, list = []) {
    if (astRoot instanceof type) {
        list.push(astRoot);
    }
    astRoot.children.forEach($ => findNodesByType($, type, list));
    return list;
}
exports.findNodesByType = findNodesByType;
function findNodesByTypeInChildren(astRoot, type, list = []) {
    astRoot.children.forEach($ => {
        if ($ instanceof type) {
            list.push($);
        }
    });
    return list;
}
exports.findNodesByTypeInChildren = findNodesByTypeInChildren;
//# sourceMappingURL=nodes.js.map