"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getAST = void 0;
const nodes_1 = require("../nodes");
const NodeError_1 = require("../NodeError");
const walker_1 = require("../walker");
const grammar_1 = require("../../grammar");
const annotations_1 = require("../annotations");
/// --- PARSING PHASE ---
const process = (0, walker_1.walkPreOrder)((token, parsingContext) => {
    if (token.errors && token.errors.length) {
        token.errors.forEach(($) => {
            if ($) {
                parsingContext.messageCollector.error(new NodeError_1.PositionCapableError($.message, token));
            }
        });
    }
});
const setModuleName = (moduleName) => (0, walker_1.walkPreOrder)((token) => {
    token.moduleName = moduleName;
});
const parsingCache = new Map();
function DJB2(input) {
    let hash = 5381;
    for (let i = 0; i < input.length; i++) {
        hash = (hash << 5) + hash + input.charCodeAt(i);
    }
    return hash;
}
function getParsingTree(moduleName, content, parsingContext) {
    const hash = moduleName + '+' + content.length.toString(16) + '_' + DJB2(content).toString(16);
    let ret = parsingCache.get(hash);
    if (!ret) {
        ret = grammar_1.parser.getAST(content, 'Document');
        parsingCache.set(hash, ret);
        setModuleName(moduleName)(ret, parsingContext);
    }
    return ret;
}
/// --- CANONICAL ---
function binaryOpVisitor(astNode) {
    let ret = visit(astNode.children[0]);
    for (let i = 1; i < astNode.children.length; i += 2) {
        const oldRet = ret;
        const opertator = astNode.children[i].text;
        const lhs = oldRet;
        const rhs = visit(astNode.children[i + 1]);
        if (opertator === 'as') {
            ret = new nodes_1.Nodes.AsExpressionNode(astNode, lhs, rhs);
        }
        else if (opertator === 'is') {
            ret = new nodes_1.Nodes.IsExpressionNode(astNode, lhs, rhs);
        }
        else if (opertator === '=') {
            ret = new nodes_1.Nodes.AssignmentNode(astNode, lhs, rhs);
        }
        else {
            const operator = new nodes_1.Nodes.NameIdentifierNode(astNode.children[i], opertator);
            ret = new nodes_1.Nodes.BinaryExpressionNode(astNode, operator, lhs, rhs);
        }
    }
    return ret;
}
const visitor = {
    ImportDirective(astNode) {
        const module = visitChildTypeOrNull(astNode, 'QName');
        const ret = new nodes_1.Nodes.ImportDirectiveNode(astNode, module);
        ret.alias = visitChildTypeOrNull(astNode, 'NameIdentifier');
        return ret;
    },
    VarDirective(astNode) {
        const decl = visit(findChildrenTypeOrFail(astNode, 'VarDeclaration'));
        const ret = new nodes_1.Nodes.VarDirectiveNode(astNode, decl);
        ret.isPublic = !findChildrenType(astNode, 'PrivateModifier');
        return ret;
    },
    ValDirective(astNode) {
        const decl = visit(findChildrenTypeOrFail(astNode, 'ValDeclaration'));
        const ret = new nodes_1.Nodes.VarDirectiveNode(astNode, decl);
        ret.isPublic = !findChildrenType(astNode, 'PrivateModifier');
        return ret;
    },
    EffectDirective(astNode) {
        const ret = new nodes_1.Nodes.EffectDirectiveNode(astNode);
        ret.isPublic = !findChildrenType(astNode, 'PrivateModifier');
        ret.effect = visit(findChildrenTypeOrFail(astNode, 'EffectDeclaration'));
        return ret;
    },
    ImplDirective(astNode) {
        const references = astNode.children.filter(($) => $.type === 'Reference');
        // the last Reference is the target
        const target = visit(references.pop());
        const directivesNode = findChildrenType(astNode, 'NamespaceElementList');
        let directives = directivesNode
            ? directivesNode.children.map(visit)
            : [];
        // TODO: warn
        const ret = new nodes_1.Nodes.ImplDirective(astNode, target, directives);
        if (references.length) {
            ret.baseImpl = visit(references.pop());
        }
        ret.isPublic = !findChildrenType(astNode, 'PrivateModifier');
        return ret;
    },
    EffectDeclaration(astNode) {
        const name = visit(findChildrenTypeOrFail(astNode, 'NameIdentifier', 'A name is required'));
        const ret = new nodes_1.Nodes.EffectDeclarationNode(astNode, name);
        const list = findChildrenType(astNode, 'EffectElementList');
        if (list) {
            ret.elements = list.children.map(($) => visit($));
        }
        else {
            // TODO: warn
        }
        return ret;
    },
    EffectMemberDeclaration(astNode) {
        const ret = new nodes_1.Nodes.EffectMemberDeclarationNode(astNode);
        ret.name = visit(findChildrenTypeOrFail(astNode, 'NameIdentifier'));
        const params = findChildrenType(astNode, 'FunctionParamsList');
        if (!params) {
            throw new NodeError_1.PositionCapableError('Missing param list in function declaration', astNode);
        }
        ret.parameters = params.children.map(($) => visit($));
        return ret;
    },
    VarDeclaration(astNode) {
        const variableName = visit(findChildrenTypeOrFail(astNode, 'NameIdentifier'));
        const value = visitLastChild(astNode);
        const ret = new nodes_1.Nodes.VarDeclarationNode(astNode, variableName, value);
        ret.variableName.annotate(new annotations_1.annotations.MutableDeclaration());
        const type = findChildrenType(astNode, 'Type');
        if (type) {
            ret.variableType = visit(type);
        }
        else {
            // TODO: warn
        }
        return ret;
    },
    ValDeclaration(astNode) {
        const variableName = visit(findChildrenTypeOrFail(astNode, 'NameIdentifier'));
        const value = visitLastChild(astNode);
        const ret = new nodes_1.Nodes.VarDeclarationNode(astNode, variableName, value);
        const type = findChildrenType(astNode, 'Type');
        if (type) {
            ret.variableType = visit(type);
        }
        else {
            // TODO: warn
        }
        return ret;
    },
    TypeDirective(astNode) {
        const children = astNode.children.slice();
        let isPublic = true;
        let child = children.shift();
        if (child.type === 'PrivateModifier') {
            isPublic = false;
            child = children.shift();
        }
        const variableName = visit(child);
        console.assert(!!variableName, 'missing variable name');
        const ret = new nodes_1.Nodes.TypeDirectiveNode(astNode, variableName);
        ret.isPublic = isPublic;
        if (children.length) {
            ret.valueType = visitLastChild(astNode);
        }
        return ret;
    },
    EnumDirective(astNode) {
        const children = astNode.children.slice();
        let child = children.shift();
        let isPublic = true;
        if (child.type === 'PrivateModifier') {
            isPublic = false;
            child = children.shift();
        }
        const variableName = visit(child);
        const typeDeclElements = findChildrenType(astNode, 'TypeDeclElements');
        const declarations = typeDeclElements ? typeDeclElements.children.map(($) => visit($)) : [];
        const ret = new nodes_1.Nodes.EnumDirectiveNode(astNode, variableName, declarations);
        ret.isPublic = isPublic;
        return ret;
    },
    TraitDirective(astNode) {
        const children = astNode.children.slice();
        let child = children.shift();
        let isPublic = true;
        if (child.type === 'PrivateModifier') {
            isPublic = false;
            child = children.shift();
        }
        const variableName = visit(child);
        const typeDeclElements = findChildrenType(astNode, 'TraitDeclElements');
        const declarations = typeDeclElements ? typeDeclElements.children.map(($) => visit($)) : [];
        declarations.forEach(($) => {
            if ($ instanceof nodes_1.Nodes.FunDirectiveNode && !$.functionNode.body) {
                $.functionNode.annotate(new annotations_1.annotations.SignatureDeclaration());
            }
        });
        const ret = new nodes_1.Nodes.TraitDirectiveNode(astNode, variableName, declarations);
        ret.isPublic = isPublic;
        return ret;
    },
    FunDeclaration(astNode) {
        const functionName = visit(findChildrenTypeOrFail(astNode, 'FunctionName', 'A function name is required').children[0]);
        const fun = new nodes_1.Nodes.FunctionNode(astNode, functionName);
        const retType = findChildrenType(astNode, 'Type');
        if (retType) {
            fun.functionReturnType = visit(retType);
        }
        else {
            // TODO: warn
        }
        const params = findChildrenType(astNode, 'FunctionParamsList');
        if (!params) {
            throw new NodeError_1.PositionCapableError('Missing param list in function declaration', astNode);
        }
        fun.parameters = params.children.map(($) => visit($));
        const body = findChildrenType(astNode, 'FunAssignExpression');
        if (body) {
            fun.body = visitLastChild(body);
        }
        return fun;
    },
    Decorator(astNode) {
        const children = astNode.children.slice();
        const name = children.shift();
        if (!name) {
            throw new NodeError_1.PositionCapableError('Missing decorator name', astNode);
        }
        const decoratorName = visit(name);
        const args = children.map(visit);
        const ret = new nodes_1.Nodes.DecoratorNode(astNode, decoratorName, args);
        return ret;
    },
    FunctionDirective(astNode) {
        const decoratorsNode = findChildrenType(astNode, 'Decorators');
        const decorators = decoratorsNode ? decoratorsNode.children.map(visit) : [];
        const functionNode = visit(findChildrenTypeOrFail(astNode, 'FunDeclaration'));
        const ret = new nodes_1.Nodes.FunDirectiveNode(astNode, functionNode, decorators);
        ret.isPublic = !findChildrenType(astNode, 'PrivateModifier');
        return ret;
    },
    CodeBlock(astNode) {
        const statements = astNode.children.map(($) => visit($)).filter(($) => !!$);
        return new nodes_1.Nodes.BlockNode(astNode, statements);
    },
    Parameter(astNode) {
        const parameterName = visit(findChildrenTypeOrFail(astNode, 'NameIdentifier', 'A parameter name is required'));
        const ret = new nodes_1.Nodes.ParameterNode(astNode, parameterName);
        const type = findChildrenType(astNode, 'Type');
        if (type) {
            ret.parameterType = visit(type);
        }
        else {
            // TODO: warn
        }
        return ret;
    },
    AssignExpression: binaryOpVisitor,
    AddExpression: binaryOpVisitor,
    OrExpression: binaryOpVisitor,
    AndExpression: binaryOpVisitor,
    BitOrExpression: binaryOpVisitor,
    BitXorExpression: binaryOpVisitor,
    BitAndExpression: binaryOpVisitor,
    AsExpression: binaryOpVisitor,
    IsExpression: binaryOpVisitor,
    RelExpression: binaryOpVisitor,
    EqExpression: binaryOpVisitor,
    ShiftExpression: binaryOpVisitor,
    MulExpression: binaryOpVisitor,
    ParenExpression(astNode) {
        const ret = visitLastChild(astNode);
        ret.hasParentheses = true;
        return ret;
    },
    AtomicExpression(astNode) {
        let ret = visit(astNode.children[0]);
        for (let currentChildren = 1; currentChildren < astNode.children.length; currentChildren++) {
            const currentNode = astNode.children[currentChildren];
            const doesItHaveCallArguments = currentNode && currentNode.type === 'CallArguments';
            const doesItHaveIndexExpression = currentNode && currentNode.type === 'IndexExpression';
            const doesItHaveMemberExpression = currentNode && currentNode.type === 'MemberExpression';
            if (doesItHaveCallArguments) {
                const argumentsNode = currentNode.children.map(($) => visit($));
                const fnCall = new nodes_1.Nodes.FunctionCallNode(currentNode, ret, argumentsNode);
                fnCall.isInfix = true;
                ret = fnCall;
            }
            else if (doesItHaveIndexExpression) {
                const argument = visit(currentNode.children[0]);
                const indexSelector = new nodes_1.Nodes.BinaryExpressionNode(currentNode, nodes_1.Nodes.NameIdentifierNode.fromString('[]', currentNode), ret, argument);
                ret = indexSelector;
            }
            else if (doesItHaveMemberExpression) {
                ret = new nodes_1.Nodes.MemberNode(astNode, ret, currentNode.children[0].text, visit(currentNode.children[1]));
            }
            else {
                if (currentNode) {
                    throw new NodeError_1.PositionCapableError("Don't know what to do with " + currentNode.type, currentNode);
                }
            }
        }
        return ret;
    },
    MatchExpression(astNode) {
        const lhs = visit(astNode.children[0]);
        const matchingSet = astNode.children[1].children.map(($) => visit($));
        return new nodes_1.Nodes.PatternMatcherNode(astNode, lhs, matchingSet);
    },
    Reference(astNode) {
        return new nodes_1.Nodes.ReferenceNode(astNode, visit(astNode.children[0]));
    },
    FunOperator(astNode) {
        return new nodes_1.Nodes.NameIdentifierNode(astNode, astNode.text.trim());
    },
    NameIdentifier(astNode) {
        return new nodes_1.Nodes.NameIdentifierNode(astNode, astNode.text.trim());
    },
    QName(astNode) {
        const names = astNode.children.map(visit);
        return new nodes_1.Nodes.QNameNode(astNode, names);
    },
    FunctionTypeParameter(astNode) {
        const type = visitChildTypeOrFail(astNode, 'Type');
        const ret = new nodes_1.Nodes.SignatureParameterNode(astNode, type);
        ret.parameterName = visitChildTypeOrNull(astNode, 'NameIdentifier');
        return ret;
    },
    FunctionTypeLiteral(child) {
        const parametersNode = findChildrenType(child, 'FunctionTypeParameters');
        const ret = new nodes_1.Nodes.FunctionTypeNode(child, parametersNode ? parametersNode.children.map(($) => visit($)) : []);
        ret.returnType = visitChildTypeOrNull(child, 'Type');
        return ret;
    },
    Type(astNode) {
        const ret = visit(astNode.children[0]);
        if (false === ret instanceof nodes_1.Nodes.TypeNode && false === ret instanceof nodes_1.Nodes.ReferenceNode) {
            console.error('Node ' + astNode.type + ' did not yield a type node');
            console.log(ret);
        }
        return ret;
    },
    CaseLiteral(x) {
        const literal = visit(x.children[0]);
        const rhs = visit(x.children[1]);
        return new nodes_1.Nodes.MatchLiteralNode(x, literal, rhs);
    },
    CaseCondition(x) {
        const condition = visit(x.children[1]);
        const rhs = visit(x.children[2]);
        const ret = new nodes_1.Nodes.MatchConditionNode(x, condition, rhs);
        ret.declaredName = visit(x.children[0]);
        return ret;
    },
    CaseIs(x) {
        const rhs = visitLastChild(x);
        const typeReference = visitChildTypeOrFail(x, 'Reference');
        const ret = new nodes_1.Nodes.MatchCaseIsNode(x, typeReference, rhs);
        ret.declaredName = visitChildTypeOrNull(x, 'NameIdentifier');
        const deconstruct = findChildrenType(x, 'DeconstructStruct');
        if (deconstruct) {
            ret.deconstructorNames = deconstruct.children.map(($) => visit($));
        }
        return ret;
    },
    CaseElse(x) {
        return new nodes_1.Nodes.MatchDefaultNode(x, visit(x.children[0]));
    },
    LoopExpression(x) {
        return new nodes_1.Nodes.LoopNode(x, visit(x.children[0]));
    },
    ContinueStatement(x) {
        return new nodes_1.Nodes.ContinueNode(x);
    },
    BreakStatement(x) {
        return new nodes_1.Nodes.BreakNode(x);
    },
    PostfixNumber(x) {
        const literal = visit(x.children[0]);
        literal.suffixReference = visit(x.children[1]);
        return literal;
    },
    NumberLiteral(x) {
        if (x.text.includes('.') || x.text.includes('E') || x.text.includes('e')) {
            return new nodes_1.Nodes.FloatLiteral(x, 'f32');
        }
        else {
            let type = 'i64';
            // tslint:disable-next-line:radix
            const intValue = parseInt(x.text);
            if (intValue >= -2147483648 && intValue <= 4294967295) {
                type = 'i32';
            }
            return new nodes_1.Nodes.IntegerLiteral(x, type);
        }
    },
    NegNumberLiteral(x) {
        if (x.text.includes('.') || x.text.includes('E') || x.text.includes('e')) {
            return new nodes_1.Nodes.FloatLiteral(x, 'f32');
        }
        else {
            return new nodes_1.Nodes.IntegerLiteral(x, 'i32');
        }
    },
    HexLiteral(x) {
        const typeName = x.text.length > '0x00000000'.length ? 'u64' : 'u32';
        return new nodes_1.Nodes.HexLiteral(x, typeName);
    },
    StringLiteral(x) {
        try {
            const ret = new nodes_1.Nodes.StringLiteral(x, 'string');
            ret.value = JSON.parse(x.text);
            return ret;
        }
        catch (e) {
            const err = 'Cannot parse string: ' + x.text;
            console.log(err);
            return new NodeError_1.PositionCapableError(err, x);
        }
    },
    BinNegExpression(x) {
        return new nodes_1.Nodes.UnaryExpressionNode(x, nodes_1.Nodes.NameIdentifierNode.fromString('~', x), visit(x.children[0]));
    },
    NegExpression(x) {
        return new nodes_1.Nodes.UnaryExpressionNode(x, nodes_1.Nodes.NameIdentifierNode.fromString('!', x), visit(x.children[0]));
    },
    UnaryMinus(x) {
        return new nodes_1.Nodes.UnaryExpressionNode(x, nodes_1.Nodes.NameIdentifierNode.fromString('-', x), visit(x.children[0]));
    },
    BooleanLiteral(x) {
        return new nodes_1.Nodes.BooleanLiteral(x, 'boolean');
    },
    Document(astNode) {
        const doc = new nodes_1.Nodes.DocumentNode(astNode);
        astNode.children.forEach(($) => doc.directives.push(visit($)));
        return doc;
    },
    IfExpression(astNode) {
        const condition = visit(astNode.children[0]);
        const truePart = visit(astNode.children[1]);
        if (astNode.children[2]) {
            const falsePart = visit(astNode.children[2]);
            return new nodes_1.Nodes.IfNode(astNode, condition, truePart, falsePart);
        }
        else {
            return new nodes_1.Nodes.IfNode(astNode, condition, truePart);
        }
    },
    SyntaxError(node) {
        return new NodeError_1.PositionCapableError(node.errors[0].message, node);
    },
    StructDirective(astNode) {
        const children = astNode.children.slice();
        let child = children[0];
        let isPublic = true;
        if (child.type === 'PrivateModifier') {
            isPublic = false;
            child = children.shift();
        }
        const ret = visit(child);
        ret.isPublic = isPublic;
        return ret;
    },
    UnknownExpression(astNode) {
        return new nodes_1.Nodes.UnknownExpressionNode(astNode);
    },
    StructLiteral(astNode) {
        const parametersNode = findChildrenTypeOrFail(astNode, 'StructParamsList');
        const parameters = parametersNode.children.filter(($) => $.type === 'Parameter').map(($) => visit($));
        return new nodes_1.Nodes.StructTypeNode(astNode, parameters);
    },
    StackLiteral(astNode) {
        const ret = new nodes_1.Nodes.StackTypeNode(astNode);
        ret.metadata = Object.create(null);
        for (let namePair of astNode.children) {
            ret.metadata[namePair.children[0].text] = visit(namePair.children[1]);
        }
        return ret;
    },
    InjectedLiteral(astNode) {
        const ret = new nodes_1.Nodes.InjectedTypeNode(astNode);
        return ret;
    },
    StructDeclaration(astNode) {
        const declaredName = visitChildTypeOrNull(astNode, 'NameIdentifier');
        const params = findChildrenType(astNode, 'FunctionParamsList');
        return new nodes_1.Nodes.StructDeclarationNode(astNode, declaredName, params ? params.children.map(($) => visit($)) : []);
    },
    UnionType(astNode) {
        const ret = new nodes_1.Nodes.UnionTypeNode(astNode);
        ret.of = astNode.children.map(($) => visit($));
        return ret;
    },
    IntersectionType(astNode) {
        const ret = new nodes_1.Nodes.IntersectionTypeNode(astNode);
        ret.of = astNode.children.map(($) => visit($));
        return ret;
    },
    TypeParen(astNode) {
        const ret = visit(astNode.children[0]);
        ret.hasParentheses = true;
        return ret;
    },
    WasmExpression(astNode) {
        const atoms = astNode.children.map(($) => visit($));
        return new nodes_1.Nodes.WasmExpressionNode(astNode, atoms);
    },
    SExpression(astNode) {
        const children = astNode.children.slice();
        const symbol = children.shift();
        const newChildren = children.map(($) => visit($));
        const ret = new nodes_1.Nodes.WasmAtomNode(astNode, symbol.text, newChildren);
        if (ret.symbol === 'call' ||
            ret.symbol === 'global.get' ||
            ret.symbol === 'global.set' ||
            ret.symbol === 'get_global' ||
            ret.symbol === 'set_global') {
            if (ret.args[0] instanceof nodes_1.Nodes.QNameNode) {
                const qname = ret.args[0];
                const varRef = new nodes_1.Nodes.ReferenceNode(ret.args[0].astNode, qname);
                if (qname.names[0].name.startsWith('$')) {
                    // TODO: fix horrible hack $
                    qname.names[0].name = qname.names[0].name.replace(/^\$/, '');
                }
                ret.args[0] = varRef;
            }
        }
        return ret;
    },
};
function visit(astNode) {
    if (!astNode) {
        console.trace();
        throw new Error('astNode is null');
    }
    if (visitor[astNode.type]) {
        const x = visitor[astNode.type](astNode);
        if (!x) {
            throw new NodeError_1.PositionCapableError('Error visiting node ' + astNode.type, astNode);
        }
        return x;
    }
    else {
        throw new NodeError_1.PositionCapableError(`Visitor not implemented for ${astNode.type}`, astNode);
    }
}
function findChildrenTypeOrFail(token, type, message) {
    const ret = token.children.find(($) => $.type === type);
    if (!ret)
        throw new NodeError_1.PositionCapableError(message || `Cannot find child node of type ${type}`, token);
    return ret;
}
function findChildrenType(token, type) {
    return token.children.find(($) => $.type === type);
}
function visitChildTypeOrNull(token, type) {
    const child = findChildrenType(token, type);
    if (!child)
        return null;
    return visit(child);
}
function visitChildTypeOrFail(token, type) {
    const child = findChildrenTypeOrFail(token, type);
    return visit(child);
}
function visitLastChild(token) {
    return visit(token.children[token.children.length - 1]);
}
function getAST(fileName, moduleName, content, parsingContext) {
    const parsingTree = getParsingTree(moduleName, content, parsingContext);
    if (!parsingTree) {
        throw new Error('parsing phase did not run or failed');
    }
    process(parsingTree, parsingContext);
    try {
        let document = visit(parsingTree);
        document.moduleName = moduleName;
        document.fileName = fileName;
        document.content = content;
        parsingContext.modulesInContext.set(moduleName, document);
        return document;
    }
    catch (e) {
        if (e instanceof NodeError_1.PositionCapableError) {
            let document = new nodes_1.Nodes.DocumentNode(parsingTree);
            document.moduleName = moduleName;
            document.fileName = fileName;
            document.content = content;
            parsingContext.modulesInContext.set(moduleName, document);
            parsingContext.messageCollector.error(e);
            return document;
        }
        throw e;
    }
}
exports.getAST = getAST;
//# sourceMappingURL=canonicalPhase.js.map