"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.executeCompilationPhase = exports.executePreCompilationPhase = exports.detectTailCall = void 0;
const nodes_1 = require("../nodes");
const walker_1 = require("../walker");
const nodeHelpers_1 = require("../nodeHelpers");
const NodeError_1 = require("../NodeError");
const annotations_1 = require("../annotations");
const types_1 = require("../types");
const helpers_1 = require("../helpers");
const helpers_2 = require("./helpers");
const assert = require("assert");
const resolveLocals = (0, walker_1.walkPreOrder)((node, _, _parent) => {
    if (node instanceof nodes_1.Nodes.PatternMatcherNode) {
        // create a local for lhs of MatchNode
        /**
         *   match aFunction(123) { ... }
         *         ^^^^^^^^^^^^^^ this part will get its own local
         */
        const fn = (0, nodeHelpers_1.findParentType)(node, nodes_1.Nodes.FunctionNode);
        if (fn) {
            const type = types_1.TypeHelpers.getNodeType(node.lhs);
            if (type) {
                const localAnnotation = new annotations_1.annotations.LocalIdentifier(fn.getTempLocal(type));
                node.annotate(localAnnotation);
                node.matchingSet.forEach($ => {
                    $.annotate(localAnnotation);
                });
            }
            else {
                throw new NodeError_1.LysCompilerError('node.lhs.ofType is undefined', node.lhs);
            }
        }
        else {
            // TODO: what if we reach here?
        }
    }
}, (node) => {
    if (node instanceof nodes_1.Nodes.PatternMatcherNode) {
        // release the local for lhs of MatchNode
        const nodeLocal = node.getAnnotation(annotations_1.annotations.LocalIdentifier);
        if (nodeLocal && nodeLocal.local instanceof nodes_1.Local) {
            const fn = (0, nodeHelpers_1.findParentType)(node, nodes_1.Nodes.FunctionNode);
            if (fn) {
                fn.freeTempLocal(nodeLocal.local);
            }
            else {
                // TODO: error
            }
        }
    }
});
const resolveVariables = (0, walker_1.walkPreOrder)((node, parsingContext) => {
    if (node instanceof nodes_1.Nodes.ReferenceNode || node instanceof nodes_1.Nodes.MemberNode) {
        if (node.resolvedReference) {
            if (node.resolvedReference.type === 'VALUE') {
                const decl = node.resolvedReference.referencedNode.parent; // NameIdentifierNode#parent
                if (decl instanceof nodes_1.Nodes.ParameterNode ||
                    decl instanceof nodes_1.Nodes.VarDeclarationNode ||
                    decl instanceof nodes_1.Nodes.MatcherNode) {
                    const declLocal = decl.getAnnotation(annotations_1.annotations.LocalIdentifier);
                    if (!declLocal) {
                        throw new NodeError_1.LysCompilerError(`Node ${decl.nodeName} has no .local`, decl);
                    }
                    node.annotate(declLocal);
                }
                else {
                    throw new NodeError_1.LysCompilerError(`Value node has no local`, decl);
                }
            }
        }
        else {
            if (node instanceof nodes_1.Nodes.ReferenceNode) {
                parsingContext.messageCollector.errorIfBranchDoesntHaveAny(new NodeError_1.LysCompilerError(`Reference was not resolved`, node));
            }
        }
    }
});
const resolveDeclarations = (0, walker_1.walkPreOrder)((node, parsingContext, parent) => {
    if (node instanceof nodes_1.Nodes.VarDeclarationNode) {
        if (parent instanceof nodes_1.Nodes.DirectiveNode) {
            node.annotate(new annotations_1.annotations.LocalIdentifier(new nodes_1.Global(node.variableName)));
        }
        else {
            const fn = (0, nodeHelpers_1.findParentType)(node, nodes_1.Nodes.FunctionNode);
            if (fn) {
                node.annotate(new annotations_1.annotations.LocalIdentifier(fn.addLocal(types_1.TypeHelpers.getNodeType(node.value), node.variableName)));
            }
            else {
                parsingContext.messageCollector.errorIfBranchDoesntHaveAny("Don't know how to generate a local for this", node);
            }
        }
    }
});
exports.detectTailCall = (0, walker_1.walkPreOrder)((node) => {
    if (node instanceof nodes_1.Nodes.FunctionNode && node.body) {
        const isTailRec = isRecursiveCallExpression(node, node.body);
        if (isTailRec) {
            node.annotate(new annotations_1.annotations.IsTailRec());
        }
    }
});
function annotateReturnExpressions(node) {
    if (!node.hasAnnotation(annotations_1.annotations.IsValueNode))
        return;
    if (node instanceof nodes_1.Nodes.PatternMatcherNode) {
        node.matchingSet.forEach($ => annotateReturnExpressions($.rhs));
    }
    else if (node instanceof nodes_1.Nodes.BlockNode) {
        if (node.statements.length > 0) {
            annotateReturnExpressions((0, helpers_1.last)(node.statements));
        }
    }
    else if (node instanceof nodes_1.Nodes.IfNode) {
        node.annotate(new annotations_1.annotations.IsReturnExpression());
        annotateReturnExpressions(node.truePart);
        if (node.falsePart) {
            annotateReturnExpressions(node.falsePart);
        }
    }
    else {
        node.annotate(new annotations_1.annotations.IsReturnExpression());
    }
}
// find the last expression of every function body and mark the return expressions
const detectReturnExpressions = (0, walker_1.walkPreOrder)((node) => {
    if (node instanceof nodes_1.Nodes.FunctionNode && node.body) {
        annotateReturnExpressions(node.body);
    }
});
/**
 * Returns true if the node is a recursive call to the specified function identifier
 *
 * @param functionNode The function identifier
 * @param node               The node to check
 * @return True if it is a recursive call to that
 */
function isRecursiveCallExpression(functionNode, node) {
    if (node instanceof nodes_1.Nodes.FunctionCallNode) {
        return isRecursiveFunctionCall(functionNode, node);
    }
    if (node instanceof nodes_1.Nodes.BlockNode) {
        return node.statements.length > 0 && isRecursiveCallExpression(functionNode, (0, helpers_1.last)(node.statements));
    }
    if (node instanceof nodes_1.Nodes.IfNode) {
        const truePart = isRecursiveCallExpression(functionNode, node.truePart);
        const falsePart = !!node.falsePart && isRecursiveCallExpression(functionNode, node.falsePart);
        return truePart || falsePart;
    }
    if (node instanceof nodes_1.Nodes.PatternMatcherNode) {
        return node.matchingSet.some(pattern => {
            return isRecursiveCallExpression(functionNode, pattern.rhs);
        });
    }
    return false;
}
function isRecursiveFunctionCall(_functionNode, fcn) {
    if (fcn.functionNode instanceof nodes_1.Nodes.ReferenceNode) {
        // Make sure the variable reference points to the same function
        const referencedType = types_1.TypeHelpers.getNodeType(fcn.functionNode);
        const functionNode = (0, nodeHelpers_1.findParentType)(fcn, nodes_1.Nodes.FunctionNode);
        if (functionNode && referencedType instanceof types_1.FunctionType) {
            const isTailRec = referencedType.name.internalIdentifier === functionNode.functionName.internalIdentifier;
            if (isTailRec) {
                fcn.annotate(new annotations_1.annotations.IsTailRecCall());
                fcn.removeAnnotation(annotations_1.annotations.IsReturnExpression);
            }
            return isTailRec;
        }
    }
    return false;
}
function findImportedModulesRecursive(moduleName, parsingContext, queue, visited = new Set()) {
    if (visited.has(moduleName))
        return;
    visited.add(moduleName);
    const document = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.TypeInitialization);
    document.importedModules.forEach(moduleName => {
        findImportedModulesRecursive(moduleName, parsingContext, queue, visited);
    });
    if (!queue.includes(document)) {
        queue.push(document);
    }
}
function executePreCompilationPhase(moduleName, parsingContext) {
    const document = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.PreCompilation - 1);
    assert(document.analysis.nextPhase === nodes_1.PhaseFlags.PreCompilation);
    (0, helpers_2.fixParents)(document, parsingContext, null);
    detectReturnExpressions(document, parsingContext, null);
    (0, exports.detectTailCall)(document, parsingContext, null);
    resolveDeclarations(document, parsingContext, null);
    document.analysis.nextPhase++;
}
exports.executePreCompilationPhase = executePreCompilationPhase;
function executeCompilationPhase(moduleName, parsingContext) {
    const document = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.Compilation - 1);
    const analysisQueue = [];
    // Find every dependency
    findImportedModulesRecursive(moduleName, parsingContext, analysisQueue);
    // Ensure every dependency has executed the PreCompilation phase
    analysisQueue.forEach(document => {
        parsingContext.getPhase(document.moduleName, nodes_1.PhaseFlags.PreCompilation);
    });
    assert(document.analysis.nextPhase === nodes_1.PhaseFlags.Compilation);
    resolveLocals(document, parsingContext, null);
    resolveVariables(document, parsingContext, null);
    (0, helpers_2.fixParents)(document, parsingContext, null);
    document.analysis.nextPhase++;
}
exports.executeCompilationPhase = executeCompilationPhase;
//# sourceMappingURL=compilationPhase.js.map