"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.executeScopePhase = exports.executeNameInitializationPhase = void 0;
const nodes_1 = require("../nodes");
const walker_1 = require("../walker");
const annotations_1 = require("../annotations");
const NodeError_1 = require("../NodeError");
const types_1 = require("../types");
const nodeHelpers_1 = require("../nodeHelpers");
const helpers_1 = require("./helpers");
const Scope_1 = require("../Scope");
const assert = require("assert");
const errorPrinter_1 = require("../../utils/errorPrinter");
const CORE_LIB = 'system::core';
const valueNodeAnnotation = new annotations_1.annotations.IsValueNode();
const findValueNodes = (0, walker_1.walkPreOrder)((node) => {
    /**
     * This phase traverses all nodes and adds an annotation to the value nodes, value nodes are those nodes that
     * must have a value.
     * Value nodes are usually the RHS of an assigment node, the LHS of match nodes, function call arguments and so on.
     */
    if (node instanceof nodes_1.Nodes.FunctionCallNode) {
        node.argumentsNode.forEach($ => $.annotate(valueNodeAnnotation));
    }
    if (node instanceof nodes_1.Nodes.VarDeclarationNode) {
        node.value.annotate(valueNodeAnnotation);
    }
    if (node instanceof nodes_1.Nodes.AssignmentNode) {
        node.rhs.annotate(valueNodeAnnotation);
        node.lhs.annotate(new annotations_1.annotations.IsAssignationLHS());
    }
    if (node instanceof nodes_1.Nodes.MemberNode) {
        if (node.lhs instanceof nodes_1.Nodes.MemberNode) {
            node.lhs.annotate(valueNodeAnnotation);
        }
    }
    if (node instanceof nodes_1.Nodes.FunctionNode) {
        let returnsVoidValue = false;
        if (node.functionReturnType) {
            // TODO: dehardcode this... Unit type may be a good approach
            returnsVoidValue =
                node.functionReturnType instanceof nodes_1.Nodes.ReferenceNode && node.functionReturnType.variable.text === 'void';
        }
        if (!returnsVoidValue) {
            if (node.body) {
                node.body.annotate(valueNodeAnnotation);
            }
            else {
                // TODO: warn
            }
        }
    }
    if (node instanceof nodes_1.Nodes.BinaryExpressionNode) {
        node.lhs.annotate(valueNodeAnnotation);
        node.rhs.annotate(valueNodeAnnotation);
    }
    if (node instanceof nodes_1.Nodes.IsExpressionNode) {
        node.lhs.annotate(valueNodeAnnotation);
    }
    if (node instanceof nodes_1.Nodes.AsExpressionNode) {
        node.lhs.annotate(valueNodeAnnotation);
    }
    if (node instanceof nodes_1.Nodes.IfNode) {
        node.condition.annotate(valueNodeAnnotation);
        if (node.falsePart) {
            if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                node.truePart.annotate(valueNodeAnnotation);
                node.falsePart.annotate(valueNodeAnnotation);
            }
        }
    }
    if (node instanceof nodes_1.Nodes.PatternMatcherNode) {
        node.lhs.annotate(valueNodeAnnotation);
        if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
            node.matchingSet.forEach($ => {
                $.annotate(valueNodeAnnotation);
                $.rhs.annotate(valueNodeAnnotation);
            });
        }
    }
    if (node instanceof nodes_1.Nodes.BlockNode) {
        if (node.hasAnnotation(annotations_1.annotations.IsValueNode) && node.statements.length > 0) {
            node.statements[node.statements.length - 1].annotate(valueNodeAnnotation);
        }
    }
});
const createScopes = (0, walker_1.walkPreOrder)((node, parsingContext, parent) => {
    if (parent) {
        if (!node.scope) {
            node.scope = parent.scope;
        }
        if (node instanceof nodes_1.Nodes.MatcherNode) {
            node.rhs.scope = node.scope.newChildScope('MatcherRHS');
            if (node.declaredName) {
                node.rhs.scope.set(node.declaredName, 'VALUE', false);
            }
            if (node instanceof nodes_1.Nodes.MatchCaseIsNode) {
                if (node.deconstructorNames) {
                    const takenNames = new Map();
                    node.deconstructorNames.forEach($ => {
                        if ($.name !== '_') {
                            if (takenNames.has($.name)) {
                                parsingContext.messageCollector.error(new NodeError_1.LysScopeError('Duplicated name', $));
                            }
                            else {
                                takenNames.set($.name, $);
                                node.rhs.scope.set($, 'VALUE', false);
                            }
                        }
                    });
                }
            }
        }
        else if (node instanceof nodes_1.Nodes.OverloadedFunctionNode) {
            node.scope.set(node.functionName, 'FUNCTION', node.isPublic);
        }
        else if (node instanceof nodes_1.Nodes.TraitDirectiveNode) {
            node.scope.set(node.traitName, 'TRAIT', node.isPublic);
            node.scope = node.scope.newChildScope(node.traitName.name + '.');
            node.selfTypeName = new nodes_1.Nodes.NameIdentifierNode(node.traitName.astNode, 'Self');
            node.scope.set(node.selfTypeName, 'TYPE', false);
        }
        else if (node instanceof nodes_1.Nodes.VarDeclarationNode) {
            if (node.variableName.name in types_1.InjectableTypes) {
                parsingContext.messageCollector.error(new NodeError_1.LysScopeError('Cannot declare a variable with the name of an system type', node.variableName));
            }
            node.value.scope = node.scope.newChildScope(node.variableName.name + '_Declaration');
            node.scope.set(node.variableName, 'VALUE', parent instanceof nodes_1.Nodes.VarDirectiveNode && parent.isPublic);
        }
        else if (node instanceof nodes_1.Nodes.ImplDirective) {
            node.scope = node.scope.newChildScope(node.targetImpl.variable.text + '.');
            if (node.baseImpl) {
                node.selfTypeName = new nodes_1.Nodes.NameIdentifierNode(node.targetImpl.astNode, 'Self');
                node.scope.set(node.selfTypeName, 'TYPE', false);
            }
        }
        else if (node instanceof nodes_1.Nodes.TypeDirectiveNode) {
            node.scope.set(node.variableName, 'TYPE', node.isPublic);
        }
        else if (node instanceof nodes_1.Nodes.FunctionNode) {
            if (node.functionName) {
                if (!(parent instanceof nodes_1.Nodes.DirectiveNode)) {
                    node.scope.set(node.functionName, 'VALUE', false);
                }
                else if (!node.functionName.internalIdentifier) {
                    node.functionName.internalIdentifier = node.scope.getInternalIdentifier(node.functionName);
                }
            }
            if (node.body) {
                // if (node.body == false) a message is emited in the semantic phase
                const functionBody = node.body;
                if (node.functionName) {
                    node.body.scope = node.scope.newChildScope(node.functionName.name + '_Body');
                }
                else {
                    node.body.scope = node.scope.newChildScope('Body');
                }
                node.parameters.forEach($ => {
                    functionBody.scope.set($.parameterName, 'VALUE', false);
                });
            }
            node.processParameters();
        }
        else if (node instanceof nodes_1.Nodes.BlockNode) {
            node.scope = node.scope.newChildScope('Block');
        }
    }
});
function collectNamespaces(namespace, directives, parsingContext) {
    const { namespaceNames } = namespace;
    const nameToShow = namespace instanceof nodes_1.Nodes.ImplDirective ? namespace.targetImpl.variable.text : namespace.traitName.name;
    function registerNameIdentifier(nameNode) {
        if (namespaceNames.has(nameNode.name) && namespaceNames.get(nameNode.name) !== nameNode) {
            parsingContext.messageCollector.error(`The name "${nameNode.name}" is already registered in the impl "${nameToShow}"`, nameNode.astNode);
            parsingContext.messageCollector.error(`This is the registered name "${nameNode.name}" of "${nameToShow}"`, namespaceNames.get(nameNode.name).astNode);
        }
        else {
            namespaceNames.set(nameNode.name, nameNode);
        }
    }
    directives.forEach(node => {
        if (node instanceof nodes_1.Nodes.OverloadedFunctionNode) {
            registerNameIdentifier(node.functionName);
        }
        else if (node instanceof nodes_1.Nodes.VarDirectiveNode) {
            registerNameIdentifier(node.decl.variableName);
        }
        else if (node instanceof nodes_1.Nodes.TypeDirectiveNode) {
            registerNameIdentifier(node.variableName);
        }
        else {
            parsingContext.messageCollector.error(`Don't know how to register this directive ${node.nodeName}`, node.astNode);
        }
    });
}
const resolveVariables = (0, walker_1.walkPreOrder)(undefined, (node, parsingContext) => {
    if (node instanceof nodes_1.Nodes.IsExpressionNode || node instanceof nodes_1.Nodes.IfNode || node instanceof nodes_1.Nodes.MatcherNode) {
        const typeName = 'boolean';
        try {
            node.scope.get(typeName);
        }
        catch (e) {
            throw new NodeError_1.LysScopeError(e.toString(), node);
        }
        const resolved = node.scope.get(typeName);
        node.booleanReference = resolved;
        node.scope.incrementUsage(typeName);
    }
    else if (node instanceof nodes_1.Nodes.LiteralNode) {
        try {
            node.scope.get(node.typeName);
        }
        catch (e) {
            throw new NodeError_1.LysScopeError(e.toString(), node);
        }
        const resolved = node.scope.get(node.typeName);
        node.resolvedReference = resolved;
        node.scope.incrementUsage(node.typeName);
    }
    else if (node instanceof nodes_1.Nodes.ReferenceNode) {
        try {
            node.scope.getQName(node.variable);
        }
        catch (e) {
            throw new NodeError_1.LysScopeError(e.toString(), node.variable);
        }
        const resolved = node.scope.getQName(node.variable);
        const document = (0, helpers_1.getDocument)(node);
        const isGlobal = !resolved.isLocalReference || resolved.scope === document.scope;
        node.isLocal = !isGlobal;
        node.resolvedReference = resolved;
        node.scope.incrementUsageQName(node.variable);
    }
    else if (node instanceof nodes_1.Nodes.ImplDirective) {
        if (node.targetImpl.resolvedReference) {
            collectNamespaces(node, node.directives, parsingContext);
        }
        else {
            throw new NodeError_1.LysScopeError(`Impl is not resolved`, node);
        }
    }
    else if (node instanceof nodes_1.Nodes.TraitDirectiveNode) {
        collectNamespaces(node, node.directives, parsingContext);
    }
    else if (node instanceof nodes_1.Nodes.ImportDirectiveNode) {
        try {
            parsingContext.getPhase(node.module.text, nodes_1.PhaseFlags.NameInitialization);
        }
        catch (e) {
            parsingContext.messageCollector.error(`Unable to load module ${node.module.text}: ` + e, node.astNode);
        }
    }
});
const findImplicitImports = (0, walker_1.walkPreOrder)((node, parsingContext) => {
    if (node instanceof nodes_1.Nodes.ImportDirectiveNode) {
        const document = (0, helpers_1.getDocument)(node);
        if (node.module.text === document.moduleName) {
            // TODO: test this
            parsingContext.messageCollector.error('Self import is not allowed', node.astNode);
        }
        const importAll = node.allItems ? new Set(['*']) : new Set();
        node.scope.registerImport(node.module.text, importAll);
    }
    else if (node instanceof nodes_1.Nodes.ReferenceNode) {
        if (node.variable.names.length > 1) {
            const { moduleName, variable } = node.variable.deconstruct();
            const document = (0, helpers_1.getDocument)(node);
            if (moduleName === document.moduleName) {
                // TODO: test this
                parsingContext.messageCollector.error('Self import is not allowed', node.astNode);
            }
            node.scope.registerImport(moduleName, new Set([variable]));
        }
    }
});
const injectImplicitCalls = (0, walker_1.walkPreOrder)((node, _) => {
    if (node instanceof nodes_1.Nodes.FunctionCallNode && node.functionNode instanceof nodes_1.Nodes.ReferenceNode) {
        if (node.functionNode.resolvedReference && node.functionNode.resolvedReference.type === 'TYPE') {
            const member = new nodes_1.Nodes.MemberNode(node.functionNode.astNode, node.functionNode, '.', new nodes_1.Nodes.NameIdentifierNode(node.functionNode.astNode, 'apply'));
            member.scope = node.functionNode.scope;
            node.functionNode = member;
        }
    }
});
function injectCoreImport(document, parsingContext) {
    if (document.moduleName.startsWith(CORE_LIB)) {
        return;
    }
    if (document.hasAnnotation(annotations_1.annotations.NoStd)) {
        return;
    }
    const coreLib = parsingContext.getPhase(CORE_LIB, nodes_1.PhaseFlags.Semantic, false);
    coreLib.directives.reverse().forEach(directive => {
        if (directive instanceof nodes_1.Nodes.ImportDirectiveNode) {
            const newDirective = new nodes_1.Nodes.ImportDirectiveNode(directive.astNode, directive.module);
            newDirective.allItems = directive.allItems;
            newDirective.alias = directive.alias;
            document.directives.unshift(newDirective);
        }
        else {
            throw new errorPrinter_1.LysError('Only import directives are allowed in system::core, found: ' + directive.nodeName);
        }
    });
}
function summarizeImports(document, parsingContext) {
    (0, helpers_1.collectImports)(document.importedModules, document, parsingContext);
    document.importedModules.delete(document.moduleName);
    document.importedModules.forEach(moduleName => {
        const requiredDocument = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.Semantic);
        if (requiredDocument !== document) {
            requiredDocument.importedBy.add(document.moduleName);
        }
    });
}
const unreachableAnnotation = new annotations_1.annotations.IsUnreachable();
const validateLoops = (0, walker_1.walkPreOrder)((node, parsingContext) => {
    if (node instanceof nodes_1.Nodes.ContinueNode || node instanceof nodes_1.Nodes.BreakNode) {
        const relevantParent = (0, nodeHelpers_1.findParentDelegate)(node, $ => {
            return ($ instanceof nodes_1.Nodes.LoopNode || $ instanceof nodes_1.Nodes.FunctionNode || $.hasAnnotation(annotations_1.annotations.IsValueNode));
        });
        if (relevantParent instanceof nodes_1.Nodes.LoopNode) {
            node.annotate(new annotations_1.annotations.CurrentLoop(relevantParent));
        }
        else {
            if (relevantParent instanceof nodes_1.Nodes.FunctionNode) {
                parsingContext.messageCollector.error(`Invalid location: No loop was found`, node.astNode);
            }
            else {
                parsingContext.messageCollector.error(`Invalid location. Parent block must return a value`, node.astNode);
                if (relevantParent) {
                    parsingContext.messageCollector.error(`Not all paths return a value`, relevantParent.astNode);
                }
            }
        }
    }
}, (node, parsingContext) => {
    if (node instanceof nodes_1.Nodes.BlockNode) {
        let nextAreUnreachable = false;
        node.statements.forEach($ => {
            if (nextAreUnreachable) {
                parsingContext.messageCollector.error(new NodeError_1.UnreachableCode($));
                $.annotate(unreachableAnnotation);
            }
            if ($ instanceof nodes_1.Nodes.ContinueNode || $ instanceof nodes_1.Nodes.BreakNode) {
                if ($.hasAnnotation(annotations_1.annotations.CurrentLoop)) {
                    nextAreUnreachable = true;
                }
            }
        });
    }
});
const validateMutability = (0, walker_1.walkPreOrder)((node, parsingContext) => {
    if (node instanceof nodes_1.Nodes.AssignmentNode) {
        if (node.lhs instanceof nodes_1.Nodes.ReferenceNode && node.lhs.resolvedReference) {
            if (!node.lhs.resolvedReference.referencedNode.hasAnnotation(annotations_1.annotations.MutableDeclaration)) {
                parsingContext.messageCollector.error(`${node.lhs.resolvedReference.referencedNode.name} is immutable, reassignment is not allowed.`, node.lhs.astNode);
            }
        }
    }
});
function executeNameInitializationPhase(moduleName, parsingContext) {
    const document = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.NameInitialization - 1);
    assert(document.analysis.nextPhase === nodes_1.PhaseFlags.NameInitialization);
    assert(document.moduleName === moduleName);
    document.scope = new Scope_1.Scope(parsingContext, document.moduleName, null, '');
    injectCoreImport(document, parsingContext);
    createScopes(document, parsingContext, null);
    (0, helpers_1.fixParents)(document, parsingContext);
    findImplicitImports(document, parsingContext, null);
    document.analysis.nextPhase++;
    return;
}
exports.executeNameInitializationPhase = executeNameInitializationPhase;
function executeScopePhase(moduleName, parsingContext) {
    const document = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.Scope - 1);
    assert(document.analysis.nextPhase === nodes_1.PhaseFlags.Scope);
    resolveVariables(document, parsingContext, null);
    findValueNodes(document, parsingContext, null);
    injectImplicitCalls(document, parsingContext, null);
    validateLoops(document, parsingContext);
    validateMutability(document, parsingContext);
    summarizeImports(document, parsingContext);
    document.analysis.nextPhase++;
}
exports.executeScopePhase = executeScopePhase;
//# sourceMappingURL=scopePhase.js.map