"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.executeTypeCheck = exports.executeTypeInitialization = void 0;
const nodes_1 = require("../nodes");
const walker_1 = require("../walker");
const types_1 = require("../types");
const NodeError_1 = require("../NodeError");
const helpers_1 = require("./helpers");
const TypeAnalyzer_1 = require("../types/TypeAnalyzer");
const assert = require("assert");
const initializeTypes = (0, walker_1.walkPreOrder)((node, parsingContext) => {
    if (node instanceof nodes_1.Nodes.ImplDirective) {
        if (node.targetImpl.resolvedReference) {
            if (node.baseImpl && node.baseImpl.resolvedReference) {
                for (let impl of node.targetImpl.resolvedReference.referencedNode.impls) {
                    if (impl.baseImpl &&
                        impl.baseImpl.resolvedReference &&
                        impl.baseImpl.resolvedReference.referencedNode === node.baseImpl.resolvedReference.referencedNode) {
                        parsingContext.messageCollector.error(`This trait is already implemented for the target type`, node.baseImpl.astNode);
                    }
                }
            }
            node.targetImpl.resolvedReference.referencedNode.impls.add(node);
        }
    }
}, (node, parsingContext, parent) => {
    if (node instanceof nodes_1.Nodes.StructTypeNode) {
        types_1.TypeHelpers.setNodeType(node, new types_1.StructType(node.parameters));
    }
    else if (node instanceof nodes_1.Nodes.StackTypeNode) {
        const lowLevelType = node.metadata['lowLevelType'];
        const byteSize = node.metadata['byteSize'];
        if (!lowLevelType) {
            parsingContext.messageCollector.error(`Missing lowLevelType schema`, node.astNode);
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes.never);
            return;
        }
        else if (!(lowLevelType instanceof nodes_1.Nodes.StringLiteral)) {
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes.never);
            parsingContext.messageCollector.error(`lowLevelType must be a string`, lowLevelType.astNode);
            return;
        }
        if (!byteSize) {
            parsingContext.messageCollector.error(`Missing byteSize schema`, node.astNode);
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes.never);
            return;
        }
        else if (!(byteSize instanceof nodes_1.Nodes.IntegerLiteral)) {
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes.never);
            parsingContext.messageCollector.error(`byteSize must be a number`, lowLevelType.astNode);
            return;
        }
        let typeName = '???';
        if (parent instanceof nodes_1.Nodes.TypeDirectiveNode) {
            typeName = parent.variableName.name;
        }
        else {
            parsingContext.messageCollector.error(`%stack can only be used in type directives`, node.astNode);
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes.never);
            return;
        }
        const type = lowLevelType.value;
        if (type in types_1.NativeTypes) {
            const newType = types_1.StackType.of(typeName, type, byteSize.value);
            types_1.TypeHelpers.setNodeType(node, newType);
        }
        else {
            parsingContext.messageCollector.error(`Unknown lowLevelType ${lowLevelType.value}`, lowLevelType.astNode);
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes.never);
        }
    }
    else if (node instanceof nodes_1.Nodes.InjectedTypeNode) {
        let typeName = '???';
        if (parent instanceof nodes_1.Nodes.TypeDirectiveNode) {
            typeName = parent.variableName.name;
        }
        else {
            parsingContext.messageCollector.error(`%injected can only be used in type directives`, node.astNode);
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes.never);
            return;
        }
        if (typeName in types_1.InjectableTypes) {
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes[typeName]);
        }
        else {
            parsingContext.messageCollector.error(`Unknown injectable type`, node.astNode);
            types_1.TypeHelpers.setNodeType(node, types_1.InjectableTypes.never);
        }
    }
    if (node instanceof nodes_1.Nodes.TraitDirectiveNode) {
        const traitType = new types_1.TraitType(node);
        types_1.TypeHelpers.setNodeType(node.traitName, traitType);
        types_1.TypeHelpers.setNodeType(node.selfTypeName, types_1.TypeType.of(new types_1.SelfType(traitType)));
    }
    if (node instanceof nodes_1.Nodes.TypeDirectiveNode) {
        if (!node.valueType) {
            parsingContext.messageCollector.error(`Missing type`, node.astNode);
        }
    }
});
/// ---
// Assign types to names, create type references
function executeTypeInitialization(moduleName, parsingContext) {
    const document = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.TypeInitialization - 1);
    assert(document.analysis.nextPhase === nodes_1.PhaseFlags.TypeInitialization);
    (0, helpers_1.fixParents)(document, parsingContext);
    initializeTypes(document, parsingContext);
    document.analysis.nextPhase++;
}
exports.executeTypeInitialization = executeTypeInitialization;
function findNonFinalizedImportsRecursive(moduleName, parsingContext, queue, visited = new Set()) {
    if (visited.has(moduleName))
        return;
    visited.add(moduleName);
    const document = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.TypeInitialization);
    document.importedModules.forEach(moduleName => {
        findNonFinalizedImportsRecursive(moduleName, parsingContext, queue, visited);
    });
    if (document.analysis.isTypeAnalysisPassNeeded) {
        if (!queue.includes(document)) {
            queue.push(document);
            if (NodeError_1.DEBUG_TYPES)
                console.log('  Scheduling: ' + document.moduleName);
        }
    }
}
function iterate(moduleName, parsingContext, _debug = false) {
    let analysisQueue = [];
    // Perform type analysis on the files in the analysis queue, which
    // is ordered in a way that should minimize the number of passes
    // we need to perform (with lower-level imports earlier in the list).
    findNonFinalizedImportsRecursive(moduleName, parsingContext, analysisQueue);
    if (analysisQueue.length === 0) {
        return;
    }
    if (NodeError_1.DEBUG_TYPES)
        console.log('Will analyze: ' + analysisQueue.map($ => $.moduleName).join(', '));
    const analyzers = new Map();
    while (true) {
        let documentToAnalyze = analysisQueue.shift();
        if (!documentToAnalyze) {
            break;
        }
        if (NodeError_1.DEBUG_TYPES) {
            console.log('Analyzing: ' + documentToAnalyze.moduleName + ' [' + analysisQueue.length + ' in queue]');
        }
        // Do a type analysis pass and determine if any internal changes occurred
        // during the pass. If so, continue to analyze until it stops changing and
        // mark all of its dependencies as needing to be reanalyzed.
        let didAnalysisChange = false;
        while (true) {
            const analyzer = new TypeAnalyzer_1.TypeAnalyzer(documentToAnalyze, parsingContext);
            analyzers.set(documentToAnalyze.moduleName, analyzer);
            if (!analyzer.analyze()) {
                if (NodeError_1.DEBUG_TYPES)
                    console.log('  Nothing changed.');
                if (analyzer.requiresTypeAnalysis()) {
                    documentToAnalyze.analysis.isTypeAnalysisPassNeeded = true;
                    if (NodeError_1.DEBUG_TYPES)
                        console.log('  Requires type analysis.');
                    const newQueue = [];
                    findNonFinalizedImportsRecursive(documentToAnalyze.moduleName, parsingContext, newQueue);
                    if (newQueue.length === 1 && newQueue[0] === documentToAnalyze) {
                        if (NodeError_1.DEBUG_TYPES)
                            console.log("  It won't converge");
                    }
                    else {
                        newQueue.forEach($ => analysisQueue.push($));
                    }
                }
                else {
                    if (documentToAnalyze.analysis.version >= TypeAnalyzer_1.MaxAnalysisPassCount) {
                        if (NodeError_1.DEBUG_TYPES)
                            console.log('  Analysis limit reached.');
                    }
                    documentToAnalyze.analysis.isTypeAnalysisPassNeeded = false;
                }
                break;
            }
            else {
                didAnalysisChange = true;
                if (NodeError_1.DEBUG_TYPES)
                    console.log('  Types has changed.');
            }
        }
        // We completed one or more updates to the file in this type
        // analysis pass, so we need to add its dependencies back
        // onto the queue if they're not already on it.
        if (didAnalysisChange) {
            for (let dependency of documentToAnalyze.importedBy) {
                if (!analysisQueue.some($ => $.moduleName === dependency)) {
                    const document = parsingContext.getPhase(dependency, nodes_1.PhaseFlags.NameInitialization);
                    if (document.analysis.isTypeAnalysisPassNeeded) {
                        if (NodeError_1.DEBUG_TYPES) {
                            console.log('  Scheduling: ' + document.moduleName);
                        }
                        analysisQueue.push(document);
                    }
                }
            }
        }
    }
    analyzers.forEach($ => {
        parsingContext.messageCollector.mergeWith($.messageCollector);
        $.nodeAnnotations.forEach((annotations, node) => {
            annotations.forEach($ => node.annotate($));
        });
    });
}
// Check and infer types
function executeTypeCheck(moduleName, parsingContext, debug = false) {
    const document = parsingContext.getPhase(moduleName, nodes_1.PhaseFlags.TypeCheck - 1);
    assert(document.analysis.nextPhase === nodes_1.PhaseFlags.TypeCheck);
    try {
        iterate(moduleName, parsingContext, debug);
    }
    catch (e) {
        if (e instanceof NodeError_1.AstNodeError) {
            parsingContext.messageCollector.error(e);
        }
        else {
            console.error(e);
            throw e;
        }
    }
    document.analysis.nextPhase++;
}
exports.executeTypeCheck = executeTypeCheck;
//# sourceMappingURL=typePhase.js.map