import { Nodes } from './nodes';
import { Scope } from './Scope';
export declare type Valtype = 'i32' | 'i64' | 'f32' | 'f64' | 'u32' | 'label';
export declare enum NativeTypes {
    i32 = "i32",
    i64 = "i64",
    f32 = "f32",
    f64 = "f64",
    funcref = "funcref",
    func = "func",
    void = "void"
}
export declare abstract class Type {
    abstract nativeType?: NativeTypes;
    get binaryenType(): Valtype | void;
    equals(otherType: Type): boolean;
    canBeAssignedTo(otherType: Type, ctx: Scope): boolean;
    abstract toString(): string;
    abstract inspect(levels: number): string;
    abstract schema(): Record<string, Type>;
    abstract getSchemaValue(name: string, ctx: Scope): any;
}
export declare function areEqualTypes(typeA: Type | null | void, typeB: Type | null | void): boolean;
export declare class FunctionSignatureType extends Type {
    readonly nativeType: NativeTypes;
    readonly parameterTypes: Type[];
    readonly parameterNames: string[];
    returnType?: Type;
    canBeAssignedTo(type: Type, ctx: Scope): boolean;
    equals(type: Type): boolean;
    toString(): string;
    inspect(_depth: number): string;
    schema(): {};
    getSchemaValue(name: string): void;
}
export declare class FunctionType extends Type {
    name: Nodes.NameIdentifierNode;
    readonly nativeType: NativeTypes;
    readonly signature: FunctionSignatureType;
    constructor(name: Nodes.NameIdentifierNode);
    equals(type: Type): boolean;
    canBeAssignedTo(type: Type, ctx: Scope): boolean;
    toString(): string;
    inspect(_depth: number): string;
    schema(): {};
    getSchemaValue(name: string): void;
}
export declare class StructType extends Type {
    parameters: Nodes.ParameterNode[];
    nativeType: NativeTypes;
    constructor(parameters: Nodes.ParameterNode[]);
    equals(type: Type): boolean;
    toString(): string;
    canBeAssignedTo(otherType: Type, ctx: Scope): boolean;
    typeSimilarity(to: Type, depth?: number): number;
    inspect(level: number): string;
    schema(): {
        byteSize: StackType;
    };
    getSchemaValue(name: string): number;
}
export declare class StackType extends Type {
    typeName: string;
    nativeType: NativeTypes;
    byteSize: number;
    private static nativeTypes;
    private constructor();
    static of(name: string, nativeType: NativeTypes, byteSize: number): StackType;
    equals(other: Type): boolean;
    canBeAssignedTo(other: Type, ctx: Scope): boolean;
    toString(): string;
    inspect(): string;
    schema(): {
        byteSize: StackType;
        allocationSize: StackType;
        stackSize: StackType;
    };
    getSchemaValue(name: string): number;
}
export declare class RefType extends Type {
    static instance: RefType;
    nativeType: NativeTypes;
    readonly byteSize = 8;
    protected constructor();
    static isRefTypeStrict(otherType: Type): boolean;
    toString(): string;
    inspect(): string;
    canBeAssignedTo(otherType: Type, _ctx: Scope): boolean;
    typeSimilarity(to: Type, depth?: number): number;
    equals(otherType: Type): boolean;
    schema(): {
        byteSize: StackType;
        stackSize: StackType;
        allocationSize: StackType;
    };
    getSchemaValue(name: string): number;
}
export declare class IntersectionType extends Type {
    of: Type[];
    nativeType: NativeTypes;
    constructor(of?: Type[]);
    toString(): string;
    canBeAssignedTo(other: Type, ctx: Scope): boolean;
    inspect(levels?: number): string;
    simplify(): Type;
    equals(other: Type): boolean;
    schema(): {};
    getSchemaValue(name: string): void;
}
export declare function getUnderlyingTypeFromAlias(type: Type): Type;
export declare class UnionType extends Type {
    of: Type[];
    readonly simplified: boolean;
    nativeType?: NativeTypes | undefined;
    get binaryenType(): Valtype;
    constructor(of?: Type[], simplified?: boolean);
    static of(x: Type[] | Type): UnionType;
    toString(): string;
    inspect(levels?: number): string;
    canBeAssignedTo(otherType: Type, ctx: Scope): boolean;
    equals(other: Type): boolean;
    /**
     * This method expands the union type made by other union types into a single
     * union made of the atoms of every member of the initial union. Recursively.
     */
    expand(ctx: Scope): Type;
    /**
     * This method removes an element from the union
     * @param type type to subtract
     */
    subtract(type: Type | null, ctx: Scope): Type;
    /**
     * This method simplifies the union type. e.g:
     *
     *   type T0 = A | A | A | B
     *   type T0Simplified = A | B
     *
     * It removes types present in unions
     *
     *   type T1 = T0Simplified | B | C | A
     *   type T1Simplified = T0Simplified | C
     *
     *
     */
    simplify(ctx: Scope): any;
    schema(): {
        byteSize: StackType;
        stackSize: StackType;
    };
    getSchemaValue(name: string, ctx: Scope): any;
}
export declare class TypeAlias extends Type {
    name: Nodes.NameIdentifierNode;
    readonly of: Type;
    get binaryenType(): void | Valtype;
    get nativeType(): NativeTypes;
    discriminant: number | null;
    traits: Map<Nodes.ImplDirective, TraitType>;
    constructor(name: Nodes.NameIdentifierNode, of: Type);
    canBeAssignedTo(other: Type, ctx: Scope): boolean;
    equals(other: Type): boolean;
    toString(): string;
    inspect(levels?: number): string;
    schema(): Record<string, Type>;
    getSchemaValue(name: string, ctx: Scope): any;
    getTypeMember(name: string, getType: (node: Nodes.Node) => Type | null): [Nodes.NameIdentifierNode, Type][];
    private getOrderedProperties;
}
export declare class TraitType extends Type {
    traitNode: Nodes.TraitDirectiveNode;
    get binaryenType(): undefined;
    get nativeType(): NativeTypes;
    constructor(traitNode: Nodes.TraitDirectiveNode);
    canBeAssignedTo(other: Type, _ctx: Scope): boolean;
    equals(other: Type): boolean;
    toString(): string;
    inspect(levels?: number): string;
    getSignatureOf(name: string): Type | null;
    schema(): Record<string, Type>;
    getSchemaValue(name: string): void;
}
export declare class TypeType extends Type {
    readonly of: Type;
    nativeType?: NativeTypes | undefined;
    static memMap: WeakMap<Type, TypeType>;
    private constructor();
    static of(of: Type): TypeType;
    canBeAssignedTo(other: Type, ctx: Scope): boolean;
    equals(other: Type): boolean;
    inspect(levels?: number): string;
    toString(): string;
    schema(): Record<string, Type>;
    getSchemaValue(name: string, ctx: Scope): any;
}
export declare class NeverType extends Type {
    static isNeverType(otherType: Type): boolean;
    get nativeType(): NativeTypes;
    toString(): string;
    inspect(): string;
    equals(other: Type): boolean;
    canBeAssignedTo(_: Type, _ctx: Scope): boolean;
    schema(): {};
    getSchemaValue(name: string): void;
}
export declare class SelfType extends Type {
    traitType: TraitType;
    get nativeType(): NativeTypes;
    constructor(traitType: TraitType);
    toString(): string;
    inspect(): string;
    canBeAssignedTo(_: Type, _ctx: Scope): boolean;
    schema(): {};
    getSchemaValue(name: string): void;
}
export declare class AnyType extends Type {
    get nativeType(): NativeTypes;
    toString(): string;
    inspect(): string;
    canBeAssignedTo(_: Type, _ctx: Scope): boolean;
    schema(): {};
    getSchemaValue(name: string): void;
}
export declare const InjectableTypes: any;
export declare const UNRESOLVED_TYPE: NeverType;
export declare namespace TypeHelpers {
    function getNodeType(node: Nodes.Node): Type | null;
    function setNodeType(node: Nodes.Node, type: Type | null): void;
}
