"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TypeHelpers = exports.UNRESOLVED_TYPE = exports.InjectableTypes = exports.AnyType = exports.SelfType = exports.NeverType = exports.TypeType = exports.TraitType = exports.TypeAlias = exports.UnionType = exports.getUnderlyingTypeFromAlias = exports.IntersectionType = exports.RefType = exports.StackType = exports.StructType = exports.FunctionType = exports.FunctionSignatureType = exports.areEqualTypes = exports.Type = exports.NativeTypes = void 0;
const typeHelpers_1 = require("./types/typeHelpers");
var NativeTypes;
(function (NativeTypes) {
    NativeTypes["i32"] = "i32";
    NativeTypes["i64"] = "i64";
    NativeTypes["f32"] = "f32";
    NativeTypes["f64"] = "f64";
    NativeTypes["funcref"] = "funcref";
    NativeTypes["func"] = "func";
    NativeTypes["void"] = "void";
})(NativeTypes = exports.NativeTypes || (exports.NativeTypes = {}));
class Type {
    get binaryenType() {
        switch (this.nativeType) {
            case NativeTypes.func:
            case NativeTypes.i32:
                return 'i32';
            case NativeTypes.f32:
                return 'f32';
            case NativeTypes.f64:
                return 'f64';
            case NativeTypes.i64:
                return 'i64';
            case NativeTypes.void:
                return undefined;
            default:
                throw new Error(`Type ${this} (${this.constructor.name}) returned a undefined binaryenType`);
        }
    }
    equals(otherType) {
        if (!otherType) {
            return false;
        }
        return otherType === this;
    }
    canBeAssignedTo(otherType, ctx) {
        if (this.equals(otherType)) {
            return true;
        }
        if (otherType instanceof TypeAlias) {
            return this.canBeAssignedTo(otherType.of, ctx);
        }
        if (otherType instanceof UnionType) {
            if (otherType.of.some($ => this.canBeAssignedTo($, ctx))) {
                return true;
            }
        }
        if (otherType instanceof IntersectionType) {
            return otherType.of.some($ => this.canBeAssignedTo($, ctx));
        }
        if (otherType instanceof AnyType) {
            return true;
        }
        if (otherType instanceof SelfType) {
            const resolved = ctx.get('Self');
            const type = TypeHelpers.getNodeType(resolved.referencedNode);
            if (type && type instanceof TypeType) {
                return this.canBeAssignedTo(type.of, ctx);
            }
        }
        return false;
    }
}
exports.Type = Type;
function areEqualTypes(typeA, typeB) {
    if (!typeA && !typeB) {
        return true;
    }
    if (typeA && typeB && typeA.equals(typeB)) {
        return true;
    }
    return false;
}
exports.areEqualTypes = areEqualTypes;
class FunctionSignatureType extends Type {
    constructor() {
        super(...arguments);
        this.nativeType = NativeTypes.func;
        this.parameterTypes = [];
        this.parameterNames = [];
    }
    canBeAssignedTo(type, ctx) {
        if (type instanceof FunctionSignatureType) {
            if (this.parameterTypes.length !== type.parameterTypes.length)
                return false;
            if (
            // can the return type be assigned?
            !this.returnType ||
                !type.returnType ||
                !this.returnType.canBeAssignedTo(type.returnType, ctx)) {
                return false;
            }
            if (
            // can every parameter be assigned?
            this.parameterTypes.some(($, $$) => !$ || !type.parameterTypes[$$] || !$.canBeAssignedTo(type.parameterTypes[$$], ctx))) {
                return false;
            }
            return true;
        }
        if (type instanceof IntersectionType) {
            return type.of.every($ => this.canBeAssignedTo($, ctx));
        }
        return super.canBeAssignedTo(type, ctx);
    }
    equals(type) {
        if (!type)
            return false;
        if (!(type instanceof FunctionSignatureType))
            return false;
        if (this.parameterTypes.length !== type.parameterTypes.length)
            return false;
        if (!areEqualTypes(this.returnType, type.returnType))
            return false;
        if (this.parameterTypes.some(($, $$) => !areEqualTypes($, type.parameterTypes[$$])))
            return false;
        return true;
    }
    toString() {
        const params = this.parameterTypes.map((type, ix) => {
            const name = this.parameterNames && this.parameterNames[ix];
            return (name ? name + ': ' : '') + (type || '?');
        });
        return `fun(${params.join(', ')}) -> ${this.returnType ? this.returnType : '?'}`;
    }
    inspect(_depth) {
        const params = this.parameterTypes.map(type => {
            if (!type) {
                return '(?)';
            }
            else {
                return type.inspect(0);
            }
        });
        return `(fun (${params.join(' ')}) ${(this.returnType && this.returnType.inspect(0)) || '?'})`;
    }
    schema() {
        return {};
    }
    getSchemaValue(name) {
        throw new Error(`Cannot read schema property ${name} of ${this.inspect(10)}`);
    }
}
exports.FunctionSignatureType = FunctionSignatureType;
class FunctionType extends Type {
    constructor(name) {
        super();
        this.name = name;
        this.nativeType = NativeTypes.func;
        this.signature = new FunctionSignatureType();
    }
    equals(type) {
        if (!type)
            return false;
        if (!(type instanceof FunctionType))
            return false;
        if (this.name !== type.name)
            return false;
        if (!this.signature.equals(type.signature))
            return false;
        return true;
    }
    canBeAssignedTo(type, ctx) {
        if (type instanceof FunctionType) {
            return this.signature.canBeAssignedTo(type.signature, ctx);
        }
        else if (type instanceof FunctionSignatureType) {
            return this.signature.canBeAssignedTo(type, ctx);
        }
        return false;
    }
    toString() {
        return this.signature.toString();
    }
    inspect(_depth) {
        const params = this.signature.parameterNames.map((_, ix) => {
            const type = this.signature.parameterTypes && this.signature.parameterTypes[ix];
            if (!type) {
                return '(?)';
            }
            else {
                return type.inspect(0);
            }
        });
        return `(fun ${JSON.stringify(this.name.name)} (${params.join(' ')}) ${(this.signature.returnType &&
            this.signature.returnType.inspect(0)) ||
            '?'})`;
    }
    schema() {
        return {};
    }
    getSchemaValue(name) {
        throw new Error(`Cannot read schema property ${name} of ${this.inspect(10)}`);
    }
}
exports.FunctionType = FunctionType;
class StructType extends Type {
    constructor(parameters) {
        super();
        this.parameters = parameters;
        this.nativeType = NativeTypes.i64;
    }
    equals(type) {
        return type === this;
    }
    toString() {
        return '%struct';
    }
    canBeAssignedTo(otherType, ctx) {
        if (super.canBeAssignedTo(otherType, ctx)) {
            return true;
        }
        if (RefType.isRefTypeStrict(otherType)) {
            return true;
        }
        return false;
    }
    typeSimilarity(to, depth = 1) {
        if (this.equals(to)) {
            return 1 / depth;
        }
        return 0;
    }
    inspect(level) {
        const types = level > 1
            ? this.parameters
                .map($ => {
                const type = $.parameterType && TypeHelpers.getNodeType($.parameterType);
                return ' ' + $.parameterName.name + ': ' + (type ? type.inspect(level - 1) : '(?)');
            })
                .join('')
            : '';
        return `(struct${types})`;
    }
    schema() {
        return {
            byteSize: StackType.of('u32', NativeTypes.i32, 4)
        };
    }
    getSchemaValue(name) {
        if (name === 'byteSize') {
            return 8;
        }
        throw new Error(`Cannot read schema property ${name} of ${this.inspect(10)}`);
    }
}
exports.StructType = StructType;
class StackType extends Type {
    constructor(typeName, nativeType, byteSize) {
        super();
        this.typeName = typeName;
        this.nativeType = nativeType;
        this.byteSize = byteSize;
    }
    static of(name, nativeType, byteSize) {
        const key = '_' + name + '_' + nativeType + '_' + byteSize;
        let ret = StackType.nativeTypes.get(key);
        if (!ret) {
            ret = new StackType(name, nativeType, byteSize);
            StackType.nativeTypes.set(key, ret);
        }
        return ret;
    }
    equals(other) {
        return other === this;
    }
    canBeAssignedTo(other, ctx) {
        const otherType = getUnderlyingTypeFromAlias(other);
        if (otherType instanceof StackType &&
            otherType.nativeType === this.nativeType &&
            otherType.typeName === this.typeName &&
            otherType.byteSize === this.byteSize) {
            return true;
        }
        return super.canBeAssignedTo(other, ctx);
    }
    toString() {
        return this.typeName;
    }
    inspect() {
        return '(native ' + this.typeName + ')';
    }
    schema() {
        return {
            byteSize: StackType.of('u32', NativeTypes.i32, 4),
            allocationSize: StackType.of('u32', NativeTypes.i32, 4),
            stackSize: StackType.of('u32', NativeTypes.i32, 4)
        };
    }
    getSchemaValue(name) {
        if (name === 'byteSize') {
            return this.byteSize;
        }
        else if (name === 'stackSize') {
            return this.byteSize;
        }
        else if (name === 'allocationSize') {
            return this.byteSize;
        }
        throw new Error(`Cannot read schema property ${name} of ${this.inspect()}`);
    }
}
exports.StackType = StackType;
StackType.nativeTypes = new Map();
const u32 = StackType.of('u32', NativeTypes.i32, 4);
const voidType = StackType.of('void', NativeTypes.void, 0);
class RefType extends Type {
    constructor() {
        super();
        this.nativeType = NativeTypes.i64;
        this.byteSize = 8;
    }
    // returns true when otherType is explicityly RefType.instance
    static isRefTypeStrict(otherType) {
        return getUnderlyingTypeFromAlias(otherType) === RefType.instance;
    }
    toString() {
        return 'ref';
    }
    inspect() {
        return '(ref ?)';
    }
    canBeAssignedTo(otherType, _ctx) {
        return RefType.isRefTypeStrict(otherType);
    }
    typeSimilarity(to, depth = 1) {
        if (this.equals(to)) {
            return 1 / depth;
        }
        return 0;
    }
    equals(otherType) {
        if (!otherType)
            return false;
        return RefType.isRefTypeStrict(otherType);
    }
    schema() {
        return {
            byteSize: StackType.of('u32', NativeTypes.i32, 4),
            stackSize: StackType.of('u32', NativeTypes.i32, 4),
            allocationSize: StackType.of('u32', NativeTypes.i32, 4)
        };
    }
    getSchemaValue(name) {
        if (name === 'byteSize') {
            return 8;
        }
        else if (name === 'stackSize') {
            return 8;
        }
        else if (name === 'allocationSize') {
            return 8;
        }
        throw new Error(`Cannot read schema property ${name} of ${this.inspect()}`);
    }
}
exports.RefType = RefType;
RefType.instance = new RefType();
class IntersectionType extends Type {
    constructor(of = []) {
        super();
        this.of = of;
        this.nativeType = NativeTypes.funcref;
    }
    toString() {
        if (this.of.length === 0)
            return '(empty intersection)';
        return this.of.map($ => ($ || '?').toString()).join(' & ');
    }
    canBeAssignedTo(other, ctx) {
        if (other instanceof IntersectionType) {
            // intersections can be assigned only if every type of
            // the other intersection can be assigned by some of the types of this intersection
            return other.of.every(otherType => this.of.some(thisType => thisType.canBeAssignedTo(otherType, ctx)));
        }
        if (this.equals(other)) {
            return true;
        }
        return this.of.some($ => $.canBeAssignedTo(other, ctx));
    }
    inspect(levels = 0) {
        return '(intersection ' + this.of.map($ => ($ ? $.inspect(levels - 1) : '(?)')).join(' ') + ')';
    }
    simplify() {
        const newTypes = [];
        function collectIntersection($) {
            if ($ instanceof IntersectionType) {
                $.of.forEach($ => {
                    collectIntersection($);
                });
            }
            else {
                if (!newTypes.some($1 => areEqualTypes($, $1))) {
                    newTypes.push($);
                }
            }
        }
        this.of.forEach(collectIntersection);
        if (newTypes.length === 1) {
            return newTypes[0];
        }
        else {
            return new IntersectionType(newTypes);
        }
    }
    equals(other) {
        if (!other)
            return false;
        return (other instanceof IntersectionType &&
            this.of.length === other.of.length &&
            this.of.every(($, ix) => areEqualTypes($, other.of[ix])));
    }
    schema() {
        return {};
    }
    getSchemaValue(name) {
        throw new Error(`Cannot read schema property ${name} of ${this.inspect(10)}`);
    }
}
exports.IntersectionType = IntersectionType;
function getUnderlyingTypeFromAlias(type) {
    if (type instanceof TypeAlias) {
        return getUnderlyingTypeFromAlias(type.of);
    }
    else {
        return type;
    }
}
exports.getUnderlyingTypeFromAlias = getUnderlyingTypeFromAlias;
class UnionType extends Type {
    constructor(of = [], simplified = false) {
        super();
        this.of = of;
        this.simplified = simplified;
    }
    get binaryenType() {
        const nativeTypes = new Set();
        this.of.forEach($ => {
            if (NeverType.isNeverType($))
                return;
            nativeTypes.add($.binaryenType);
        });
        if (nativeTypes.size === 0) {
            throw new Error('Cannot find a suitable low level type for ' + this.toString() + ' (0)');
        }
        if (nativeTypes.size === 1) {
            return nativeTypes.values().next().value;
        }
        else {
            throw new Error('Cannot find a suitable low level type for ' + this.toString());
        }
    }
    static of(x) {
        if (x instanceof UnionType) {
            return x;
        }
        else if (x instanceof Type) {
            return new UnionType([x], true);
        }
        else if (x instanceof Array) {
            return new UnionType(x);
        }
        throw new Error('Cannot create UnionType');
    }
    toString() {
        if (this.of.length === 0)
            return '(empty union)';
        return this.of.map($ => ($ || '?').toString()).join(' | ');
    }
    inspect(levels = 0) {
        if (this.of.length === 0)
            return `(union EMPTY)`;
        return '(union ' + this.of.map($ => ($ ? $.inspect(levels - 1) : '(?)')).join(' ') + ')';
    }
    canBeAssignedTo(otherType, ctx) {
        return this.of.every($ => $.canBeAssignedTo(otherType, ctx));
    }
    equals(other) {
        if (!other)
            return false;
        return (other instanceof UnionType &&
            this.of.length === other.of.length &&
            this.of.every(($, ix) => areEqualTypes($, other.of[ix])));
    }
    /**
     * This method expands the union type made by other union types into a single
     * union made of the atoms of every member of the initial union. Recursively.
     */
    expand(ctx) {
        const newSet = new Set();
        function add(typeToAdd) {
            const candidate = getUnderlyingTypeFromAlias(typeToAdd);
            if (candidate instanceof UnionType) {
                // If it is an union, we must expand it for each atom
                candidate.of.forEach($ => add($));
            }
            else {
                // finalize if already present
                if (newSet.has(typeToAdd)) {
                    return;
                }
                for (let $ of newSet) {
                    // finalize if we already have a type A == B && B == A
                    if ($.equals(typeToAdd) && typeToAdd.equals($)) {
                        return;
                    }
                }
                newSet.add(typeToAdd);
            }
        }
        this.of.forEach(add);
        const newTypes = Array.from(newSet.values());
        return new UnionType(newTypes).simplify(ctx);
    }
    /**
     * This method removes an element from the union
     * @param type type to subtract
     */
    subtract(type, ctx) {
        if (!type)
            return this;
        const removingRefType = RefType.isRefTypeStrict(type);
        if (!this.simplified) {
            return UnionType.of(this.expand(ctx)).subtract(type, ctx);
        }
        const newSet = new Set();
        for (let $ of this.of) {
            if ((!removingRefType && RefType.isRefTypeStrict($)) || !$.canBeAssignedTo(type, ctx)) {
                newSet.add($);
            }
        }
        if (newSet.size === 0) {
            return exports.InjectableTypes.never;
        }
        const newTypes = Array.from(newSet.values());
        if (newTypes.length === 1) {
            return newTypes[0];
        }
        return new UnionType(newTypes);
    }
    /**
     * This method simplifies the union type. e.g:
     *
     *   type T0 = A | A | A | B
     *   type T0Simplified = A | B
     *
     * It removes types present in unions
     *
     *   type T1 = T0Simplified | B | C | A
     *   type T1Simplified = T0Simplified | C
     *
     *
     */
    simplify(ctx) {
        if (this.of.length === 1) {
            return this.of[0];
        }
        let newTypes = [];
        this.of.forEach($ => {
            if (NeverType.isNeverType($))
                return;
            if (!newTypes.some($1 => areEqualTypes($1, $)) && $) {
                newTypes.push($);
            }
        });
        if (newTypes.length === 0) {
            return exports.InjectableTypes.never;
        }
        let unions = [];
        function collectUnion($) {
            const candidate = getUnderlyingTypeFromAlias($);
            if (candidate instanceof UnionType) {
                if (!unions.includes(candidate)) {
                    unions.push(candidate);
                    candidate.of.forEach($ => {
                        collectUnion($);
                    });
                }
            }
        }
        // Collect unions
        newTypes.forEach(collectUnion);
        // This are the unions that generate some conflict with other unions,
        // therefore we need to expand those unions to the atoms
        const blackListedUnionAtoms = [];
        const unionsToRemove = [];
        // Find the conflictive atoms
        unions.forEach((union, ix) => {
            const expanded = UnionType.of(union.expand(ctx));
            const hasConflict = expanded.of.some(atom => unions.some(($, $$) => $$ !== ix && atom.canBeAssignedTo($, ctx)));
            if (hasConflict) {
                blackListedUnionAtoms.push(...expanded.of);
                // we are removing the union, it might have elements not present in the
                // newTypes, we add them
                expanded.of.forEach($ => {
                    if (!newTypes.some($1 => $1.equals($))) {
                        newTypes.push($);
                    }
                });
                unionsToRemove.push(union);
            }
        });
        unions = unions.filter($ => !unionsToRemove.includes($));
        // Eliminate types present in unions
        newTypes.forEach((newType, i) => {
            const candidate = getUnderlyingTypeFromAlias(newType);
            if (unionsToRemove.includes(candidate)) {
                newTypes[i] = null;
                return;
            }
            if (RefType.isRefTypeStrict(candidate) || blackListedUnionAtoms.some(x => candidate.canBeAssignedTo(x, ctx))) {
                return;
            }
            if (unions.some(union => !union.equals(candidate) && candidate.canBeAssignedTo(union, ctx))) {
                newTypes[i] = null;
            }
        });
        // Remove eliminated types
        newTypes = newTypes.filter($ => !!$);
        if (newTypes.length === 1) {
            return newTypes[0];
        }
        else {
            return new UnionType(newTypes, true);
        }
    }
    schema() {
        return {
            byteSize: u32,
            stackSize: u32
        };
    }
    getSchemaValue(name, ctx) {
        if (name === 'byteSize') {
            const nativeTypes = new Set();
            this.of.forEach($ => {
                if (NeverType.isNeverType($))
                    return;
                nativeTypes.add($.getSchemaValue('byteSize', ctx));
            });
            if (nativeTypes.size === 0) {
                throw new Error('Cannot find a consistent byteSize in ' + this.inspect(100) + ' (0)');
            }
            if (nativeTypes.size === 1) {
                return nativeTypes.values().next().value;
            }
            else {
                throw new Error('Cannot find a consistent byteSize in ' + this.inspect(100));
            }
        }
        else if (name === 'allocationSize') {
            return 8;
        }
        else if (name === 'stackSize') {
            return 8;
        }
        throw new Error(`Cannot read schema property ${name} of ${this.inspect(10)}`);
    }
}
exports.UnionType = UnionType;
function mapAinB(a, b) {
    for (let [key, value] of a) {
        if (!b.has(key))
            return false;
        if (b.get(key) !== value)
            return false;
    }
    return true;
}
function areMapsEqual(a, b) {
    if (a.size !== b.size)
        return false;
    if (!mapAinB(a, b))
        return false;
    if (!mapAinB(b, a))
        return false;
    return true;
}
class TypeAlias extends Type {
    constructor(name, of) {
        super();
        this.name = name;
        this.of = of;
        this.discriminant = null;
        this.traits = new Map();
    }
    get binaryenType() {
        return this.of.binaryenType;
    }
    get nativeType() {
        return this.of.nativeType;
    }
    canBeAssignedTo(other, ctx) {
        // | if (other instanceof TraitType) {
        // |   for (let [, trait] of this.traits) {
        // |     if (trait.equals(other)) {
        // |       return true;
        // |     }
        // |   }
        // | }
        if (other instanceof SelfType) {
            const resolved = ctx.get('Self');
            const type = TypeHelpers.getNodeType(resolved.referencedNode);
            if (type && type instanceof TypeType) {
                return this.canBeAssignedTo(type.of, ctx);
            }
        }
        return this.of.canBeAssignedTo(other, ctx);
    }
    equals(other) {
        if (!(other instanceof TypeAlias))
            return false;
        if (other.name !== this.name)
            return false;
        if (other.discriminant !== this.discriminant)
            return false;
        if (!areMapsEqual(this.traits, other.traits)) {
            return false;
        }
        if (!areEqualTypes(this.of, other.of))
            return false;
        return true;
    }
    toString() {
        return this.name.name;
    }
    inspect(levels = 0) {
        const ofString = levels > 0 ? ' ' + (this.of ? this.of.inspect(levels - 1) : '(?)') : '';
        return `(alias ${this.name.name}${ofString})`;
    }
    schema() {
        const result = {
            ...this.of.schema(),
            discriminant: u32
        };
        const baseType = getUnderlyingTypeFromAlias(this);
        if (baseType instanceof StructType) {
            result['allocationSize'] = u32;
            result['stackSize'] = u32;
            const properties = this.getOrderedProperties();
            properties.forEach(prop => {
                result[`property$${prop.index}_offset`] = u32;
                result[`property$${prop.index}_allocationSize`] = u32;
            });
        }
        return result;
    }
    getSchemaValue(name, ctx) {
        if (name === 'discriminant') {
            if (this.discriminant === null) {
                throw new Error('empty discriminant');
            }
            return this.discriminant;
        }
        else {
            const baseType = getUnderlyingTypeFromAlias(this);
            if (baseType instanceof StructType) {
                if (name === 'stackSize') {
                    return 8;
                }
                else if (name === 'allocationSize') {
                    const properties = this.getOrderedProperties();
                    let offset = 0;
                    for (let prop of properties) {
                        const fn = getNonVoidFunction(TypeHelpers.getNodeType(prop.name), ctx);
                        offset += fn.signature.returnType.getSchemaValue('byteSize', ctx);
                    }
                    return offset;
                }
                else if (name.startsWith('property$')) {
                    const properties = this.getOrderedProperties();
                    const index = parseInt(name.substr('property$'.length), 10);
                    const property = properties.find($ => $.index === index);
                    if (!property) {
                        throw new Error('cannot find property index ' + index);
                    }
                    if (name.endsWith('_offset')) {
                        let offset = 0;
                        for (let prop of properties) {
                            if (prop.index === index) {
                                break;
                            }
                            const fn = getNonVoidFunction(TypeHelpers.getNodeType(prop.name), ctx);
                            offset += fn.signature.returnType.getSchemaValue('stackSize', ctx);
                        }
                        return offset;
                    }
                    else if (name.endsWith('_allocationSize')) {
                        const fn = getNonVoidFunction(TypeHelpers.getNodeType(property.name), ctx);
                        return fn.signature.returnType.getSchemaValue('allocationSize', ctx);
                    }
                }
            }
        }
        return this.of.getSchemaValue(name, ctx);
    }
    getTypeMember(name, getType) {
        const ret = [];
        for (let impl of this.name.impls) {
            const resolvedName = impl.namespaceNames.get(name);
            if (resolvedName) {
                const memberType = getType(resolvedName);
                if (!(0, typeHelpers_1.isValidType)(memberType)) {
                    return [];
                }
                ret.push([resolvedName, memberType]);
            }
        }
        if (this.of instanceof TypeAlias) {
            ret.push(...this.of.getTypeMember(name, getType));
        }
        return ret;
    }
    getOrderedProperties() {
        const properties = [];
        // TODO: review this.
        for (let impl of this.name.impls) {
            for (let [name, nameIdentifierNode] of impl.namespaceNames) {
                if (name.startsWith('property$')) {
                    const index = parseInt(name.substr('property$'.length), 10);
                    properties.push({ index, name: nameIdentifierNode });
                }
            }
        }
        properties.sort((a, b) => {
            if (a.index > b.index) {
                return 1;
            }
            else {
                return -1;
            }
        });
        return properties;
    }
}
exports.TypeAlias = TypeAlias;
class TraitType extends Type {
    constructor(traitNode) {
        super();
        this.traitNode = traitNode;
    }
    get binaryenType() {
        return undefined;
    }
    get nativeType() {
        return NativeTypes.void;
    }
    canBeAssignedTo(other, _ctx) {
        return this === other;
    }
    equals(other) {
        return other === this;
    }
    toString() {
        return this.traitNode.traitName.name;
    }
    inspect(levels = 0) {
        const subtypes = [];
        if (levels > 1) {
            this.traitNode.namespaceNames.forEach((value, key) => {
                const t = TypeHelpers.getNodeType(value);
                subtypes.push('[' + key + ': ' + (t ? t.inspect(levels - 1) : 'null') + ']');
            });
        }
        return `(trait ${this.traitNode.traitName.name}${subtypes.map($ => ' ' + $).join('')})`;
    }
    getSignatureOf(name) {
        const declName = this.traitNode.namespaceNames.get(name);
        if (!declName)
            return null;
        return TypeHelpers.getNodeType(declName);
    }
    schema() {
        const result = {};
        return result;
    }
    getSchemaValue(name) {
        throw new Error(`Cannot read schema property ${name} of ${this.inspect()}`);
    }
}
exports.TraitType = TraitType;
function getNonVoidFunction(type, ctx) {
    const functions = type.of;
    for (let fn of functions) {
        if (fn.signature.returnType && !voidType.canBeAssignedTo(fn.signature.returnType, ctx)) {
            return fn;
        }
    }
    return null;
}
class TypeType extends Type {
    constructor(of) {
        super();
        this.of = of;
    }
    static of(of) {
        let ret = this.memMap.get(of);
        if (!ret) {
            ret = new TypeType(of);
            this.memMap.set(of, ret);
        }
        return ret;
    }
    canBeAssignedTo(other, ctx) {
        const otherType = getUnderlyingTypeFromAlias(other);
        if (otherType instanceof TypeType) {
            return this.of.canBeAssignedTo(otherType.of, ctx);
        }
        return false;
    }
    equals(other) {
        if (!other)
            return false;
        return other instanceof TypeType && areEqualTypes(other.of, this.of);
    }
    inspect(levels = 0) {
        return `(type ${this.of.inspect(levels - 1)})`;
    }
    toString() {
        return `Type<${this.of}>`;
    }
    schema() {
        return this.of.schema();
    }
    getSchemaValue(name, ctx) {
        return this.of.getSchemaValue(name, ctx);
    }
}
exports.TypeType = TypeType;
TypeType.memMap = new WeakMap();
// https://en.wikipedia.org/wiki/Bottom_type
// https://en.wikipedia.org/wiki/Fail-stop
class NeverType extends Type {
    static isNeverType(otherType) {
        return getUnderlyingTypeFromAlias(otherType) instanceof NeverType;
    }
    get nativeType() {
        return NativeTypes.void;
    }
    toString() {
        return 'never';
    }
    inspect() {
        return '(never)';
    }
    equals(other) {
        if (NeverType.isNeverType(other)) {
            return true;
        }
        if (other instanceof UnionType) {
            if (other.of.length === 0) {
                return true;
            }
            if (other.of.length === 1 && this.equals(other.of[0])) {
                return true;
            }
        }
        return super.equals(other);
    }
    canBeAssignedTo(_, _ctx) {
        return true;
    }
    schema() {
        return {};
    }
    getSchemaValue(name) {
        throw new Error(`Cannot read schema property ${name} of ${this.inspect()}`);
    }
}
exports.NeverType = NeverType;
class SelfType extends Type {
    constructor(traitType) {
        super();
        this.traitType = traitType;
    }
    get nativeType() {
        return NativeTypes.funcref;
    }
    toString() {
        return 'Self';
    }
    inspect() {
        return `(self ${this.traitType.inspect(0)})`;
    }
    canBeAssignedTo(_, _ctx) {
        console.log(_ctx.inspect(true), _.inspect(1));
        return false;
    }
    schema() {
        return {};
    }
    getSchemaValue(name) {
        throw new Error(`Cannot read schema property ${name} of ${this.inspect()}`);
    }
}
exports.SelfType = SelfType;
// https://en.wikipedia.org/wiki/Top_type
class AnyType extends Type {
    get nativeType() {
        return NativeTypes.funcref;
    }
    toString() {
        return 'any';
    }
    inspect() {
        return '(any)';
    }
    canBeAssignedTo(_, _ctx) {
        return false;
    }
    schema() {
        return {};
    }
    getSchemaValue(name) {
        throw new Error(`Cannot read schema property ${name} of ${this.inspect()}`);
    }
}
exports.AnyType = AnyType;
exports.InjectableTypes = Object.assign(Object.create(null), {
    void: voidType,
    ref: RefType.instance,
    never: new NeverType(),
    AnyType: TypeType.of(new AnyType()),
    Any: new AnyType()
});
exports.UNRESOLVED_TYPE = new NeverType();
var TypeHelpers;
(function (TypeHelpers) {
    const ofTypeSymbol = Symbol('ofType');
    function getNodeType(node) {
        if (!node)
            return null;
        return node[ofTypeSymbol] || null;
    }
    TypeHelpers.getNodeType = getNodeType;
    function setNodeType(node, type) {
        node[ofTypeSymbol] = type;
        node.isTypeResolved = (0, typeHelpers_1.isValidType)(type);
    }
    TypeHelpers.setNodeType = setNodeType;
})(TypeHelpers = exports.TypeHelpers || (exports.TypeHelpers = {}));
//# sourceMappingURL=types.js.map