"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TypeAnalyzer = exports.TypeResolver = exports.MaxAnalysisPassCount = void 0;
const walker_1 = require("../walker");
const nodes_1 = require("../nodes");
const types_1 = require("../types");
const typeHelpers_1 = require("./typeHelpers");
const annotations_1 = require("../annotations");
const NodeError_1 = require("../NodeError");
const helpers_1 = require("../helpers");
const MessageCollector_1 = require("../MessageCollector");
const astPrinter_1 = require("../../utils/astPrinter");
const helpers_2 = require("../phases/helpers");
// At some point, we'll cut off the analysis passes and assume
// we're making no forward progress. This should happen only
// on the case of bugs in the analyzer.
exports.MaxAnalysisPassCount = 100;
function createTypeAlias(name, value, parsingContext) {
    const document = (0, helpers_2.getDocument)(name);
    if (!document)
        throw new NodeError_1.LysTypeError('Cannot find document', name);
    const discriminant = parsingContext.getTypeDiscriminant(document.moduleName, name.name);
    const alias = new types_1.TypeAlias(name, value);
    alias.discriminant = discriminant;
    return alias;
}
class TypeResolver {
    constructor(documentNode, parsingContext, messageCollector = new MessageCollector_1.MessageCollector()) {
        this.documentNode = documentNode;
        this.parsingContext = parsingContext;
        this.messageCollector = messageCollector;
        this.didAnalysisChange = false;
        this.nodeAnnotations = new Map();
        this.nonResolvedNodes = new Set();
        this.walk = (0, walker_1.walkPreOrder)(this.onEnterNode.bind(this), this.onLeaveNode.bind(this));
    }
    analyze() {
        this.didAnalysisChange = false;
        this.messageCollector.errors.length = 0;
        this.nonResolvedNodes.clear();
        this.nodeAnnotations.clear();
        this.walk(this.documentNode, this.parsingContext);
        this.documentNode.analysis.version++;
        // If we've already analyzed the file the max number of times,
        // just give up and admit defeat. This should happen only in
        // the case of analyzer bugs.
        if (this.documentNode.analysis.version >= exports.MaxAnalysisPassCount) {
            this.messageCollector.error(`Hit max analysis pass count for ${this.documentNode.moduleName}`, this.documentNode.astNode);
            return false;
        }
        return this.didAnalysisChange;
    }
    requiresTypeAnalysis() {
        if (this.documentNode.analysis.version >= exports.MaxAnalysisPassCount) {
            return false;
        }
        if (this.didAnalysisChange) {
            return true;
        }
        let allTypesResolved = true;
        (0, walker_1.walk)(this.documentNode, this.parsingContext, node => {
            if (node instanceof nodes_1.Nodes.ExpressionNode) {
                if (!node.isTypeResolved) {
                    allTypesResolved = false;
                    this.nonResolvedNodes.add(node);
                    return false;
                }
            }
            if (node instanceof nodes_1.Nodes.FunctionNode) {
                if (!(0, typeHelpers_1.isValidType)(this.getType(node.functionName))) {
                    allTypesResolved = false;
                    this.nonResolvedNodes.add(node);
                    return false;
                }
            }
            if (node instanceof nodes_1.Nodes.TypeNode) {
                if (!(0, typeHelpers_1.isValidType)(this.getType(node))) {
                    allTypesResolved = false;
                    this.nonResolvedNodes.add(node);
                    return false;
                }
            }
            if (node instanceof nodes_1.Nodes.TypeDirectiveNode) {
                if (!(0, typeHelpers_1.isValidType)(this.getType(node.variableName))) {
                    allTypesResolved = false;
                    this.nonResolvedNodes.add(node);
                    return false;
                }
            }
            if (node instanceof nodes_1.Nodes.OverloadedFunctionNode) {
                if (!(0, typeHelpers_1.isValidType)(this.getType(node.functionName))) {
                    allTypesResolved = false;
                    this.nonResolvedNodes.add(node);
                    return false;
                }
            }
        });
        if (NodeError_1.DEBUG_TYPES) {
            this.nonResolvedNodes.forEach($ => console.log((0, astPrinter_1.printAST)($)));
        }
        this.documentNode.analysis.areTypesResolved = allTypesResolved;
        return !allTypesResolved;
    }
    setType(node, type) {
        const currentType = types_1.TypeHelpers.getNodeType(node);
        const areEqual = (currentType && type && (0, types_1.areEqualTypes)(type, currentType)) || false;
        types_1.TypeHelpers.setNodeType(node, type);
        if (type === types_1.UNRESOLVED_TYPE) {
            node.isTypeResolved = false;
        }
        if (!areEqual && type) {
            this.setAnalysisChanged();
        }
    }
    removeStashedNodeAnnotation(node, annotation) {
        const annotations = this.nodeAnnotations.get(node);
        if (!annotations)
            return false;
        return annotations.delete(annotation);
    }
    getStashedNodeAnnotation(node, klass) {
        const annotations = this.nodeAnnotations.get(node);
        if (!annotations)
            return null;
        for (let annotation of annotations) {
            if (annotation instanceof klass)
                return annotation;
        }
        return null;
    }
    stashNodeAnnotation(node, annotation) {
        const annotations = this.nodeAnnotations.get(node) || new Set();
        if (!this.nodeAnnotations.has(node)) {
            this.nodeAnnotations.set(node, annotations);
        }
        annotations.add(annotation);
    }
    getType(node) {
        const type = types_1.TypeHelpers.getNodeType(node);
        return type;
    }
    setAnalysisChanged() {
        this.didAnalysisChange = true;
    }
}
exports.TypeResolver = TypeResolver;
class TypeAnalyzer extends TypeResolver {
    onLeaveNode(node, _parsingContext, parent) {
        if (!node.scope) {
            this.messageCollector.errorIfBranchDoesntHaveAny(`Node ${node.nodeName} has no scope`, node);
            return;
        }
        const nodeScope = node.scope;
        if (node instanceof nodes_1.Nodes.MatcherNode) {
            console.assert(parent instanceof nodes_1.Nodes.PatternMatcherNode);
            this.processMatcherNode(node, parent);
        }
        else if (node instanceof nodes_1.Nodes.LiteralNode) {
            if (node.resolvedReference) {
                const type = this.getType(node.resolvedReference.referencedNode);
                this.setType(node, this.getTypeTypeType(node, type, this.messageCollector));
            }
        }
        else if (node instanceof nodes_1.Nodes.ReferenceNode) {
            if (node.resolvedReference) {
                const type = this.getType(node.resolvedReference.referencedNode);
                if ((0, typeHelpers_1.isValidType)(type)) {
                    if (type instanceof types_1.TypeType && node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                        if (type.of instanceof types_1.TypeAlias) {
                            const fnType = this.resolveTypeMember(node, type.of, 'apply', this.messageCollector, nodeScope);
                            // TODO: a better error would be X is not callable
                            if (fnType) {
                                const fun = this.findFunctionOverload(fnType, [], node, null, false, this.messageCollector, true, nodeScope);
                                if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                                    this.annotateImplicitCall(node, fun, [], this.messageCollector, nodeScope);
                                    if (fun.signature.returnType) {
                                        this.setType(node, fun.signature.returnType);
                                        return;
                                    }
                                }
                                this.messageCollector.errorIfBranchDoesntHaveAny(type.inspect(100), node);
                            }
                        }
                        this.messageCollector.error(new NodeError_1.UnexpectedType(type, node));
                    }
                    else if (type instanceof types_1.FunctionSignatureType && node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                        const currentAnnotation = this.getStashedNodeAnnotation(node, annotations_1.annotations.IsFunctionReference);
                        if (currentAnnotation) {
                            currentAnnotation.fun = type;
                        }
                        else {
                            this.stashNodeAnnotation(node, new annotations_1.annotations.IsFunctionReference(type));
                        }
                    }
                    this.setType(node, type);
                }
                else {
                    this.messageCollector.errorIfBranchDoesntHaveAny(`Unable to resolve name "${node.resolvedReference.toString()}"${(type &&
                        ' ' + type.inspect(100)) ||
                        (0, astPrinter_1.printAST)(node.resolvedReference.referencedNode)}`, node);
                    if (node.resolvedReference.moduleName) {
                        if (!this.documentNode.importedModules.has(node.resolvedReference.moduleName)) {
                            this.messageCollector.errorIfBranchDoesntHaveAny(`Document is not importing "${node.resolvedReference.moduleName}"`, node);
                        }
                    }
                    this.setType(node, types_1.UNRESOLVED_TYPE);
                }
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny('Unable to resolve name', node);
            }
        }
        else if (node instanceof nodes_1.Nodes.IsExpressionNode) {
            const booleanType = node.booleanReference
                ? this.getTypeTypeType(node, this.getType(node.booleanReference.referencedNode), this.messageCollector)
                : null;
            const lhsType = this.getType(node.lhs);
            const rhsType = this.getTypeTypeType(node.rhs, this.getType(node.rhs), this.messageCollector);
            if ((0, typeHelpers_1.isValidType)(lhsType)) {
                if (!this.canBeAssigned(lhsType, types_1.RefType.instance, nodeScope)) {
                    this.messageCollector.error(`"is" expression can only be used with reference types, used with: "${lhsType}"`, node.astNode);
                    this.setType(node, booleanType);
                    return;
                }
                if ((0, typeHelpers_1.isValidType)(rhsType)) {
                    const valueType = this.resolveTypeMember(node.rhs, rhsType, 'is', this.messageCollector, nodeScope);
                    if (valueType) {
                        const fun = this.findFunctionOverload(valueType, [rhsType], node, null, false, this.messageCollector, false, nodeScope);
                        if (fun instanceof types_1.FunctionType) {
                            node.resolvedFunctionType = fun;
                            this.setType(node, fun.signature.returnType);
                            return;
                        }
                        else {
                            this.messageCollector.error(`This statement is always false, type "${lhsType}" can never be "${rhsType}". "${rhsType}" has no overload for \`fun is(arg: ${lhsType}): boolean\`.`, node.astNode);
                            this.setType(node, booleanType);
                        }
                    }
                    else {
                        this.messageCollector.error(`This statement is always false, type "${lhsType}" can never be "${rhsType}". "${rhsType}" has no "is" function implementation.`, node.astNode);
                        this.setType(node, booleanType);
                    }
                }
                else {
                    this.messageCollector.errorIfBranchDoesntHaveAny(`Error with "is"`, node.rhs);
                    this.setType(node, booleanType);
                }
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny(`Error with "is"`, node.lhs);
                this.setType(node, booleanType);
            }
        }
        else if (node instanceof nodes_1.Nodes.AsExpressionNode) {
            const lhsType = this.getType(node.lhs);
            if ((0, typeHelpers_1.isValidType)(lhsType)) {
                const rhsType = this.getTypeTypeType(node.rhs, this.getType(node.rhs), this.messageCollector);
                if ((0, typeHelpers_1.isValidType)(rhsType)) {
                    if (lhsType.equals(rhsType) && rhsType.equals(lhsType)) {
                        this.messageCollector.warning(`This cast is useless "${lhsType}" as "${rhsType}"`, node);
                        if (!node.hasAnnotation(annotations_1.annotations.ByPassFunction)) {
                            this.stashNodeAnnotation(node, new annotations_1.annotations.ByPassFunction());
                        }
                        this.setType(node, rhsType);
                        return;
                    }
                    const memberType = this.resolveTypeMember(node.lhs, lhsType, 'as', this.messageCollector, nodeScope);
                    if (memberType) {
                        const fun = this.findFunctionOverload(memberType, [lhsType], node, rhsType, false, new MessageCollector_1.MessageCollector(), false, nodeScope);
                        if (fun instanceof types_1.FunctionType && (0, typeHelpers_1.isValidType)(fun.signature.returnType)) {
                            this.annotateImplicitCall(node, fun, [node.lhs], this.messageCollector, nodeScope);
                            this.setType(node, fun.signature.returnType);
                        }
                        else {
                            this.messageCollector.errorIfBranchDoesntHaveAny(new NodeError_1.LysTypeError(`Cannot convert type "${lhsType}" into "${rhsType}"`, node.lhs));
                            this.setType(node, types_1.UNRESOLVED_TYPE);
                        }
                    }
                }
                else {
                    this.messageCollector.errorIfBranchDoesntHaveAny('Cannot resolve type 1', node.rhs);
                }
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny('Cannot resolve type 2', node.lhs);
            }
        }
        else if (node instanceof nodes_1.Nodes.BinaryExpressionNode) {
            const lhsType = this.getType(node.lhs);
            const rhsType = this.getType(node.rhs);
            if ((0, typeHelpers_1.isValidType)(lhsType)) {
                const memberName = node.operator.name;
                const memberType = this.resolveTypeMember(node, lhsType, memberName, this.messageCollector, nodeScope);
                if (memberType) {
                    if ((0, typeHelpers_1.isValidType)(rhsType)) {
                        const argumentTypes = [lhsType, rhsType];
                        const fun = this.findFunctionOverload(memberType, argumentTypes, node, null, false, this.messageCollector, true, nodeScope);
                        if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                            this.annotateImplicitCall(node, fun, [node.lhs, node.rhs], this.messageCollector, nodeScope);
                            this.setType(node, fun.signature.returnType);
                        }
                        else {
                            this.setType(node, types_1.UNRESOLVED_TYPE);
                        }
                    }
                    else {
                        this.messageCollector.errorIfBranchDoesntHaveAny('Cannot resolve type 3', node.rhs);
                    }
                }
                else {
                    this.setType(node, types_1.UNRESOLVED_TYPE);
                }
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny('Cannot resolve type 4 ' + (0, astPrinter_1.printAST)(node.lhs), node.lhs);
                this.setType(node, types_1.UNRESOLVED_TYPE);
            }
        }
        else if (node instanceof nodes_1.Nodes.UnaryExpressionNode) {
            const rhsType = this.getType(node.rhs);
            if ((0, typeHelpers_1.isValidType)(rhsType)) {
                const memberName = node.operator.name;
                const memberType = this.resolveTypeMember(node, rhsType, memberName, this.messageCollector, nodeScope);
                if (memberType) {
                    const argumentTypes = [rhsType];
                    const fun = this.findFunctionOverload(memberType, argumentTypes, node, null, false, new MessageCollector_1.MessageCollector(), true, nodeScope);
                    if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                        this.annotateImplicitCall(node, fun, node.argumentsNode, this.messageCollector, nodeScope);
                        this.setType(node, fun.signature.returnType);
                    }
                    else {
                        this.messageCollector.errorIfBranchDoesntHaveAny(`Cannot apply operator "${memberName}" in type "${rhsType}"`, node);
                        this.setType(node, rhsType);
                    }
                }
            }
        }
        else if (node instanceof nodes_1.Nodes.IfNode) {
            const condition = this.getType(node.condition);
            const booleanType = node.booleanReference
                ? this.getTypeTypeType(node.condition, this.getType(node.booleanReference.referencedNode), this.messageCollector)
                : null;
            const truePart = this.getType(node.truePart);
            if (condition) {
                if (booleanType) {
                    this.ensureCanBeAssignedWithImplicitConversion(condition, booleanType, node.condition, this.messageCollector, nodeScope);
                }
                else {
                    this.messageCollector.error('Cannot resolve "boolean"', node.condition.astNode);
                }
            }
            if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                if (!node.falsePart) {
                    this.messageCollector.error('A ternary operation requires an else branch', node.astNode);
                }
                if ((0, typeHelpers_1.isValidType)(truePart)) {
                    const falsePart = node.falsePart ? this.getType(node.falsePart) : null;
                    if ((0, typeHelpers_1.isValidType)(falsePart)) {
                        this.setType(node, new types_1.UnionType([truePart, falsePart]).simplify(nodeScope));
                    }
                    else {
                        this.messageCollector.error('Else not resolved', node.astNode);
                        this.setType(node, new types_1.UnionType([truePart, types_1.UNRESOLVED_TYPE]).simplify(nodeScope));
                    }
                    return;
                }
            }
            else {
                this.setType(node, types_1.InjectableTypes.void);
            }
        }
        else if (node instanceof nodes_1.Nodes.AssignmentNode) {
            const lhsType = this.getType(node.lhs);
            const rhsType = this.getType(node.rhs);
            if (!(0, typeHelpers_1.isValidType)(lhsType)) {
                this.messageCollector.error('Cannot resolve type', node.lhs.astNode);
                return;
            }
            if (!(0, typeHelpers_1.isValidType)(rhsType)) {
                this.messageCollector.error('Cannot resolve type', node.rhs.astNode);
                return;
            }
            // TODO: unhack my heart
            if (node.lhs instanceof nodes_1.Nodes.MemberNode) {
                const call = this.getStashedNodeAnnotation(node.lhs, annotations_1.annotations.ImplicitCall);
                if (call) {
                    this.messageCollector.error('!!!1 This node already has an ' + call.toString(), node.lhs.astNode);
                    return;
                }
                const prop = this.describeProperty(lhsType);
                if (prop) {
                    if (prop.setter) {
                        const memberLhsType = this.getType(node.lhs.lhs);
                        const fun = this.findFunctionOverload(lhsType, [memberLhsType, rhsType], node.lhs, null, false, this.messageCollector, true, nodeScope);
                        if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                            if (!fun.name.hasAnnotation(annotations_1.annotations.Setter)) {
                                this.messageCollector.error(new NodeError_1.LysTypeError(`Cannot assign to "${node.lhs.memberName.name}" because it is a read-only property.`, node.lhs.memberName));
                                this.setType(node, types_1.UNRESOLVED_TYPE);
                            }
                            else {
                                this.annotateImplicitCall(node, fun, [node.lhs.lhs, node.rhs], this.messageCollector, nodeScope);
                                this.setType(node, fun.signature.returnType);
                            }
                        }
                        else {
                            this.messageCollector.error('Overload not found', node.rhs.astNode);
                            this.setType(node, types_1.UNRESOLVED_TYPE);
                            return;
                        }
                    }
                    else {
                        this.messageCollector.error(new NodeError_1.LysTypeError(`Cannot assign to "${node.lhs.memberName.name}" because it is a read-only property.`, node.lhs.memberName));
                        this.setType(node, types_1.UNRESOLVED_TYPE);
                    }
                }
                else {
                    if (!this.messageCollector.hasErrorForBranch(node.lhs.astNode)) {
                        this.messageCollector.error('Invalid property !', node.lhs.astNode);
                    }
                    this.setType(node, types_1.UNRESOLVED_TYPE);
                    return;
                }
            }
            else if (node.lhs instanceof nodes_1.Nodes.BinaryExpressionNode) {
                const call = this.getStashedNodeAnnotation(node.lhs, annotations_1.annotations.ImplicitCall);
                if (call) {
                    this.removeStashedNodeAnnotation(node.lhs, call);
                }
                const memberLhsType = this.getType(node.lhs.lhs);
                const memberRhsType = this.getType(node.lhs.rhs);
                if (!(0, typeHelpers_1.isValidType)(memberLhsType)) {
                    this.messageCollector.error('Cannot resolve type', node.lhs.lhs.astNode);
                    return;
                }
                if (!(0, typeHelpers_1.isValidType)(memberRhsType)) {
                    this.messageCollector.error('Cannot resolve type', node.lhs.rhs.astNode);
                    return;
                }
                const memberName = node.lhs.operator.name;
                const memberType = this.resolveTypeMember(node, memberLhsType, memberName, this.messageCollector, nodeScope);
                if (memberType) {
                    const fun = this.findFunctionOverload(memberType, [memberLhsType, memberRhsType, rhsType], node.lhs, null, false, this.messageCollector, true, node.lhs.scope);
                    if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                        this.annotateImplicitCall(node, fun, [node.lhs.lhs, node.lhs.rhs, node.rhs], this.messageCollector, nodeScope);
                        this.setType(node, fun.signature.returnType);
                    }
                    else {
                        this.messageCollector.error('Overload not found', node.rhs.astNode);
                    }
                }
                else {
                    this.setType(node, types_1.UNRESOLVED_TYPE);
                }
            }
            else {
                const result = this.ensureCanBeAssignedWithImplicitConversion(rhsType, lhsType, node.rhs, this.messageCollector, nodeScope);
                if (!types_1.NeverType.isNeverType(rhsType) && rhsType.nativeType === types_1.NativeTypes.void) {
                    this.messageCollector.error('The expression returns a void value, which cannot be assigned to any value', node.rhs.astNode);
                }
                if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                    this.setType(node, result.type);
                }
                else {
                    this.setType(node, types_1.InjectableTypes.void);
                }
            }
        }
        else if (node instanceof nodes_1.Nodes.FunctionCallNode) {
            const functionType = this.getType(node.functionNode);
            if ((0, typeHelpers_1.isValidType)(functionType)) {
                const argumentTypes = node.argumentsNode.map($ => this.getType($));
                const validArguments = argumentTypes.every($ => (0, typeHelpers_1.isValidType)($));
                if (validArguments && (0, typeHelpers_1.isValidType)(functionType)) {
                    if (node.functionNode instanceof nodes_1.Nodes.MemberNode &&
                        this.getStashedNodeAnnotation(node.functionNode, annotations_1.annotations.MethodCall)) {
                        // Detect if we are calling a <instance>.method(...)
                        // to inject the "self" argument
                        const fun = this.findFunctionOverload(functionType, [this.getType(node.functionNode.lhs), ...argumentTypes], node, null, false, this.messageCollector, true, nodeScope);
                        if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                            this.annotateImplicitCall(node, fun, [node.functionNode.lhs, ...node.argumentsNode], this.messageCollector, nodeScope);
                            this.setType(node, fun.signature.returnType);
                        }
                        else {
                            this.setType(node, types_1.UNRESOLVED_TYPE);
                        }
                    }
                    else {
                        const fun = this.findFunctionOverload(functionType, argumentTypes, node, null, false, this.messageCollector, true, nodeScope);
                        if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                            this.annotateImplicitCall(node, fun, node.argumentsNode, this.messageCollector, nodeScope);
                            this.setType(node, fun.signature.returnType);
                        }
                        else if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionSignatureType) {
                            const currentAnnotation = this.getStashedNodeAnnotation(node, annotations_1.annotations.IsFunctionReference);
                            if (currentAnnotation) {
                                currentAnnotation.fun = fun;
                            }
                            else {
                                this.stashNodeAnnotation(node, new annotations_1.annotations.IsFunctionReference(fun));
                            }
                            this.setType(node, fun.returnType);
                        }
                        else {
                            this.setType(node, types_1.UNRESOLVED_TYPE);
                        }
                    }
                }
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny('Not a function', node.functionNode);
            }
        }
        else if (node instanceof nodes_1.Nodes.FunctionNode) {
            const functionType = new types_1.FunctionType(node.functionName);
            node.parameters.forEach(($, ix) => {
                const parameterType = this.getType($.parameterName);
                functionType.signature.parameterNames[ix] = $.parameterName.name;
                functionType.signature.parameterTypes[ix] = parameterType;
            });
            const retType = node.functionReturnType
                ? this.getTypeFromTypeNode(node.functionReturnType, this.messageCollector)
                : null;
            if ((0, typeHelpers_1.isValidType)(retType)) {
                functionType.signature.returnType = retType;
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny('Cannot resolve return type', node.functionReturnType || node);
            }
            if (node.body) {
                const inferedReturnType = this.getType(node.body);
                if (inferedReturnType) {
                    if (inferedReturnType instanceof types_1.TypeType) {
                        this.messageCollector.error(new NodeError_1.UnexpectedType(inferedReturnType, node.body));
                    }
                    else {
                        this.ensureCanBeAssignedWithImplicitConversion(inferedReturnType, functionType.signature.returnType, node.body, this.messageCollector, nodeScope);
                    }
                }
                else {
                    this.messageCollector.errorIfBranchDoesntHaveAny(new NodeError_1.CannotInferReturnType(node.body || node));
                }
            }
            this.setType(node.functionName, functionType);
        }
        else if (node instanceof nodes_1.Nodes.ParameterNode) {
            // TODO: const valueType = TypeHelpers.getNodeType(node.defaultValue);
            // TODO: verify assignability
            const typeType = node.parameterType ? this.getTypeFromTypeNode(node.parameterType, this.messageCollector) : null;
            if ((0, typeHelpers_1.isValidType)(typeType)) {
                this.setType(node.parameterName, typeType);
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny('Cannot infer type 6' + typeType, node);
            }
        }
        else if (node instanceof nodes_1.Nodes.VarDeclarationNode) {
            const valueType = types_1.TypeHelpers.getNodeType(node.value);
            const typeType = node.variableType ? this.getTypeFromTypeNode(node.variableType, this.messageCollector) : null;
            if (node.variableType) {
                if ((0, typeHelpers_1.isValidType)(valueType) && (0, typeHelpers_1.isValidType)(typeType)) {
                    const ret = this.ensureCanBeAssignedWithImplicitConversion(valueType, typeType, node.value, this.messageCollector, nodeScope);
                    node.value = ret.node;
                    this.setType(node.variableName, ret.type);
                }
                else if ((0, typeHelpers_1.isValidType)(typeType)) {
                    this.setType(node.variableName, typeType);
                }
                else if ((0, typeHelpers_1.isValidType)(valueType)) {
                    this.setType(node.variableName, valueType);
                }
                else {
                    this.setType(node.variableName, types_1.UNRESOLVED_TYPE);
                }
            }
            else if (valueType) {
                this.setType(node.variableName, valueType);
            }
            else {
                this.setType(node.variableName, types_1.UNRESOLVED_TYPE);
            }
        }
        else if (node instanceof nodes_1.Nodes.BlockNode) {
            if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                if (node.statements.length === 0) {
                    this.setType(node, types_1.InjectableTypes.void);
                }
                else {
                    const lastStatement = (0, helpers_1.last)(node.statements);
                    const type = this.getType(lastStatement);
                    if ((0, typeHelpers_1.isValidType)(type)) {
                        this.setType(node, type);
                    }
                    else {
                        this.messageCollector.errorIfBranchDoesntHaveAny('Cannot resolve type 6', lastStatement);
                    }
                }
            }
            else {
                this.setType(node, types_1.InjectableTypes.void);
            }
        }
        else if (node instanceof nodes_1.Nodes.MemberNode) {
            const lhsType = this.getType(node.lhs);
            if (!(0, typeHelpers_1.isValidType)(lhsType)) {
                this.messageCollector.errorIfBranchDoesntHaveAny('Cannot resolve type 5', node.lhs);
                this.setType(node, types_1.UNRESOLVED_TYPE);
                return;
            }
            if (node.operator === '.^') {
                if (lhsType instanceof types_1.TypeType) {
                    const allowedTypeSchemas = lhsType.schema();
                    if (node.memberName.name in allowedTypeSchemas) {
                        this.setType(node, allowedTypeSchemas[node.memberName.name]);
                        return;
                    }
                    else {
                        const keys = Object.keys(allowedTypeSchemas);
                        if (keys.length) {
                            this.messageCollector.error(new NodeError_1.LysTypeError(`Invalid schema property. Available options: ${keys.join(', ')}`, node.memberName));
                        }
                        else {
                            this.messageCollector.error(new NodeError_1.LysTypeError(`The type "${lhsType}" has no schema`, node.lhs));
                        }
                    }
                }
                else {
                    this.messageCollector.errorIfBranchDoesntHaveAny(new NodeError_1.NotAValidType(node.lhs, lhsType));
                }
                this.setType(node, types_1.UNRESOLVED_TYPE);
            }
            else {
                if (lhsType instanceof types_1.TypeType) {
                    const memberName = node.memberName.name;
                    const memberType = this.resolveTypeMember(node.memberName, lhsType.of, memberName, this.messageCollector, nodeScope);
                    if (memberType) {
                        if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                            /**
                             * This branch covers Type.'variable
                             *
                             * impl Test {
                             *   var CONST = 1
                             * }
                             *
                             * Test.CONST
                             * ^^^^^^^^^^
                             */
                            // a if (memberType.referencedNode) {
                            // a   node.resolvedReference = new Reference(
                            // a     memberType.referencedNode,
                            // a     nodeScope,
                            // a     'VALUE',
                            // a     memberType.referencedNode.astNode.moduleName
                            // a   );
                            // a } else {
                            // a   this.messageCollector.errorIfBranchDoesntHaveAny(new NotAValidType(node.lhs));
                            // a }
                            if (!node.resolvedReference) {
                                this.messageCollector.errorIfBranchDoesntHaveAny('Reference not resolved', node.lhs);
                            }
                        }
                        this.setType(node, memberType);
                    }
                    else {
                        this.setType(node, types_1.UNRESOLVED_TYPE);
                    }
                }
                else if (lhsType instanceof types_1.TypeAlias) {
                    // we are inside if (node is MemberNode)
                    // lhsType is typeOf(node.lhs)
                    // lhsType is TypeAlias means we are receiving a value as LHS
                    // i.e: "asd".length
                    const memberName = node.memberName.name;
                    const memberType = this.resolveTypeMember(node.memberName, lhsType, memberName, this.messageCollector, nodeScope);
                    if (memberType) {
                        const isValueNode = !node.hasAnnotation(annotations_1.annotations.IsAssignationLHS);
                        if (isValueNode) {
                            const prop = this.describeProperty(memberType);
                            if (prop) {
                                if (prop.getter) {
                                    const fun = this.findFunctionOverload(memberType, [lhsType], node, null, false, this.messageCollector, false, nodeScope);
                                    if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                                        this.annotateImplicitCall(node, fun, [node.lhs], this.messageCollector, nodeScope);
                                        this.setType(node, fun.signature.returnType);
                                    }
                                    else {
                                        this.messageCollector.error(new NodeError_1.LysTypeError(`${lhsType}.${memberName} is not a valid property getter ${fun} ${memberType.inspect(100)}`, node.memberName));
                                        this.setType(node, types_1.UNRESOLVED_TYPE);
                                    }
                                }
                                else if (prop.method) {
                                    this.stashNodeAnnotation(node, new annotations_1.annotations.MethodCall());
                                    this.setType(node, memberType);
                                }
                                else {
                                    this.messageCollector.error(new NodeError_1.LysTypeError(`${lhsType}.${memberName} is not a getter or method`, node.memberName));
                                    this.setType(node, types_1.UNRESOLVED_TYPE);
                                }
                            }
                            else {
                                this.messageCollector.error(new NodeError_1.LysTypeError(`${lhsType}.${memberName} is not a getter or method`, node.memberName));
                                this.setType(node, types_1.UNRESOLVED_TYPE);
                            }
                        }
                        else {
                            this.setType(node, memberType);
                        }
                    }
                    else {
                        this.setType(node, types_1.UNRESOLVED_TYPE);
                    }
                }
                else {
                    this.messageCollector.errorIfBranchDoesntHaveAny(`Don't know how to find a member in "${lhsType}"`, node.memberName);
                }
            }
        }
        else if (node instanceof nodes_1.Nodes.ImplDirective) {
            if (node.baseImpl && node.baseImpl.resolvedReference) {
                const trait = this.getType(node.baseImpl.resolvedReference.referencedNode);
                if (node.targetImpl.resolvedReference) {
                    const targetType = this.getType(node.targetImpl.resolvedReference.referencedNode);
                    if (node.selfTypeName && targetType) {
                        this.setType(node.selfTypeName, targetType);
                    }
                }
                if (trait) {
                    if (trait instanceof types_1.TraitType) {
                        if (!(0, typeHelpers_1.isValidType)(trait)) {
                            this.messageCollector.errorIfBranchDoesntHaveAny('Trait not ready', node.baseImpl);
                        }
                        else {
                            const requiredImplementations = new Set(trait.traitNode.namespaceNames.keys());
                            const givenImplementations = [];
                            // verify every implementation function against the trait's signatures
                            for (let [name, nameNode] of node.namespaceNames) {
                                givenImplementations.push(name);
                                const type = this.getType(nameNode);
                                const signature = trait.getSignatureOf(name);
                                if (signature && type && !type.canBeAssignedTo(signature, nodeScope)) {
                                    this.messageCollector.errorIfBranchDoesntHaveAny(new NodeError_1.TypeMismatch(type, signature, nameNode));
                                }
                                if (requiredImplementations.has(name)) {
                                    requiredImplementations.delete(name);
                                }
                                else {
                                    // TODO: check extra names and suggest private implementations
                                }
                            }
                            if (requiredImplementations.size) {
                                this.messageCollector.errorIfBranchDoesntHaveAny(`Not all functions are implemented, missing: ${Array.from(requiredImplementations).join(', ')}`, node.baseImpl);
                            }
                        }
                    }
                    else {
                        // TODO: test this
                        this.messageCollector.errorIfBranchDoesntHaveAny(`${trait.inspect(100)} is not a trait`, node.baseImpl);
                    }
                }
            }
            if (node.targetImpl.resolvedReference) {
                // this will fail if the resolved reference is not a type
                this.getTypeTypeType(node.targetImpl, this.getType(node.targetImpl.resolvedReference.referencedNode), this.messageCollector);
            }
        }
        else if (node instanceof nodes_1.Nodes.TypeDirectiveNode) {
            if (node.valueType) {
                let valueType = this.getTypeFromTypeNode(node.valueType, this.messageCollector);
                if ((0, typeHelpers_1.isValidType)(valueType)) {
                    const type = createTypeAlias(node.variableName, valueType, this.parsingContext);
                    type.name.impls.forEach(impl => {
                        if (impl.baseImpl && impl.baseImpl.resolvedReference) {
                            const trait = this.getType(impl.baseImpl.resolvedReference.referencedNode);
                            if (trait instanceof types_1.TraitType) {
                                type.traits.set(impl, trait);
                            }
                        }
                    });
                    this.setType(node.variableName, types_1.TypeType.of(type));
                }
                else {
                    this.messageCollector.errorIfBranchDoesntHaveAny('Invalid type', node.valueType || node);
                }
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny('Missing value type', node);
            }
        }
        else if (node instanceof nodes_1.Nodes.FunctionTypeNode) {
            const signatureType = new types_1.FunctionSignatureType();
            node.parameters.forEach(($, ix) => {
                const parameterType = this.getTypeFromTypeNode($.parameterType, this.messageCollector);
                signatureType.parameterTypes[ix] = parameterType;
            });
            const retType = node.returnType ? this.getTypeFromTypeNode(node.returnType, this.messageCollector) : null;
            if ((0, typeHelpers_1.isValidType)(retType)) {
                signatureType.returnType = retType;
            }
            else {
                this.messageCollector.errorIfBranchDoesntHaveAny('Cannot resolve return type', node.returnType || node);
            }
            this.setType(node, signatureType);
        }
        else if (node instanceof nodes_1.Nodes.IntersectionTypeNode) {
            const of = node.of.map($ => this.getTypeTypeType($, this.getType($), this.messageCollector));
            this.setType(node, new types_1.IntersectionType(of));
        }
        else if (node instanceof nodes_1.Nodes.UnionTypeNode) {
            const of = node.of.map($ => this.getTypeTypeType($, this.getType($), this.messageCollector));
            this.setType(node, new types_1.UnionType(of));
        }
        else if (node instanceof nodes_1.Nodes.OverloadedFunctionNode) {
            const incomingTypes = node.functions.map(fun => {
                if (!fun.functionNode.hasAnnotation(annotations_1.annotations.IsOverloaded)) {
                    this.stashNodeAnnotation(fun.functionNode, new annotations_1.annotations.IsOverloaded());
                }
                return this.getType(fun.functionNode.functionName);
            });
            const incomingFunctionTypes = [];
            for (let incomingType of incomingTypes) {
                if (incomingType instanceof types_1.FunctionType) {
                    incomingFunctionTypes.push(incomingType);
                }
                else {
                    this.messageCollector.error(`All members of an overloaded function should be functions, but found ${incomingType}`, node.astNode);
                }
            }
            const newType = new types_1.IntersectionType(incomingFunctionTypes);
            this.setType(node.functionName, newType);
        }
        else if (node instanceof nodes_1.Nodes.WasmExpressionNode) {
            this.setType(node, types_1.InjectableTypes.never);
        }
        else if (node instanceof nodes_1.Nodes.UnknownExpressionNode) {
            this.setType(node, types_1.InjectableTypes.never);
        }
        else if (node instanceof nodes_1.Nodes.LoopNode) {
            this.setType(node, types_1.InjectableTypes.void);
        }
        else if (node instanceof nodes_1.Nodes.PatternMatcherNode) {
            if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                const restType = node.carryType;
                if ((0, typeHelpers_1.isValidType)(restType) && !types_1.NeverType.isNeverType(restType)) {
                    this.messageCollector.error(`Match is not exhaustive, not covered types: "${restType}"`, node.astNode);
                }
                const retType = types_1.UnionType.of(node.matchingSet.map($ => this.getType($))).simplify(nodeScope);
                this.setType(node, retType);
            }
            else {
                this.setType(node, types_1.InjectableTypes.void);
            }
            // TODO: warn if empty matcher
        }
        else if (node instanceof nodes_1.Nodes.VarDeclarationNode) {
            if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
                this.messageCollector.error('This should be an expression. A declaration was found.', node.astNode);
            }
            this.setType(node, types_1.InjectableTypes.void);
        }
        else if (node instanceof nodes_1.Nodes.ExpressionNode) {
            const type = this.getType(node);
            if (!(0, typeHelpers_1.isValidType)(type)) {
                this.messageCollector.errorIfBranchDoesntHaveAny(`Cannot resolve type of node ${node.nodeName}`, node);
            }
        }
    }
    onEnterNode(node) {
        if (node instanceof nodes_1.Nodes.PatternMatcherNode) {
            // this value must be cleared in every execution
            delete node.carryType;
        }
    }
    processMatcherNode(node, parent) {
        if (!node.scope) {
            this.messageCollector.errorIfBranchDoesntHaveAny(`Node ${node.nodeName} has no scope`, node);
            return;
        }
        const nodeScope = node.scope;
        const carryType = parent.carryType || this.getType(parent.lhs);
        if (!(0, typeHelpers_1.isValidType)(carryType)) {
            return;
        }
        if (!parent.carryType) {
            parent.carryType = carryType;
        }
        const booleanType = node.booleanReference
            ? this.getTypeTypeType(node, this.getType(node.booleanReference.referencedNode), this.messageCollector)
            : null;
        if (!(0, typeHelpers_1.isValidType)(booleanType)) {
            return;
        }
        if (types_1.NeverType.isNeverType(carryType)) {
            this.messageCollector.error(new NodeError_1.UnreachableCode(node.rhs));
            this.messageCollector.error(carryType.inspect(10), node.rhs.astNode);
            this.stashNodeAnnotation(node.rhs, new annotations_1.annotations.IsUnreachable());
            this.setType(node, types_1.UNRESOLVED_TYPE);
            return;
        }
        if (node instanceof nodes_1.Nodes.MatchDefaultNode) {
            this.removeTypeFromMatcherFlow(parent, carryType);
        }
        else if (node instanceof nodes_1.Nodes.MatchLiteralNode) {
            const argumentType = this.getType(node.literal);
            if (!(0, typeHelpers_1.isValidType)(argumentType)) {
                return;
            }
            const eqFunction = this.resolveTypeMember(node.literal, carryType, '==', this.messageCollector, nodeScope);
            if (eqFunction) {
                const argTypes = [carryType, argumentType];
                const fun = this.findFunctionOverload(eqFunction, argTypes, node.literal, null, false, new MessageCollector_1.MessageCollector(), true, nodeScope);
                if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                    node.resolvedFunctionType = fun;
                    if (!this.canBeAssigned(fun.signature.returnType, booleanType, nodeScope)) {
                        this.messageCollector.error(new NodeError_1.TypeMismatch(fun.signature.returnType, booleanType, node));
                    }
                }
                else {
                    this.messageCollector.error(new NodeError_1.TypeMismatch(carryType, argumentType, node.literal));
                    this.messageCollector.error(new NodeError_1.UnreachableCode(node.rhs));
                    this.stashNodeAnnotation(node.rhs, new annotations_1.annotations.IsUnreachable());
                }
            }
        }
        else if (node instanceof nodes_1.Nodes.MatchCaseIsNode) {
            const argumentType = this.getTypeTypeType(node.typeReference, this.getType(node.typeReference), this.messageCollector);
            if (!(0, typeHelpers_1.isValidType)(argumentType)) {
                return;
            }
            if (node.declaredName) {
                this.setType(node.declaredName, argumentType);
            }
            if (!this.canBeAssigned(carryType, types_1.RefType.instance, nodeScope)) {
                this.messageCollector.error(`"is" expression can only be used with reference types, used with: "${carryType}"`, node.astNode);
            }
            else if (!this.canBeAssigned(carryType, argumentType, nodeScope) &&
                !this.canBeAssigned(argumentType, carryType, nodeScope)) {
                this.messageCollector.error(new NodeError_1.TypeMismatch(carryType, argumentType, node.typeReference));
                this.messageCollector.error(new NodeError_1.UnreachableCode(node.rhs));
                this.stashNodeAnnotation(node.rhs, new annotations_1.annotations.IsUnreachable());
            }
            else {
                const eqFunction = this.resolveTypeMember(node.typeReference, argumentType, 'is', this.messageCollector, nodeScope);
                if (eqFunction) {
                    const fun = this.findFunctionOverload(eqFunction, [argumentType], node, null, false, new MessageCollector_1.MessageCollector(), false, nodeScope);
                    if ((0, typeHelpers_1.isValidType)(fun) && fun instanceof types_1.FunctionType) {
                        node.resolvedFunctionType = fun;
                        if (!this.canBeAssigned(fun.signature.returnType, booleanType, nodeScope)) {
                            this.messageCollector.error(new NodeError_1.TypeMismatch(fun.signature.returnType, booleanType, node));
                        }
                        this.removeTypeFromMatcherFlow(parent, argumentType);
                    }
                    else {
                        this.messageCollector.error(new NodeError_1.TypeMismatch(carryType, argumentType, node.typeReference));
                        this.messageCollector.error(new NodeError_1.UnreachableCode(node.rhs));
                        this.stashNodeAnnotation(node.rhs, new annotations_1.annotations.IsUnreachable());
                    }
                }
            }
        }
        const result = this.getType(node.rhs);
        if (node.hasAnnotation(annotations_1.annotations.IsValueNode)) {
            this.setType(node, result);
        }
        else {
            this.setType(node, types_1.InjectableTypes.void);
        }
    }
    annotateImplicitCall(nodeToAnnotate, fun, argumentNodes, messageCollector, scope) {
        const oldAnnotation = this.getStashedNodeAnnotation(nodeToAnnotate, annotations_1.annotations.ImplicitCall);
        const newAnnotation = new annotations_1.annotations.ImplicitCall(this.createImplicitCall(nodeToAnnotate, fun, argumentNodes, messageCollector, scope));
        if (oldAnnotation) {
            if (!fun.equals(oldAnnotation.implicitCall.resolvedFunctionType)) {
                messageCollector.error(new NodeError_1.LysTypeError(`This node already has an ${oldAnnotation} trying to add ${newAnnotation}`, nodeToAnnotate));
            }
        }
        else {
            this.stashNodeAnnotation(nodeToAnnotate, newAnnotation);
        }
    }
    removeTypeFromMatcherFlow(node, typeToRemove) {
        let carryType = node.carryType;
        if (!(0, typeHelpers_1.isValidType)(carryType)) {
            return;
        }
        if (!(0, typeHelpers_1.isValidType)(typeToRemove)) {
            return;
        }
        if (carryType instanceof types_1.UnionType) {
            carryType = carryType.expand(node.scope);
        }
        else {
            carryType = types_1.UnionType.of(carryType).expand(node.scope);
        }
        let matchingValueType = carryType instanceof types_1.UnionType ? carryType : types_1.UnionType.of(carryType);
        const newType = matchingValueType.subtract(typeToRemove, node.scope);
        matchingValueType = types_1.UnionType.of(newType);
        node.carryType = matchingValueType.simplify(node.scope);
    }
    getTypeTypeType(node, type, messageCollector) {
        if (type instanceof types_1.TypeType) {
            return type.of;
        }
        else {
            if (type) {
                messageCollector.errorIfBranchDoesntHaveAny(new NodeError_1.NotAValidType(node, type));
            }
            return null;
        }
    }
    getTypeFromTypeNode(node, messageCollector) {
        if (node instanceof nodes_1.Nodes.ReferenceNode) {
            return this.getTypeTypeType(node, types_1.TypeHelpers.getNodeType(node), messageCollector);
        }
        else if (node instanceof nodes_1.Nodes.FunctionTypeNode ||
            node instanceof nodes_1.Nodes.IntersectionTypeNode ||
            node instanceof nodes_1.Nodes.UnionTypeNode ||
            node instanceof nodes_1.Nodes.InjectedTypeNode ||
            node instanceof nodes_1.Nodes.StructTypeNode ||
            node instanceof nodes_1.Nodes.StackTypeNode) {
            return types_1.TypeHelpers.getNodeType(node);
        }
        else {
            messageCollector.errorIfBranchDoesntHaveAny(new NodeError_1.NotAValidType(node, null));
            return null;
        }
    }
    resolveTypeMember(errorNode, type, memberName, messageCollector, scope) {
        if (type && type instanceof types_1.TypeAlias) {
            const types = type.getTypeMember(memberName, $ => this.getType($));
            if (types.length) {
                const ret = [];
                for (let result of types) {
                    const [referencedNode, memberType] = result;
                    const parent = referencedNode.parent;
                    if (!(0, typeHelpers_1.isValidType)(memberType)) {
                        return false;
                    }
                    if (parent && parent instanceof nodes_1.Nodes.DirectiveNode) {
                        if (!parent.isPublic) {
                            const declScope = referencedNode.scope;
                            if (!declScope) {
                                if (errorNode) {
                                    messageCollector.error(new NodeError_1.LysTypeError(`Name's parent has no scope`, type.name));
                                }
                                return false;
                            }
                            if (!scope) {
                                if (errorNode) {
                                    messageCollector.error(new NodeError_1.LysTypeError(`Scope is null`, errorNode));
                                    console.trace();
                                }
                                return false;
                            }
                            if (!scope.isDescendantOf(declScope)) {
                                if (errorNode) {
                                    messageCollector.error(new NodeError_1.LysTypeError(`Name "${memberName}" is private in ${type.name.name}`, errorNode));
                                }
                                continue;
                            }
                        }
                    }
                    ret.push(memberType);
                }
                if (ret.length > 1) {
                    return new types_1.IntersectionType(ret).simplify();
                }
                if (ret.length === 1) {
                    return ret[0];
                }
                if (errorNode) {
                    messageCollector.errorIfBranchDoesntHaveAny(new NodeError_1.LysTypeError(`Type is not ready`, errorNode));
                }
                return false;
            }
            else {
                if (errorNode) {
                    messageCollector.error(new NodeError_1.LysTypeError(`Property "${memberName}" doesn't exist on type "${type}".`, errorNode));
                }
                // Since the members are calculated in the scope phase we return
                // `never` type so the type check execution can continue
                return types_1.InjectableTypes.never;
            }
        }
        else {
            if (errorNode) {
                if (!types_1.NeverType.isNeverType(type)) {
                    messageCollector.error(new NodeError_1.LysTypeError(`Type ${type.inspect(10)} has no members. Members are stored in type aliases.`, errorNode));
                }
            }
        }
        return false;
    }
    findFunctionOverload(incommingType, argTypes, errorNode, returnType, strict, messageCollector, automaticCoercion, scope) {
        if (incommingType instanceof types_1.TypeType) {
            return this.findFunctionOverload(incommingType.of, argTypes, errorNode, returnType, strict, messageCollector, automaticCoercion, scope);
        }
        if (incommingType instanceof types_1.IntersectionType) {
            const matchList = [];
            for (let fun of incommingType.of) {
                if (fun instanceof types_1.FunctionType) {
                    if (strict) {
                        if (this.acceptsTypes(fun.signature, argTypes, strict, automaticCoercion, scope)) {
                            if (!returnType || fun.signature.returnType.equals(returnType)) {
                                return fun;
                            }
                        }
                    }
                    else {
                        const result = this.acceptsTypes(fun.signature, argTypes, strict, automaticCoercion, scope);
                        if (result) {
                            if (!returnType || this.canBeAssigned(fun.signature.returnType, returnType, scope)) {
                                matchList.push({ fun, score: result.score, casts: result.casts });
                            }
                        }
                    }
                }
                else {
                    if (errorNode) {
                        messageCollector.error(new NodeError_1.NotAFunction(fun, errorNode));
                    }
                    return null;
                }
            }
            if (matchList.length === 1) {
                return matchList[0].fun;
            }
            else if (matchList.length > 1) {
                matchList.sort((a, b) => {
                    if (a.score < b.score) {
                        return 1;
                    }
                    else if (a.score > b.score) {
                        return -1;
                    }
                    else {
                        return 0;
                    }
                });
                if (matchList[0].score > matchList[1].score) {
                    return matchList[0].fun;
                }
                let selectedOverload = 0;
                if (errorNode) {
                    messageCollector.warning(`Implicit overload with ambiguity.\nPicking overload ${matchList[selectedOverload].fun.toString()} for types (${argTypes
                        .map($ => $.toString())
                        .join(', ')})\nComplete list of overloads:\n${matchList
                        .map($ => '  ' + $.fun.toString() + ' score: ' + $.score)
                        .join('\n')}`, errorNode);
                }
                return matchList[selectedOverload].fun;
            }
            if (errorNode) {
                if (incommingType.of.length > 1) {
                    messageCollector.error(new NodeError_1.InvalidOverload(incommingType, argTypes, errorNode));
                }
                else {
                    const fun = incommingType.of[0];
                    messageCollector.error(new NodeError_1.InvalidCall(fun.signature.parameterTypes, argTypes, errorNode));
                }
                return null;
            }
            return new types_1.UnionType(incommingType.of.map(($) => $.signature.returnType));
        }
        else if (incommingType instanceof types_1.FunctionType) {
            const queryResult = this.acceptsTypes(incommingType.signature, argTypes, strict, automaticCoercion, scope);
            if (!queryResult) {
                if (errorNode) {
                    messageCollector.error(new NodeError_1.InvalidCall(incommingType.signature.parameterTypes, argTypes, errorNode));
                }
                else {
                    return null;
                }
            }
            return incommingType;
        }
        else if (incommingType instanceof types_1.FunctionSignatureType) {
            const queryResult = this.acceptsTypes(incommingType, argTypes, strict, automaticCoercion, scope);
            if (!queryResult) {
                if (errorNode) {
                    messageCollector.error(new NodeError_1.InvalidCall(incommingType.parameterTypes, argTypes, errorNode));
                }
                else {
                    return null;
                }
            }
            return incommingType;
        }
        else {
            if (errorNode && !types_1.NeverType.isNeverType(incommingType)) {
                messageCollector.error(new NodeError_1.NotAFunction(incommingType, errorNode));
            }
            return null;
        }
    }
    findImplicitTypeCasting(errorNode, from, to, messageCollector, scope) {
        if (this.canBeAssigned(from, to, scope)) {
            return null;
        }
        if (from instanceof types_1.TypeAlias) {
            try {
                const fnType = this.resolveTypeMember(errorNode, from, 'as', messageCollector, scope);
                if (fnType) {
                    const fun = this.findFunctionOverload(fnType, [from], null, to, true, messageCollector, true, scope);
                    if (fun instanceof types_1.FunctionType) {
                        if (!fun.name.hasAnnotation(annotations_1.annotations.Explicit)) {
                            if (this.canBeAssigned(fun.signature.returnType, to, scope)) {
                                return fun;
                            }
                        }
                    }
                }
            }
            catch (e) {
                return null;
            }
        }
        return null;
    }
    sizeInBytes(type, defaultSize) {
        if (type instanceof types_1.StackType) {
            return type.byteSize;
        }
        if (type instanceof types_1.RefType) {
            return type.byteSize;
        }
        switch (type.binaryenType) {
            case types_1.NativeTypes.i32:
            case types_1.NativeTypes.f32:
                return 4;
            case types_1.NativeTypes.i64:
            case types_1.NativeTypes.f64:
                return 8;
        }
        return defaultSize;
    }
    acceptsTypes(type, types, strict, automaticCoercion, scope) {
        if (type.parameterTypes.length !== types.length) {
            return null;
        }
        let score = 1;
        let casts = [];
        if (type.parameterTypes.length === 0) {
            return { score: Infinity, casts: [] };
        }
        for (let index = 0; index < types.length; index++) {
            const argumentType = types[index];
            const parameterType = type.parameterTypes[index];
            const equals = argumentType.equals(parameterType);
            if (equals) {
                score += 1;
                casts.push(null);
            }
            else if (!strict) {
                const cleanAssignation = this.canBeAssigned(argumentType, parameterType, scope);
                if (cleanAssignation) {
                    score += this.getTypeSimilarity(argumentType, parameterType);
                }
                else if (automaticCoercion) {
                    try {
                        const implicitCast = this.findImplicitTypeCasting(null, argumentType, parameterType, new MessageCollector_1.MessageCollector(), scope);
                        if (implicitCast) {
                            casts.push(implicitCast);
                            const dataLossInBytes = this.sizeInBytes(parameterType, 0) / this.sizeInBytes(argumentType, 1);
                            if (isNaN(dataLossInBytes))
                                return null;
                            score += 0.5 * Math.min(((types.length - index) / types.length) * dataLossInBytes, 1);
                        }
                        else {
                            return null;
                        }
                    }
                    catch {
                        return null;
                    }
                }
                else {
                    return null;
                }
            }
            else {
                return null;
            }
        }
        return {
            score,
            casts
        };
    }
    downToRefTypes(type) {
        let argType = type;
        while (true) {
            if (argType instanceof types_1.StructType) {
                return [argType];
            }
            else if (argType instanceof types_1.RefType) {
                return [argType];
            }
            else if (argType instanceof types_1.TypeAlias) {
                argType = argType.of;
            }
            else if (argType instanceof types_1.UnionType) {
                return (0, helpers_1.flatten)(argType.of.map($ => this.downToRefTypes($)));
            }
            else {
                return [];
            }
        }
    }
    getTypeSimilarity(lhs, rhs) {
        if (rhs.equals(lhs) && lhs.equals(rhs)) {
            return 1;
        }
        const lhsTypes = this.downToRefTypes(lhs);
        if (lhsTypes.length === 0)
            return 0;
        const rhsTypes = this.downToRefTypes(rhs);
        if (rhsTypes.length === 0)
            return 0;
        let results = [];
        lhsTypes.forEach(lhs => rhsTypes.forEach(rhs => results.push(lhs.typeSimilarity(rhs))));
        return Math.max.apply(Math, results);
    }
    canBeAssigned(sourceType, targetType, scope) {
        if (!sourceType || !sourceType.canBeAssignedTo) {
            console.trace();
            console.log(sourceType, sourceType.toString());
            console.log(sourceType.inspect(10));
        }
        return sourceType.canBeAssignedTo(targetType, scope);
    }
    ensureCanBeAssignedWithImplicitConversion(sourceType, targetType, node, messageCollector, scope) {
        if (sourceType instanceof types_1.IntersectionType &&
            targetType instanceof types_1.FunctionSignatureType &&
            this.canBeAssigned(sourceType, targetType, scope)) {
            for (let fun of sourceType.of) {
                if (this.canBeAssigned(fun, targetType, scope)) {
                    if (fun instanceof types_1.FunctionType) {
                        const currentAnnotation = this.getStashedNodeAnnotation(node, annotations_1.annotations.FunctionInTable);
                        if (currentAnnotation) {
                            currentAnnotation.nameIdentifier = fun.name;
                        }
                        else {
                            this.stashNodeAnnotation(node, new annotations_1.annotations.FunctionInTable(fun.name));
                        }
                    }
                    return { node, type: targetType };
                }
            }
        }
        else if (this.canBeAssigned(sourceType, targetType, scope)) {
            return { node, type: targetType };
        }
        else {
            const implicitCast = this.findImplicitTypeCasting(node, sourceType, targetType, new MessageCollector_1.MessageCollector(), scope);
            if (implicitCast) {
                return { node: this.createImplicitCall(node, implicitCast, [node], messageCollector, scope), type: targetType };
            }
        }
        messageCollector.error(new NodeError_1.TypeMismatch(sourceType, targetType, node));
        return { node, type: targetType };
    }
    createImplicitCall(node, fun, args, messageCollector, scope) {
        const functionCallNode = new nodes_1.Nodes.InjectedFunctionCallNode(node.astNode);
        functionCallNode.scope = node.scope;
        functionCallNode.argumentsNode = args;
        functionCallNode.annotate(new annotations_1.annotations.Injected());
        functionCallNode.resolvedFunctionType = fun;
        this.ensureArgumentCoercion(functionCallNode, fun.signature, args.map($ => this.getType($)), messageCollector, scope);
        types_1.TypeHelpers.setNodeType(functionCallNode, fun.signature.returnType);
        return functionCallNode;
    }
    ensureArgumentCoercion(node, fun, argsTypes, messageCollector, scope) {
        node.argumentsNode.forEach((argNode, i) => {
            const argType = argsTypes[i];
            if (argType) {
                const ret = this.ensureCanBeAssignedWithImplicitConversion(argType, fun.parameterTypes[i], argNode, messageCollector, scope);
                node.argumentsNode[i] = ret.node;
            }
            else {
                messageCollector.error(new NodeError_1.LysTypeError('type is undefined', node.argumentsNode[i]));
            }
        });
    }
    describeProperty(type) {
        if (type instanceof types_1.FunctionType) {
            return {
                type,
                method: type.name.hasAnnotation(annotations_1.annotations.Method),
                getter: type.name.hasAnnotation(annotations_1.annotations.Getter),
                setter: type.name.hasAnnotation(annotations_1.annotations.Setter)
            };
        }
        if (type instanceof types_1.IntersectionType) {
            const ret = {
                type,
                getter: false,
                setter: false,
                method: false
            };
            type.of
                .map($ => this.describeProperty($))
                .forEach($ => {
                if ($) {
                    ret.getter = ret.getter || $.getter;
                    ret.setter = ret.setter || $.setter;
                    ret.method = ret.method || $.method;
                }
            });
            if (ret.getter || ret.setter || ret.method) {
                return ret;
            }
        }
        return void 0;
    }
}
exports.TypeAnalyzer = TypeAnalyzer;
//# sourceMappingURL=TypeAnalyzer.js.map