"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.main = void 0;
const arg_1 = __importDefault(require("arg"));
const ParsingContext_1 = require("./compiler/ParsingContext");
const path_1 = require("path");
const testEnvironment_1 = require("./utils/testEnvironment");
const test_1 = require("./utils/libs/test");
const NodeSystem_1 = require("./support/NodeSystem");
const fs_1 = require("fs");
const colors_1 = require("./utils/colors");
const errorPrinter_1 = require("./utils/errorPrinter");
const index_1 = require("./index");
const loadFromMD_1 = require("./utils/loadFromMD");
const nodePrinter_1 = require("./utils/nodePrinter");
function mkdirRecursive(dir) {
    // we explicitly don't use `path.sep` to have it platform independent;
    let sep = '/';
    let segments = dir.split(sep);
    let current = '';
    let i = 0;
    while (i < segments.length) {
        current = current + sep + segments[i];
        NodeSystem_1.nodeSystem.createDirectory(current);
        i++;
    }
}
function getParsingContext(cwd, argv) {
    NodeSystem_1.nodeSystem.cwd = cwd;
    const parsingContext = new ParsingContext_1.ParsingContext(NodeSystem_1.nodeSystem);
    parsingContext.paths.push(NodeSystem_1.nodeSystem.resolvePath(__dirname, '../stdlib'));
    let libs = [];
    let libPaths = [];
    const args = (0, arg_1.default)({
        '--output': String,
        '-o': '--output',
        '--no-optimize': Boolean,
        '--wast': Boolean,
        '--lib': [String],
        '--test': Boolean,
        '--debug': Boolean,
        '--desugar': Boolean,
        '--watch': Boolean,
        '--run': '--test'
    }, {
        permissive: true,
        argv
    });
    args['--lib'] = args['--lib'] || [];
    const DEBUG = !!args['--debug'];
    const DESUGAR = !!args['--desugar'];
    const WAST = !!args['--wast'];
    const WATCH = !!args['--watch'];
    const OPTIMIZE = !args['--no-optimize'];
    const file = args._[0];
    let customAssertions = {};
    if (!file) {
        throw new errorPrinter_1.LysError('Error: You did not specify an input file. \n\n  Usage: $ lys mainFile.lys');
    }
    if (!NodeSystem_1.nodeSystem.fileExists(file)) {
        throw new errorPrinter_1.LysError(`Error: File ${file} does not exist.`);
    }
    let mainModule = parsingContext.getModuleFQNForFile(file);
    if (file.endsWith('.md')) {
        const MD = (0, loadFromMD_1.loadFromMD)(parsingContext, parsingContext.system.readFile(file));
        mainModule = MD.mainModule;
        for (let path in MD.jsFiles) {
            if (path === 'assertions.js') {
                customAssertions[path] = MD.jsFiles[path];
            }
            else {
                libs.push(MD.jsFiles[path]);
            }
        }
        args['--test'] = true;
    }
    if (args['--test']) {
        args['--lib'].push(NodeSystem_1.nodeSystem.resolvePath(__dirname, 'utils/libs/env.js'));
        args['--lib'].push(NodeSystem_1.nodeSystem.resolvePath(__dirname, 'utils/libs/test.js'));
    }
    if (args['--lib'] && args['--lib'].length) {
        for (let libPath of args['--lib']) {
            const lib = NodeSystem_1.nodeSystem.resolvePath(libPath);
            if (!NodeSystem_1.nodeSystem.fileExists(lib)) {
                throw new errorPrinter_1.LysError(`Cannot find lib: ${lib}`);
            }
            const r = require(lib);
            if (!r.default) {
                throw new errorPrinter_1.LysError(`Library ${lib} has no "default" export`);
            }
            if (typeof r.default !== 'function') {
                throw new errorPrinter_1.LysError(`"default" is not a function in ${lib}`);
            }
            libPaths.push(lib);
            libs.push(r.default);
        }
    }
    if (!args['--output']) {
        args['--output'] = 'build/' + (0, path_1.basename)(file, '.lys');
    }
    const outFileFullWithoutExtension = NodeSystem_1.nodeSystem.resolvePath(NodeSystem_1.nodeSystem.getCurrentDirectory(), args['--output']);
    const outPath = (0, path_1.dirname)(outFileFullWithoutExtension);
    mkdirRecursive(outPath);
    return Object.assign(parsingContext, {
        mainModule,
        options: {
            DEBUG,
            DESUGAR,
            WAST,
            OPTIMIZE,
            WATCH,
            TEST: args['--test']
        },
        outFileFullWithoutExtension,
        outPath,
        libPaths,
        libs,
        customAssertions
    });
}
async function emit(parsingContext) {
    const codeGen = (0, index_1.compile)(parsingContext, parsingContext.mainModule, parsingContext.options.DEBUG);
    await codeGen.validate(parsingContext.options.OPTIMIZE, parsingContext.options.DEBUG || parsingContext.options.WAST);
    if (parsingContext.options.WAST) {
        NodeSystem_1.nodeSystem.writeFile(parsingContext.outFileFullWithoutExtension + '.wast', await codeGen.emitText());
    }
    if (codeGen.sourceMap) {
        NodeSystem_1.nodeSystem.writeFile(NodeSystem_1.nodeSystem.resolvePath(parsingContext.outPath, 'sourceMap.map'), codeGen.sourceMap);
    }
    if (!codeGen.buffer) {
        throw new errorPrinter_1.LysError('Could not generate WASM binary');
    }
    (0, fs_1.writeFileSync)(parsingContext.outFileFullWithoutExtension + '.wasm', codeGen.buffer);
    let src = [];
    src.push('Object.defineProperty(exports, "__esModule", { value: true });');
    src.push('const modules = [];');
    for (let i in parsingContext.libPaths) {
        const path = parsingContext.libPaths[i];
        src.push(`modules.push(require(${JSON.stringify((0, path_1.relative)((0, path_1.dirname)(parsingContext.outFileFullWithoutExtension), path))}).default);`);
    }
    const values = [];
    codeGen.buffer.forEach(($) => values.push($));
    src.push(`const buffer = new Uint8Array(${JSON.stringify(values)})`);
    src.push(`
exports.default = async function() {
  let instance = null;

  const getInstance = () => instance;

  let injectedModules = {};

  modules.forEach($ => {
    const generatedModule = $(getInstance);

    if (generatedModule) {
      injectedModules = { ...injectedModules, ...generatedModule };
    }
  });

  const compiled = await WebAssembly.compile(buffer);

  instance = new WebAssembly.Instance(compiled, injectedModules);

  return instance;
}
  `);
    NodeSystem_1.nodeSystem.writeFile(parsingContext.outFileFullWithoutExtension + '.js', src.join('\n'));
    if (parsingContext.options.DESUGAR) {
        parsingContext.modulesInContext.forEach(module => {
            if (module.fileName.startsWith(NodeSystem_1.nodeSystem.cwd)) {
                const relativePath = NodeSystem_1.nodeSystem.relative(NodeSystem_1.nodeSystem.cwd, module.fileName);
                const targetFile = NodeSystem_1.nodeSystem.resolvePath(parsingContext.outPath + '/desugar/', relativePath);
                mkdirRecursive((0, path_1.dirname)(targetFile));
                NodeSystem_1.nodeSystem.writeFile(targetFile, (0, nodePrinter_1.printNode)(module));
            }
        });
    }
    if (parsingContext.options.TEST) {
        const testInstance = await (0, testEnvironment_1.generateTestInstance)(codeGen.buffer, parsingContext.libs);
        if (typeof testInstance.exports.test !== 'function') {
            if (typeof testInstance.exports.main !== 'function') {
                throw new errorPrinter_1.LysError('There is no exported function named "main" or "test"');
            }
            else {
                testInstance.exports.main();
            }
        }
        else {
            testInstance.exports.test();
        }
        const testResults = (0, test_1.getTestResults)(testInstance);
        for (let path in parsingContext.customAssertions) {
            const startTime = Date.now();
            try {
                parsingContext.customAssertions[path](() => testInstance);
                testResults.push({
                    title: path,
                    endTime: Date.now(),
                    startTime,
                    passed: true,
                    skipped: false
                });
            }
            catch (e) {
                console.error(e);
                testResults.push({
                    title: path,
                    endTime: Date.now(),
                    startTime,
                    passed: false,
                    skipped: false
                });
            }
        }
        let totalTime = 0;
        let totalTests = 0;
        let totalPass = 0;
        const summarizeItem = (result) => {
            totalTime += result.endTime - result.startTime;
            totalTests++;
            if (result.passed) {
                totalPass++;
            }
            result.tests && result.tests.forEach(summarizeItem);
        };
        const summarize = (results) => {
            results.forEach(summarizeItem);
        };
        if (testResults && testResults.length) {
            summarize(testResults);
            NodeSystem_1.nodeSystem.writeFile('_lys_test_results.json', JSON.stringify(testResults, null, 2));
            if (totalPass !== totalTests) {
                console.error((0, colors_1.formatColorAndReset)(`\n\n  Some tests failed. Passed: ${totalPass} Failed: ${totalTests - totalPass} (${(totalTime / 1000).toFixed(2)}s)\n\n`, colors_1.ForegroundColors.Red));
                return false;
            }
            else {
                console.error((0, colors_1.formatColorAndReset)(`\n\n  All tests passed. Passed: ${totalPass} Failed: ${totalTests - totalPass} (${(totalTime / 1000).toFixed(2)}s)\n\n`, colors_1.ForegroundColors.Green));
                return true;
            }
        }
    }
    return parsingContext.messageCollector.errors.length === 0;
}
async function main(cwd, argv) {
    const parsingContext = getParsingContext(cwd, argv);
    if (parsingContext.options.WATCH) {
        let invalidated = true;
        let running = false;
        (0, fs_1.watch)(cwd, { recursive: true }, (event, fileName) => {
            const fqn = parsingContext.getModuleFQNForFile(fileName);
            const inContext = parsingContext.modulesInContext.has(fqn);
            if (inContext) {
                console.log(event, fileName);
                parsingContext.invalidateModule(fqn);
                invalidated = true;
            }
        });
        setInterval(async () => {
            if (invalidated && !running) {
                invalidated = false;
                const start = Date.now();
                console.log('Starting incremental compilation...');
                running = true;
                emit(parsingContext)
                    .then(() => {
                    running = false;
                    console.log((0, colors_1.formatColorAndReset)('[OK] ' + (Date.now() - start).toFixed(1) + 'ms', colors_1.ForegroundColors.Green));
                })
                    .catch(() => {
                    running = false;
                    console.log((0, colors_1.formatColorAndReset)('[ERROR] ' + (Date.now() - start).toFixed(1) + 'ms', colors_1.ForegroundColors.Red));
                });
            }
        }, 2000);
        await new Promise(() => void 0); // wait forever
    }
    else {
        if ((await emit(parsingContext)) === false) {
            process.exit(1);
        }
    }
}
exports.main = main;
//# sourceMappingURL=index-bin.js.map