"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.nodeSystem = exports.NodeSystem = void 0;
const _os = require("os");
const _fs = require("fs");
const _path = require("path");
const platform = _os.platform();
// NodeJS detects "\uFEFF" at the start of the string and *replaces* it with the actual
// byte order mark from the specified encoding. Using any other byte order mark does
// not actually work.
const byteOrderMarkIndicator = '\uFEFF';
/** Convert all lowercase chars to uppercase, and vice-versa */
function swapCase(s) {
    return s.replace(/\w/g, ch => {
        const up = ch.toUpperCase();
        return ch === up ? ch.toLowerCase() : up;
    });
}
function fileSystemEntryExists(path, entryKind) {
    try {
        const stat = _fs.statSync(path);
        switch (entryKind) {
            case 0 /* FileSystemEntryKind.File */:
                return stat.isFile();
            case 1 /* FileSystemEntryKind.Directory */:
                return stat.isDirectory();
            default:
                return false;
        }
    }
    catch (e) {
        return false;
    }
}
class NodeSystem {
    constructor() {
        this.args = process.argv.slice(2);
        this.newLine = _os.EOL;
        this.useCaseSensitiveFileNames = this.isFileSystemCaseSensitive();
        this.cwd = process.cwd();
    }
    isFileSystemCaseSensitive() {
        // win32\win64 are case insensitive platforms
        if (platform === 'win32' || platform === 'win64') {
            return false;
        }
        // If this file exists under a different case, we must be case-insensitve.
        return !this.fileExists(swapCase(__filename));
    }
    directoryExists(path) {
        return fileSystemEntryExists(path, 1 /* FileSystemEntryKind.Directory */);
    }
    getDirectories(path) {
        return _fs
            .readdirSync(path)
            .filter(dir => fileSystemEntryExists(this.resolvePath(path, dir), 1 /* FileSystemEntryKind.Directory */));
    }
    write(s) {
        process.stdout.write(s);
    }
    writeOutputIsTTY() {
        return process.stdout.isTTY || false;
    }
    readFile(fileName, _encoding) {
        if (!this.fileExists(fileName)) {
            return undefined;
        }
        const buffer = _fs.readFileSync(fileName);
        let len = buffer.length;
        if (len >= 2 && buffer[0] === 0xfe && buffer[1] === 0xff) {
            // Big endian UTF-16 byte order mark detected. Since big endian is not supported by node.js,
            // flip all byte pairs and treat as little endian.
            len &= ~1; // Round down to a multiple of 2
            for (let i = 0; i < len; i += 2) {
                const temp = buffer[i];
                buffer[i] = buffer[i + 1];
                buffer[i + 1] = temp;
            }
            return buffer.toString('utf16le', 2);
        }
        if (len >= 2 && buffer[0] === 0xff && buffer[1] === 0xfe) {
            // Little endian UTF-16 byte order mark detected
            return buffer.toString('utf16le', 2);
        }
        if (len >= 3 && buffer[0] === 0xef && buffer[1] === 0xbb && buffer[2] === 0xbf) {
            // UTF-8 byte order mark detected
            return buffer.toString('utf8', 3);
        }
        // Default is UTF-8 with no byte order mark
        return buffer.toString('utf8');
    }
    writeFile(fileName, data, writeByteOrderMark) {
        // If a BOM is required, emit one
        if (writeByteOrderMark) {
            // tslint:disable-next-line:no-parameter-reassignment
            data = byteOrderMarkIndicator + data;
        }
        let fd;
        try {
            fd = _fs.openSync(fileName, 'w');
            _fs.writeSync(fd, data, /*position*/ undefined, 'utf8');
        }
        finally {
            if (fd !== undefined) {
                _fs.closeSync(fd);
            }
        }
    }
    resolvePath(...paths) {
        return _path.resolve(...paths);
    }
    fileExists(path) {
        return fileSystemEntryExists(path, 0 /* FileSystemEntryKind.File */);
    }
    getCurrentDirectory() {
        return this.cwd;
    }
    relative(from, to) {
        return _path.relative(from, to);
    }
    createDirectory(directoryName) {
        if (!this.directoryExists(directoryName)) {
            _fs.mkdirSync(directoryName);
        }
    }
}
exports.NodeSystem = NodeSystem;
exports.nodeSystem = new NodeSystem();
//# sourceMappingURL=NodeSystem.js.map