"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.printErrors = exports.LysError = void 0;
const LineMapper_1 = require("./LineMapper");
const astPrinter_1 = require("./astPrinter");
const colors_1 = require("./colors");
function mapSet(set, fn) {
    const out = [];
    set.forEach($ => out.push(fn($)));
    return out;
}
class LysError extends Error {
}
exports.LysError = LysError;
const ansiRegex = new RegExp([
    '[\\u001B\\u009B][[\\]()#;?]*(?:(?:(?:[a-zA-Z\\d]*(?:;[a-zA-Z\\d]*)*)?\\u0007)',
    '(?:(?:\\d{1,4}(?:;\\d{0,4})*)?[\\dA-PRZcf-ntqry=><~]))'
].join('|'), 'g');
function printErrors(parsingContext, stripAnsi = false, messageCollector) {
    const errorByFile = new Map();
    const messageCollectorToPrint = messageCollector || parsingContext.messageCollector;
    messageCollectorToPrint.errors.forEach($ => {
        const document = ($.position && $.position.moduleName) || '(no document)';
        if (!errorByFile.has(document)) {
            errorByFile.set(document, []);
        }
        errorByFile.get(document).push($);
    });
    const out = [];
    errorByFile.forEach((errors, moduleName) => {
        out.push(printErrors_(moduleName, parsingContext, errors, stripAnsi));
    });
    return out.join('\n');
}
exports.printErrors = printErrors;
function printErrors_(moduleName, parsingContext, errors, stripAnsi = false) {
    const printLines = [];
    const parsing = parsingContext.getExistingParsingPhaseForModule(moduleName);
    if (!parsing) {
        throw new Error('Cannot find moduleName:' + moduleName);
        errors.forEach((err) => {
            printLines.push((0, colors_1.formatColorAndReset)(err.message, colors_1.ForegroundColors.Red));
            if (err.stack) {
                printLines.push((0, astPrinter_1.indent)((0, colors_1.formatColorAndReset)(err.stack, colors_1.ForegroundColors.Grey), '  '));
            }
        });
        if (stripAnsi) {
            return printLines.join('\n').replace(ansiRegex, '');
        }
        return printLines.join('\n');
    }
    let fileNameToPrint = parsing.fileName || moduleName;
    fileNameToPrint = fileNameToPrint.replace(parsingContext.system.getCurrentDirectory() + '/', '');
    fileNameToPrint = fileNameToPrint.replace(parsingContext.system.getCurrentDirectory() + '\\', '');
    // normalize filename for windows
    fileNameToPrint = fileNameToPrint.replace(/\\/g, '/');
    printLines.push((0, colors_1.formatColorAndReset)(fileNameToPrint || '(no file)', colors_1.gutterStyleSequence));
    const source = parsing.content;
    let lineMapper = new LineMapper_1.LineMapper(source, Math.random().toString());
    if (errors.length === 0)
        return '';
    lineMapper.initMapping();
    let lines = source.split(/\r\n|\r|\n/g);
    const errorOnLines = new Array(lines.length + 1).fill(null).map(_ => new Set());
    const printableErrors = new Set();
    errors.forEach((err) => {
        try {
            if (!err.position) {
                throw null;
            }
            const start = lineMapper.position(err.position.start);
            const end = lineMapper.position(err.position.end);
            if (start.line >= 0) {
                errorOnLines[start.line] = errorOnLines[start.line] || new Set();
                const error = { start, end, message: err.message || err.toString(), stack: err.stack, warning: err.warning };
                printableErrors.add(error);
                for (let l = start.line; l <= end.line; l++) {
                    errorOnLines[l].add(error);
                }
            }
        }
        catch (e) {
            printableErrors.add({
                start: null,
                end: null,
                message: err.message || err.toString(),
                stack: err.stack,
                warning: false
            });
        }
    });
    const blackPadding = (0, colors_1.formatColorAndReset)('     ' + colors_1.gutterSeparator, colors_1.gutterStyleSequence);
    const ln = (n) => (0, colors_1.formatColorAndReset)(('    ' + (n + 1).toString()).substr(-5) + '' + colors_1.gutterSeparator, colors_1.gutterStyleSequence);
    const printedErrors = new Set();
    const printableLines = [];
    const linesAbove = 10;
    let lastLine = 0;
    errorOnLines.forEach((x, line) => {
        if (x && x.size) {
            for (let i = 0; i < linesAbove; i++) {
                printableLines[Math.max(line - i, 0)] = true;
                printableLines[Math.min(line + i, lines.length)] = true;
            }
            printableLines[line] = true;
        }
    });
    if (!errors.length) {
        lines.forEach((_, y) => {
            printableLines[y] = true;
        });
    }
    printableLines.forEach((printable, i) => {
        if (!printable)
            return;
        let line = lines[i];
        if (i !== lastLine) {
            printLines.push((0, colors_1.formatColorAndReset)('  ...' + colors_1.gutterSeparator, colors_1.gutterStyleSequence));
        }
        lastLine = i + 1;
        if (typeof line !== 'string') {
            return;
        }
        if (errorOnLines[i] && errorOnLines[i].size) {
            printLines.push(ln(i) +
                (0, colors_1.formatColorAndReset)(line, colors_1.ForegroundColors.ErrorLine) +
                mapSet(errorOnLines[i], x => {
                    let message = '';
                    const color = x.warning ? colors_1.ForegroundColors.Yellow : colors_1.ForegroundColors.Red;
                    if (x.end && x.start) {
                        if (i === x.end.line || x.end.line === x.start.line) {
                            if (x.start.column <= x.end.column && x.end.line === x.start.line) {
                                message = '\n' + blackPadding + new Array(x.start.column + 1).join(' ');
                                message =
                                    message + (0, colors_1.formatColorAndReset)(new Array(x.end.column + 1 - x.start.column).join('^'), color) + ' ';
                            }
                            else {
                                message = '\n' + blackPadding + new Array(x.end.column).join(' ');
                                message = message + (0, colors_1.formatColorAndReset)('^ ', color);
                            }
                            message = message + (0, colors_1.formatColorAndReset)(x.message, color);
                        }
                        printedErrors.add(x);
                    }
                    return message;
                }).join(''));
        }
        else {
            printLines.push(ln(i) + line);
        }
    });
    if (lines.length !== lastLine - 1) {
        printLines.push((0, colors_1.formatColorAndReset)('  ...' + colors_1.gutterSeparator, colors_1.gutterStyleSequence));
    }
    printableErrors.forEach(err => {
        if (!printedErrors.has(err)) {
            printLines.push((0, colors_1.formatColorAndReset)(err.message, colors_1.ForegroundColors.Red));
        }
    });
    if (stripAnsi) {
        return printLines.join('\n').replace(ansiRegex, '');
    }
    return printLines.join('\n');
}
//# sourceMappingURL=errorPrinter.js.map